'''
/*****************************************************************************************************
*                                   NOTICE
*
* THIS COMPUTER SOFTWARE CONTAINS PROPRIETARY INFORMATION OF USDA-ARS/SPRU.
* NEITHER RECEIPT NOR POSSESSION THEREOF COVERS ANY RIGHT TO
* REPRODUCE OR USE, OR DISCLOSE IN WHOLE OR IN PART, ANY SUCH INFORMATION
* WITHOUT WRITTEN AUTHORIZATION FROM USDA-ARS/SPRU.  
* 
* ****************************************************************************************************
* Program: IMAP (Image Mapping & Analytics for Phenotyping)
* ****************************************************************************************************
* Purpose: HTP Image Analytics GUI
* Author: James Kim (james.y.kim@usda.gov)
* Version:  V1.2
* Date:  2018-12-11  
* Revision: 2024-03-18
* ****************************************************************************************************/
'''

path = '../IMAP/'
import tkinter as tk #cf)Tkinter for Python 2.*
import tkinter.filedialog   #cf)tkFileDialog for Python 2.*
from tkinter import messagebox,simpledialog,ttk

import cv2
import numpy as np
from PIL import ImageTk, Image, ImageStat
from matplotlib import pyplot as plt
from matplotlib.path import Path as pltPath

from osgeo import gdal, osr, ogr
import osgeo.gdalnumeric as gdn
import rasterio
#from geographiclib.geodesic import Geodesic
#from pyproj import Proj

import os, time, sys, struct
import re
import skimage as sk    #50MB, install scikit-image
from math import * 
import imutils as imu
import json as js
import fnmatch
import keyboard as kb

#Local libraries
sys.path.append('Lib/')
import Pic2Video as P2V
#import fmCalendar as fmC

#list array: row=len(im[0]), col=len(im[0][0])
#ds: image raster(BGRN), ds.GetGeoTransform() #ulx, xres, xskew, uly, yskew, yres
#im: image array (cv2:BGR,np:RGB), im.shape (0:H,1:W,2:ch), im.dtype, ch0=im[:,:,0], rgb=im[:,:,:3], px=im[:,:,0][y,x], cv2.imshow('im',im), cv2.imwrite(fname,im), crop:im[y0:y1,x0:x1]
#img: image matrix (PIL.Image.open:RGB,mat.image.imread:RGB) img.size (0:W,1:H), img.format, img.getchannel(0)=img.split()[0], px=img.getpixel((x,y)), img.show(), img.save(fname), img.crop(x1,y1,x2,y2), img.close(), img.mode:RGB
#RGB->BGR for cv2: im[:,:,::-1] or cv2.cvtColor(im,cv2.COLOR_RGB2BGR)
#im2img: img = Image.fromarray(im)   
#img2im: im = np.array(img) 
#gdal2cv: im = ds.ReadAsArray(), im = np.transpose(im, [1,2,0])
#pass parameter to event handling: command=lambda: callback(1)

# Convert 1b image to heatmap based on ROI grid
def ConvertIm2Heatmap(im,ROI):  
    X1, Y1, X2, Y2  = ROI[0:4]
    imGRID = im[Y1:Y2,X1:X2]
    minP,maxP = np.nanpercentile(imGRID,(5,95))
    #print(np.max(imGRID),'percentile_GRID=',minP,maxP)
    colormap = plt.get_cmap('RdYlGn')
    heatmap = (colormap(im/maxP) * 2**8).astype(np.uint8)[:,:,:3]
    heatmap = cv2.cvtColor(heatmap, cv2.COLOR_RGB2BGR)                        
    #heatmap = cv2.applyColorMap(im, cv2.COLORMAP_DEEPGREEN)
    del imGRID,colormap    
    return heatmap

    
def getFence(x0, y0, x1, y1, x2, y2):        
    if x1 == x2: # vertical line
        if x0 < x1: return 'WW'
        elif x0 > x1: return 'EE'
        else:       return 'On'
    else: m = (y2-y1)/(x2-x1)    

    if y1 == y2: # horizontal line
        if y0 < y1: return 'SS'
        elif y0 > y1: return 'NN'
        else:       return 'On'

    # check y-coord on y = m*x + (y1 - m*x1)
    NS='O'
    if y0 == m*x0 + (y1 - m*x1): return 'On'
    elif y0 < m*x0 + (y1 - m*x1): NS = 'S'
    else:                         NS = 'N'

    # check x-coord on (y - (y1 - m*x1))/m = x 
    if x0 == (y0 - (y1 - m*x1))/m: return 'On'
    elif x0 < (y0 - (y1 - m*x1))/m: return NS+'W'
    else:                           return NS+'E'

# check GeoFence
def isInFence(E0, N0, FB):
    if E0>0:
        #polygon = [(FB[0],FB[1]),(FB[2],FB[3]),(FB[4],FB[5]),(FB[6],FB[7])]
        #path = pltPath(polygon)
        #points = [(E0,N0)]
        #inside = path.contains_points(points)
        ##print('inside=',inside)
        #return inside
    
        GF = [None]*4
        #print(' GF=',end='')
        for i in range(4):
            GF[i] = getFence(E0,N0,FB[(i*2)%8],FB[(i*2+1)%8],FB[(i*2+2)%8],FB[(i*2+3)%8]) 
            #print(GF[i],end=',')
            if GF[i]=='On': return True
        if GF[0][0]=='S' and GF[1][1]=='W' and GF[2][0]=='N' and GF[3][1]=='E': return True
        else: return False

# Extract ROI px within tROI by GeoFencing, NOET: verification only due to computation load
def Extract_ROI_px(tROI):        
    i1,j1,i2,j2,i3,j3,i4,j4 = tROI[:8]  # NW(i1,j1), SE(i2,j2), NE(i3,j3), SW(i4,j4)

    ROIpx = []

    # calculate slope (a) & y-intercept (b)
    if (i4-i1) == 0: # non-rotated rectangle
        x_min,y_min,x_max,y_max = min(i1,i2,i3,i4), min(j1,j2,j3,j4), max(i1,i2,i3,i4), max(j1,j2,j3,j4)
        for j in range(j1,j2+1):
            for i in range(i1,i2+1):
                ROIpx.append([i,j])    
    else:
        x1,x2,x3,x4 = i1,i2,i3,i4   # reverse j-sign for ij->xy coord
        y1,y2,y3,y4 = -j1,-j2,-j3,-j4   # reverse j-sign for ij->xy coord
        FB = [x1,y1,x3,y3,x2,y2,x4,y4]  # NW(x1,y1), NE(x3,y3), SE(x2,y2), SW(x4,y4)

        x_min,y_min,x_max,y_max = min(x1,x2,x3,x4), min(y1,y2,y3,y4), max(x1,x2,x3,x4), max(y1,y2,y3,y4)
        
        for y in range(y_max,y_min,-1):    # reverse y-order for xy->ij coord
            for x in range(x_min,x_max):
                # check Geo-Fence                
                ret1 = isInFence(x,y,FB)       
                if ret1:
                    i,j = x,-y   # reverse back y-sign for xy->ij coord
                    ROIpx.append([i,j])    
    return ROIpx

# Extract ROI px within standing ROI
def Extract_ROIpx(ROI):        
    i1,j1,i2,j2 = ROI[:4]  # NW(i1,j1), SE(i2,j2), NE(i3,j3), SW(i4,j4)
    ROIpx = []
    for j in range(j1,j2+1):
        for i in range(i1,i2+1):
            ROIpx.append([i,j])       
    return ROIpx

# Extract subROI px within a single 4-corner subtROI
def Extract_tROIpx(tROI):        
    if len(tROI)>=8:
        i1,j1,i2,j2,i3,j3,i4,j4 = tROI[:8]  # NW(i1,j1), SE(i2,j2), NE(i3,j3), SW(i4,j4)
    else: 
        i1,j1,i2,j2 = tROI[:4]  # NW(i1,j1), SE(i2,j2), NE(i3,j3), SW(i4,j4)
        i3,j3,i4,j4 = i2, j1, i1, j2
        #messagebox.showerror(title='Error',message='Missing ROI! Redraw a ROI!')
        #return tROI
    #print('tROI=',tROI)    
    
    ROIpx = []

    # calculate slope (a) & y-intercept (b)
    if (i4-i1) == 0: ROIpx = Extract_ROIpx(tROI) # non-rotated rectangle
    else:
        x1,x2,x3,x4 = i1,i2,i3,i4   # reverse j-sign for ij->xy coord
        y1,y2,y3,y4 = -j1,-j2,-j3,-j4   # reverse j-sign for ij->xy coord

        x_min,y_min,x_max,y_max = min(x1,x2,x3,x4), min(y1,y2,y3,y4), max(x1,x2,x3,x4), max(y1,y2,y3,y4)
                
        if x3==x1 or x2==x4: ROIpx = Extract_ROIpx(tROI) # undefinable rectangle
        else:
            a1 = (y3-y1)/(x3-x1) # slope
            b1 = y1 - a1*x1      # NW(x1,y1) on y = a1*x + b1 (Top W-E)
            #a3 = a1 # NOT extactly equal when low resol, get it from 2-pt
            a3 = (y2-y4)/(x2-x4) # slope
            b3 = y2 - a3*x2     # SE(x2,y2) on y = a3*x + b3 (Bottom W-E)
            if x3==x2 or x1==x4 or y1==y3 or y2==y4: ROIpx = Extract_ROIpx(tROI) # non-rotated rectangle
            else:
                #a2 = -1/a1  # NOT extactly equal when low resol, get it from 2-pt
                a2 = (y3-y2)/(x3-x2) # slope
                b2 = y2 - a2*x2     # SE(x2,y2) on y = a2*x + b4 (Right N-S)        
                #a4 = a2 # NOT extactly equal when low resol, get it from 2-pt
                a4 = (y1-y4)/(x1-x4) # slope
                b4 = y1 - a4*x1     # NW(x1,y1) on y = a4*x + b2 (Left N-S)
                #print('a1,b1,a2,b2,a3,b3,a4,b4=',a1,b1,a2,b2,a3,b3,a4,b4)
            
                # find subROI px by geo-fencing
                #CntIn, CntOut = 0, 0
                for y in range(y_max,y_min-1,-1):    # reverse y-order for xy->ij coord
                    for x in range(x_min,x_max+1):
                        min13, max13 = min(a3*x+b3,a1*x+b1), max(a3*x+b3,a1*x+b1)
                        min24, max24 = min(a4*x+b4,a2*x+b2), max(a4*x+b4,a2*x+b2)
                        if min13 < y <= max13 and min24 < y <= max24:
                            #print('x,y=',x,y,'min13,max13=',min13,max13,'min24,max24=',min24,max24,'others=',a3*x+b3,a1*x+b1,a4*x+b4,a2*x+b2,a3*x+b3<y,y<a1*x+b1,a4*x+b4<y,y<a2*x+b2)
                            i,j = x,-y   # reverse back y-sign for xy->ij coord
                            ROIpx.append([i,j])    
                            #CntIn = CntIn+1
                        #else: CntOut = CntOut+1
        #print('x_min,y_min,x_max,y_max=',x_min,y_min,x_max,y_max)
        #print('CntIn, CntOut=',CntIn, CntOut,'total=',CntIn+CntOut)
        InnerBoundary = [x_min,-y_max,x_max,-y_min] # reverse back y-sign and y_min*y_max for xy->ij coord
    return ROIpx
    
# Add a ROI grid on the image
def Add_Grid(im, ROI, GRID, color=None, LineSize=None, rot=None):        
    if len(im.shape)>2 and im.shape[2]==3:
        im = cv2.cvtColor(im,cv2.COLOR_RGB2BGR)   # RGB->BGR for cv2 operations
    if color == None: 
        color = (255,0,0)   #BGR: cyon=(255,255,0)
    if LineSize==None: LineSize = 4
    
    if rot==0:
        cv2.rectangle(im, (ROI[0],ROI[1]), (ROI[2],ROI[3]), color, LineSize)   # draw an outer box, Cyan when cvtColor(im,cv2.COLOR_BGR2RGB)
        # draw inner lines in the outer box
        if(GRID[0]>1):  #Row
            for i in range(1,GRID[0]):
                xx1 = ROI[0]  
                yy1 = ROI[1] + round((i)*(ROI[3]-ROI[1])/GRID[0])
                xx2 = ROI[2]
                yy2 = yy1
                cv2.line(im, (xx1,yy1), (xx2,yy2), color, LineSize)   # H-grid in ROI, Cyan when cvtColor(im,cv2.COLOR_BGR2RGB)
        if(GRID[1]>1):   #Col
            for i in range(1,GRID[1]):
                xx1 = ROI[0] + round((i)*(ROI[2]-ROI[0])/GRID[1])
                yy1 = ROI[1]
                xx2 = xx1
                yy2 = ROI[3]
                cv2.line(im, (xx1,yy1), (xx2,yy2), color, LineSize)   # V-grid in ROI, Cyan when cvtColor(im,cv2.COLOR_BGR2RGB)
    else:    
        #print('_ROI=',ROI)
        rotGRID,a,b,r,min_theta, max_theta = Calculate_GRID_corners([ROI],rot) 
        i1,j1,i2,j2,i3,j3,i4,j4 = rotGRID
        R,C = GRID
        #print('im.shape=',im.shape)
        for j in range(R):        
            for i in range(C):    
                if i==0:
                    hi1,hj1,hi3,hj3 = round((i4-i1)*j/R+i1), round((j4-j1)*j/R+j1), round((i2-i3)*j/R+i3), round((j2-j3)*j/R+j3)
                    hi4,hj4,hi2,hj2 = round((i4-i1)*(j+1)/R+i1), round((j4-j1)*(j+1)/R+j1), round((i2-i3)*(j+1)/R+i3), round((j2-j3)*(j+1)/R+j3)
                    cv2.line(im, (hi1,hj1), (hi3,hj3), color, LineSize)   # H-grid in ROI, Cyan when cvtColor(im,cv2.COLOR_BGR2RGB)
                    cv2.line(im, (hi4,hj4), (hi2,hj2), color, LineSize)   # H-grid in ROI, Cyan when cvtColor(im,cv2.COLOR_BGR2RGB)
                    #print('hi1,hj1,hi3,hj3=',hi1,hj1,hi3,hj3,'hi4,hj4,hi2,hj2=',hi4,hj4,hi2,hj2)
                if j==0:
                    vi1,vj1,vi4,vj4 = round((i3-i1)*i/C+i1), round(j1-(j1-j3)*i/C), round((i2-i4)*i/C+i4), round(j4-(j4-j2)*i/C)                
                    vi3,vj3,vi2,vj2 = round((i3-i1)*(i+1)/C+i1), round(j1-(j1-j3)*(i+1)/C), round((i2-i4)*(i+1)/C+i4), round(j4-(j4-j2)*(i+1)/C)                
                    cv2.line(im, (vi1,vj1), (vi4,vj4), color, LineSize)   # H-grid in ROI, Cyan when cvtColor(im,cv2.COLOR_BGR2RGB)
                    cv2.line(im, (vi3,vj3), (vi2,vj2), color, LineSize)   # H-grid in ROI, Cyan when cvtColor(im,cv2.COLOR_BGR2RGB)
                    #print('vi1,vj1,vi4,vj4=',vi1,vj1,vi4,vj4,'vi3,vj3,vi2,vj2=',vi3,vj3,vi2,vj2)

    #cv2.imshow('im',im), cv2.waitKey(0)
    if len(im.shape)>2 and im.shape[2]==3:
        im = cv2.cvtColor(im,cv2.COLOR_BGR2RGB)   # BGR->RGB when cv2 operations done for display
    return im    

# a subclass of Canvas for dealing with resizing of windows
class ResizingCanvas(tk.Canvas):
    def __init__(self,parent,**kwargs):
        tk.Canvas.__init__(self,parent,**kwargs)
        self.bind("<Configure>", self.on_resize)
        self.height = self.winfo_reqheight()
        self.width = self.winfo_reqwidth()

    def on_resize(self,event):
        # determine the ratio of old width/height to new width/height
        wscale = float(event.width)/self.width
        hscale = float(event.height)/self.height
        self.width = event.width
        self.height = event.height
        # resize the canvas 
        self.config(width=self.width, height=self.height)
        # rescale all the objects tagged with the "all" tag
        self.scale("all",0,0,wscale,hscale)
        print('HxW=',self.height,self.width)
        
# Calculates the heading between two points.
def Calculate_Heading(LL1, LL2):
    """
    Parameters:
      - LL1: tuple for lat/lon (deg) of 1st point.
      - LL2: tuple for lat/lon (deg) of  2nd point.
    The formulae: theta = atan2(sin(dlong)*cos(lat2), cos(lat1)*sin(lat2) - sin(lat1)*cos(lat2)*cos(dlong))
    Returns: The heading in radian
    Returns Type: float
    """
    if (type(LL1) != tuple) or (type(LL2) != tuple):
        raise TypeError("Only tuples are supported as arguments")

    lat1 = radians(LL1[0])
    lat2 = radians(LL2[0])
    diffLong = radians(LL2[1] - LL1[1])
    y = sin(diffLong)*cos(lat2)
    x = cos(lat1)*sin(lat2) - sin(lat1)*cos(lat2)*cos(diffLong)

    initial_heading = atan2(y, x)
    # normalize the initial heading, as atan2 return values from -180 to + 180 which is not for a compass heading
    compass_heading = (initial_heading + 2*pi) % (2*pi)
    
    return compass_heading

# correct GPS coords by applying offsets
def Correct_GPSoffset(Easting0,Northing0,a,b,yaw):          
    Easting = Easting0 + sqrt(a**2+b**2)*sin(yaw+atan(b/abs(a)))    # NOTE: abs() needed to avoid double negatives of a & b. Negativity is determined by b. 
    Northing = Northing0 + sqrt(a**2+b**2)*cos(yaw+atan(b/abs(a)))
    #Easting = Easting0 + (a*sin(yaw)+b*cos(yaw))    # takes 10 sec longer
    #Northing = Northing0 + (a*cos(yaw)-b*sin(yaw))
    return Easting,Northing

def Create_Polygon(coords): #Lat, Lon
    ring = ogr.Geometry(ogr.wkbLinearRing)
    for coord in coords:
        ring.AddPoint(coord[0], coord[1])

    # Create polygon
    poly = ogr.Geometry(ogr.wkbPolygon)
    poly.AddGeometry(ring)
    return poly.ExportToWkt()

# Read a shapefile to access vertices
def Read_Shapefile(in_shp):
    file = ogr.Open(in_shp)
    layer = file.GetLayer(0)
    no_of_points = layer.GetFeatureCount()
    print('no_of_points=',no_of_points)            
    
    ## get projected spatial reference
    #sr = layer.GetSpatialRef()
    ## get geographic spatial reference 
    #geogr_sr = sr.CloneGeogCS()    
    ## define reprojection
    #proj_to_geog = osr.CoordinateTransformation(sr, geogr_sr)    

    poly = [0]*4  # a polygon w [Lat,Lon] of 4 corners
    polyE = []    # a set of polygons extended 2D  
    polys = []    # a set of polygons 3D
    point = [0]   # a point w [Lat,Lon]
    pointE = []   # a set of points extended 2D  
    points = []   # a set of points 3D
    for i,area in enumerate(layer):
        area_shape = area.GetGeometryRef()
        #print('area_shape=',area_shape)
        #print (area_shape.Centroid().ExportToWkt())    

        # transform coordinates
        #area_polygon.Transform(proj_to_geog)        
                
        if area_shape.GetGeometryName() == 'POLYGON':
            area_polygon = area_shape.GetGeometryRef(0)
            #print('area_polygon=',area_polygon)
            no_of_polygon_vertices = area_polygon.GetPointCount()
            #print(i,'no_of_polygon_vertices=',no_of_polygon_vertices)
            for j in range(no_of_polygon_vertices-1): 
                lon, lat, z = area_polygon.GetPoint(j)   # can be either LonLat or EastingNorthing    
                poly[j] = [lat,lon]
                #print(i,j,'lon, lat=',lon, lat)
            polyE.extend(poly)  # NOTE: adding every item in the iterable objects to the list cf)append: adding one object of any type to a list
            #print(i,'polyE=',polyE)            
        elif area_shape.GetGeometryName() == 'POINT':
            E, N, z = area_shape.GetPoint()   # can be either LonLat or EastingNorthing    
            point = [E,N]
            #print(i,'E, N=',E, N)
            #pointE.extend(point)  # NOTE: adding every item in the iterable objects to the list cf)append: adding one object of any type to a list
            pointE.append(point)  # NOTE: adding every item in the iterable objects to the list cf)append: adding one object of aimH type to a list
            #print(i,'pointE=',pointE)            
        else:
            messagebox.showerror(title='Error',message='Neither POLYGON nor POINT! Please load POLYGON or POINT.')
            return
    
    if area_shape.GetGeometryName() == 'POLYGON':
        for i in range(0,len(polyE),+4): polys.append(polyE[i:4+i])
        #print(i,'polys=',polys)
        return 'POLYGON', polys
    elif area_shape.GetGeometryName() == 'POINT':
        #print('pointE=',pointE)        
        return 'POINT', pointE    

# Convert polygons to a shapefile with OGR    
def Write_shapefile(fdir,plots,file=None):    
    print('plots=',plots)
    print('file=',file)
    
    #file = "C:/Users/james.kim2/Documents/Research/Python/IMAP/Config/ROI.txt"
    if file == None: 
        filename = tkinter.filedialog.asksaveasfilename(initialdir=fdir,title='Save as a shp file',filetypes = (("shp files","*.shp"),("all files","*.*")))
    else: filename = file

    #GRID,ROI,ROI_rot,plots = Read_ROI(filename)    
    #print('GRID=',GRID)
    #print('ROI=',ROI)
    
    # convert ROI to GPS coord.
    #ConvertROI2UTM()
    
    fdir,fname,fbasename,ftype = ExtractFilename(filename)
    out_shp = fdir + fbasename + '.shp'
    print('out_shp=',out_shp)        

    # set up the shapefile driver
    driver = ogr.GetDriverByName('ESRI Shapefile')
    
    # create the data source
    data_source = driver.CreateDataSource(out_shp)
    
    # create the spatial reference, WGS84
    srs = osr.SpatialReference()
    srs.ImportFromEPSG(4326)
    
    # create the layer
    layer = data_source.CreateLayer('', srs, ogr.wkbPolygon)    # cf) ogr.wkbPoint    
    
    # Add the field
    #field_name = ogr.FieldDefn("Name", ogr.OFTString)
    #field_name.SetWidth(24)    
    #layer.CreateField(field_name)
    layer.CreateField(ogr.FieldDefn('id', ogr.OFTInteger))    
    #layer.CreateField(ogr.FieldDefn("Latitude", ogr.OFTReal))

    polys = []
    for i in range(len(plots)):        
        plots[i].append(plots[i][0])    # add the first coord to close a polygon
        coords = plots[i]        
        #coords = [(-106.6472953, 24.0370137), (-106.4933356, 24.05293569), (-106.4941789, 24.01969175), (-106.4927777, 23.98804445), (-106.4922614, 23.95582128), (-106.4925834, 23.92302327), (-106.4924068, 23.89048039), (-106.4925695, 23.85771361), (-106.4932479, 23.82457675), (-106.4928676, 23.7922049), (-106.4925072, 23.75980241), (-106.492388, 23.72722475), (-106.4922574, 23.69464296), (-106.4921181, 23.6620529), (-106.4922734, 23.62926733), (-106.4917201, 23.59697561), (-106.4914134, 23.56449628), (-106.4912558, 23.5319045), (-106.491146, 23.49926362), (-106.4911734, 23.46653561), (-106.4910181, 23.43392476), (-106.4910156, 23.40119976), (-106.4909501, 23.3685223), (-106.4908165, 23.33586566), (-106.4907735, 23.30314904), (-106.4906954, 23.27044931), (-106.4906366, 23.23771759), (-106.4905894, 23.20499124), (-106.4905432, 23.17226022), (-106.4904748, 23.1395177), (-106.4904187, 23.10676788), (-106.4903676, 23.07401321), (-106.4903098, 23.04126832), (-106.4902512, 23.00849426), (-106.4901979, 22.97572025), (-106.490196, 22.97401001), (-106.6481193, 22.95609832), (-106.6481156, 22.95801668), (-106.6480697, 22.99082052), (-106.6480307, 23.02362441), (-106.6479937, 23.0563977), (-106.6479473, 23.0891833), (-106.647902, 23.12196713), (-106.6478733, 23.15474057), (-106.6478237, 23.18750353), (-106.6477752, 23.22026138), (-106.6477389, 23.25302505), (-106.647701, 23.28577123), (-106.6476562, 23.31851549), (-106.6476211, 23.3512557), (-106.6475745, 23.38397935), (-106.6475231, 23.41671055), (-106.6474863, 23.44942382), (-106.6474432, 23.48213255), (-106.6474017, 23.51484861), (-106.6474626, 23.54747418), (-106.647766, 23.57991134), (-106.6482374, 23.61220905), (-106.6484783, 23.64467084), (-106.6482308, 23.6775148), (-106.6479338, 23.7103854), (-106.6478395, 23.74309074), (-106.6472376, 23.77618646), (-106.6472982, 23.80876072), (-106.647127, 23.84151129), (-106.6471277, 23.8741312), (-106.6473995, 23.90656505), (-106.6473138, 23.93916488), (-106.6473408, 23.97172031), (-106.6473796, 24.00435261), (-106.6472953, 24.0370137)]
        #print(i,'coords=',coords)
        polys.append(Create_Polygon(coords))            
        #print(i,'ploys=',polys)
    #print('polys[0]=',polys[0])
    
    print('len(polys)=',len(polys))
    # Process the text file and add the attributes and features to the shapefile
    for i in range(len(polys)):
        # create the WKT for the feature using Python string formatting
        #wkt = "POINT(%f %f)"%(float(row['Longitude']), float(row['Latitude']))
        wkt = polys[i]
    
        # Create a new feature (attribute and geometry)
        feature = ogr.Feature(layer.GetLayerDefn())
        
        # Set the attributes using the values from the delimited text file
        feature.SetField('id', 123+i)
        #feature.SetField("Latitude", row['Lat'])
            
        # Create the point from the Well Known Txt
        point = ogr.CreateGeometryFromWkt(wkt)
                
        # Set the feature geometry using the point
        feature.SetGeometry(point)

        # Create the feature in the layer (shapefile) 
        layer.CreateFeature(feature)
    
    # Dereference the feature
    data_source = layer = feature = geom = None    

# Convert LL in csv to a shapefile
def Convert_CSV2SHP(fdir, file=None): 
    #file = r"C:\Users\james.kim2\Documents\Research\Python\Border\FD_ALARC.csv"
    if file == None: filename = LoadFile(fdir,'CSV')     
    else: filename = file
    # exit if canceled to open a file
    if len(filename)==0:    
        return   
    
    FD_GPS = Read_FD_GPS(filename) #(Lon, Lat)
    FDs = list(FD_GPS.values())
    print('len(FDs)=',len(FDs))
    print('FDs=',FDs)
    
    fdir,fname,fbasename,ftype = ExtractFilename(filename)
    out_shp = fdir + fbasename + '.shp'
    print('out_shp=',out_shp)
    
    Write_shapefile(fdir, FDs, out_shp)    
    print('done')

# convert ROI to UTM coord.
def Convert_ROI2LL(ROI,ds,flag_im):         
    gt = ds.GetGeoTransform()   #ulx, xres, xskew, uly, yskew, yres
    if gt:
        minx = gt[0]
        maxy = gt[3]  
        E_res = gt[1]
        N_res = gt[5]            
        
    # Read GPS coordinate from geotiff (a raster layer)                
    plots = []
    for i in range(len(ROI)):
        E_NW = minx + ROI[i][0]*E_res #UTM of original EPSG
        N_NW = maxy + ROI[i][1]*N_res #UTM of original EPSG
        E_SE = minx + ROI[i][2]*E_res #UTM of original EPSG
        N_SE = maxy + ROI[i][3]*N_res #UTM of original EPSG
        E_NE = minx + ROI[i][4]*E_res #UTM of original EPSG
        N_NE = maxy + ROI[i][5]*N_res #UTM of original EPSG
        E_SW = minx + ROI[i][6]*E_res #UTM of original EPSG
        N_SW = maxy + ROI[i][7]*N_res #UTM of original EPSG
        #Convert UTM to WGS 84 Lat\Lon
        if flag_im: 
            Lat_NW, Lon_NW = Convert_UTM2LL(E_NW,N_NW,None,ds.GetProjection())  #LL of WGS 84
            Lat_SE, Lon_SE = Convert_UTM2LL(E_SE,N_SE,None,ds.GetProjection())  #LL of WGS 84
            Lat_NE, Lon_NE = Convert_UTM2LL(E_NE,N_NE,None,ds.GetProjection())  #LL of WGS 84
            Lat_SW, Lon_SW = Convert_UTM2LL(E_SW,N_SW,None,ds.GetProjection())  #LL of WGS 84
        else:          
            Lat_NW, Lon_NW = Convert_UTM2LL(E_NW,N_NW,12)  #LL of WGS 84, UTM projection 12 (AZ) as default (change if differnt loc) 
            Lat_SE, Lon_SE = Convert_UTM2LL(E_SE,N_SE,12)  #LL of WGS 84, UTM projection 12 (AZ) as default (change if differnt loc) 
            Lat_NE, Lon_NE = Convert_UTM2LL(E_NE,N_NE,12)  #LL of WGS 84, UTM projection 12 (AZ) as default (change if differnt loc) 
            Lat_SW, Lon_SW = Convert_UTM2LL(E_SW,N_SW,12)  #LL of WGS 84, UTM projection 12 (AZ) as default (change if differnt loc) 
        
        plot = [(Lon_NW,Lat_NW),(Lon_NE,Lat_NE),(Lon_SE,Lat_SE),(Lon_SW,Lat_SW)]                
        plots.append(plot)
    #print('len(plots)=',len(plots),'plots=',plots)            
    return plots


#Load a file from Windows Explorer
def LoadFile(fdir,filetype=None):    
    print('fdir=',fdir)
    if filetype == None or filetype.lower() == "img":
        filename = tkinter.filedialog.askopenfilename(initialdir=fdir,title='Open an image file',filetypes = (("All Image Files","*.tif;*.bmp;*.jpg"),("TIF Files","*.tif"),("TIFF Files","*.tiff"),("BMP Files","*.bmp"),("JPEG Files","*.jpg"),("PNG Files","*.png"),("All Files","*.*")))
        #filename = path+"veg.jpg"            
    elif filetype.lower() == "txt":
        filename = tkinter.filedialog.askopenfilename(initialdir=fdir,title='Open a txt file',filetypes = (("TXT Files","*.txt"),("All Files","*.*")))
    elif filetype.lower() == "csv":
        filename = tkinter.filedialog.askopenfilename(initialdir=fdir,title='Open a csv datafile',filetypes = (("CSV Files","*.csv"),("All Files","*.*")))
    elif filetype.lower() == "xlsx":
        filename = tkinter.filedialog.askopenfilename(initialdir=fdir,title='Open a xlsx datafile',filetypes = (("XLSX Files","*.xlsx"),("All Files","*.*")))
    elif filetype.lower() == "shp":
        filename = tkinter.filedialog.askopenfilename(initialdir=fdir,title='Open a shp file',filetypes = (("SHP Files","*.shp"),("All Files","*.*")))
    elif filetype.lower() == "json":
        filename = tkinter.filedialog.askopenfilename(initialdir=fdir,title='Open a json file',filetypes = (("JSON Files","*.json"),("All Files","*.*")))
    else:
        messagebox.showerror(title='Error',message='File type was not found!')        
    #print('filename:%s\n'%(filename))
    return filename

def Read_ROI(filename=None):
    if(filename==None): filename = os.getcwd()+'\Config\ROI.txt'         
    print('filename=',filename)
    with open(filename) as f:
        lines = f.read().split('\n')        
    #print('lines=',len(lines),lines)
    
    # delete if a blank line
    for i,line in enumerate(lines):
        if line == "": del lines[i]
    #print('lines=',len(lines),lines)
    
    GRID = [0]*2    # [Row, Col]
    plot = [0]*4 # 4 corners of plot boundary
    plots = []
    for i,line in enumerate(lines):
        str_val = re.split(r', ',line) 
        #print (i,'str_val=',str_val)
        if i==0: 
            GRID[0] = int(str_val[0]) # GRID: RxC
            GRID[1] = int(str_val[1]) # GRID: RxC
            ROI_n = int(str_val[2])
            ROI_shape = str_val[3]
            ROI_rot = float(str_val[4])
            print ('GRID=',GRID,'ROI_n=',ROI_n,'ROI_shape=',ROI_shape,'ROI_rot=',ROI_rot)            
            ROI = [[0]*9 for i in range(ROI_n)] # 8 corners (NW,SE,NE,SW), ROI shape ('r' or 'c')
        if i>=1 and i<ROI_n+1: 
            for j in range(8):  # 8 values of 4 corners
                ROI[i-1][j] = int(str_val[j+1])            
            #print('len(str_val)=',len(str_val)) 
            if(len(str_val))>=17:
                for j in range(4):  # 4 xy-coord
                    plot[j] = (float(str_val[2*j+9]),float(str_val[2*j+10]))    #[Lon_NW,Lat_NW),(Lon_SE,Lat_NW),(Lon_SE,Lat_SE),(Lon_NW,Lat_SE)]
                plots.append(plot)
            ROI[i-1][8] = ROI_shape
    return GRID,ROI,ROI_rot,plots

# load a json to read GPS offsets
def LoadJSON(fdir, file=None, display=None):      
    if file == None or not os.path.exists(file): filename = LoadFile(fdir,'JSON')     
    else: filename = file
    if len(filename)==0:
        return        
    #print('filename=',filename)
    
    # read json contents
    jstr = open(filename,'r').read()
    jdata = js.loads(jstr)
    #print(jdata)
    #generate a dictionary for each sensor
    IRt = {}
    CC = {}
    Pe = {}
    Ho = {}
    for p in jdata['sensor_types']:
        #print('type: ' + p['type'])
        for i in range(len(p['sensors'])):
            #print('\t name: ', p['sensors'][i]['name'])
            #print('\t offset: ')
            #print('\t\t x: ', p['sensors'][i]['offset']['x'])
            #print('\t\t y: ', p['sensors'][i]['offset']['y'])
            #print('\t\t z: ', p['sensors'][i]['offset']['z'])    
            if p['type'] == 'cropcircle':
                CC[p['sensors'][i]['name'][-2:]] = [float(p['sensors'][i]['offset']['x']), float(p['sensors'][i]['offset']['y'])]
            if p['type'] == 'irt':
                IRt[p['sensors'][i]['name'][-2:]] = [float(p['sensors'][i]['offset']['x']), float(p['sensors'][i]['offset']['y'])]
            if p['type'] == 'pepperl':
                Pe[p['sensors'][i]['name'][-1]] = [float(p['sensors'][i]['offset']['x']), float(p['sensors'][i]['offset']['y'])]
            if p['type'] == 'honeywell':
                Ho[p['sensors'][i]['name'][-1]] = [float(p['sensors'][i]['offset']['x']), float(p['sensors'][i]['offset']['y'])]
    return CC,IRt,Pe,Ho

# read meta data of Avg platform
def Read_Platform4Avg(filename=None):
    if(filename==None): filename = os.getcwd()+'\Config\Avg_meta.dat'        
    #Detect # of lines
    num_lines = sum(1 for line in open(filename))
    print(num_lines,'filename=',filename)        
    
    lst = [0]*num_lines
    with open(filename) as f:
        for i, line in enumerate(f):    
            lst[i] = line.split(',')       
            #print (i,lst[i],lst[i][1])
            if   i == 0: CC2BB = float(lst[i][0])
            elif i == 1: yaw_off = float(lst[i][0])    #yaw offset in deg
            elif i == 2: a = float(lst[i][0])     #GPS heading offset
            elif i == 3: ctr_off = float(lst[i][0])    #GPS center offset
            elif i == 4: arm_n = int(lst[i][0])    #number of sensor arms
            elif i == 5: RS = int(lst[i][0])    #row sapcing [inch]
            elif i == 6: Heading_n = int(lst[i][0])
            elif i == 7: 
                GPS = int(lst[i][0])
                GPS_n = int(lst[i][1])   #number of GPS Lat\Lon per timestamp
            elif i == 8: 
                IMU = int(lst[i][0])
                IMU_n = int(lst[i][1])   #number of IMU per timestamp
            elif i == 9: 
                CC = int(lst[i][0])
                CC_n = int(lst[i][1])   #number of CC per row
            elif i == 10: 
                T= int(lst[i][0])
                T_n = int(lst[i][1])    #number of T per row
            elif i == 11: 
                Tb= int(lst[i][0])
                Tb_n = int(lst[i][1])    #number of Tb per row
            elif i == 12: 
                H = int(lst[i][0])
                H_n = int(lst[i][1])    #number of H per row
            else:        
                P = int(lst[i][0])
                P_n = int(lst[i][1])    #number of P per row
    return CC2BB,yaw_off,a,ctr_off,arm_n,RS,Heading_n,GPS,IMU,CC,T,Tb,H,P,GPS_n,IMU_n,CC_n,T_n,Tb_n,H_n,P_n

def Read_FD_GPS(filename=None):
    if(filename==None): filename = os.getcwd()+'\Config\ALARC_FD_GPS.csv'        
    #Detect # of lines
    num_lines = sum(1 for line in open(filename))
    print(num_lines,'filename=',filename)        
    
    lst = [0]*num_lines
    FD = [] #(Lon, Lat)
    # generate a dictionary for FD_GPS = {"FID":[(Lon1,Lat1),(Lon2,Lat2),(Lon3,Lat3),(Lon4,Lat4)],"":[(),(),(),()], ...}
    FD_GPS = {}
    with open(filename) as f:
        for i, line in enumerate(f):    
            if i>0 and i<num_lines:
                lst[i] = line.split(',')       
                #print (i,lst[i],lst[i][1])                    
                if i%4 == 1: 
                    FID = lst[i][0]
                    #print('FID=',FID)
                if float(lst[i][2])>180:  # convert UTM2LL if UTM 
                    Lat_deg, Lon_deg = Convert_UTM2LL(float(lst[i][1]),float(lst[i][2]),zone=12)
                    FD.append((Lon_deg,Lat_deg))
                else: FD.append((float(lst[i][2]),float(lst[i][1])))
                #print('FD=',FD)
                if i%4 == 0:
                    FD_GPS[FID] = FD
                    FD = []
    return FD_GPS

def Temperature2Color(Tmp):
    LB = 0
    UB = 255
    T1 = LB+(UB-LB)/4; 
    T2 = LB+2*(UB-LB)/4;
    T3 = LB+3*(UB-LB)/4;      
    #print(Tmp,T1,T2,T3)

    if Tmp>=LB and Tmp<=T1:  # Black~Blue
        R = 0
        G = 0
        B = 255-int((T1-Tmp)*255/(T1-LB))
    elif Tmp>T1 and Tmp<=T2: # Blue~Cyan
        R = 0
        G = 255-int((T2-Tmp)*255/(T2-T1))
        B = 255
    elif Tmp>T2 and Tmp<=T3:  # Cyan~Yellow
        R = 255-int((T3-Tmp)*255/(T3-T2))
        G = 255
        B = int((T3-Tmp)*255/(T3-T2))
    elif Tmp>T3 and Tmp<=UB:  # Yellow~Red
        R = 255
        G = int((UB-Tmp)*255/(UB-T3))
        B = 0
    elif Tmp>UB: R = G = B = 255
    else: R = G = B = 0	        
    return R,G,B    


def VI2Color(VI, cmap=None):
    LB = 0
    UB = 255
    T1 = LB+(UB-LB)/4; 
    T2 = LB+2*(UB-LB)/4;
    T3 = LB+3*(UB-LB)/4;      
    #print(VI,T1,T2,T3)
    
    if cmap == None or cmap == 'red':    
        if VI>=LB and VI<=T1:  # Black~Red Black~Blue
            R = 255-int((T1-VI)*255/(T1-LB))
            G = 0
            B = 0
        elif VI>T1 and VI<=T2: # Red~Yellow Blue~Cyan
            R = 255
            G = 255-int((T2-VI)*255/(T2-T1))
            B = 0
        elif VI>T2 and VI<=T3:  # Yellow~Green Cyan~Yellow
            R = int((T3-VI)*255/(T3-T2))
            G = 255
            B = 0
        elif VI>T3 and VI<=UB:  # Green~Dark Green Yellow~Red
            R = 0
            G = 127+int((UB-VI)*128/(UB-T3))
            B = 0
        elif VI == -1: # filtered non-numerics and percentiles to White
            R = G = B = 255
        else: R = G = B = 0	        

    else:
        if VI>=LB and VI<=T1:  # Black~Blue
            R = 0
            G = 0
            B = 255-int((T1-VI)*255/(T1-LB))
        elif VI>T1 and VI<=T2: # Blue~Cyan
            R = 0
            G = 255-int((T2-VI)*255/(T2-T1))
            B = 255
        elif VI>T2 and VI<=T3:  # Cyan~Yellow
            R = 255-int((T3-VI)*255/(T3-T2))
            G = 255
            B = int((T3-VI)*255/(T3-T2))
        elif VI>T3 and VI<=UB:  # Yellow~Red
            R = 255
            G = int((UB-VI)*255/(UB-T3))
            B = 0
        elif VI == -1: # filter non-numerics to White
            R = G = B = 255
        else: R = G = B = 0	        
        
    return R,G,B    

#Loading geotiff image
def LoadGeoTiff(file):
    (fileRoot, fileExt) = os.path.splitext(file)
    outFileName = fileRoot + "_mod" + fileExt
    #import scipy
    #im = scipy.misc.imread(file)
    ds = gdal.Open(file, gdal.GA_ReadOnly)    
    
    ##Getting Dataset info
    #print ('ds=',ds)
    if ds == None:
        messagebox.showwarning(title='Warning',message='ds is NoneType!')
    else:
        nband = ds.RasterCount
        #print ('Driver: {}/{}'.format(ds.GetDriver().ShortName,ds.GetDriver().LongName))
        #print('Size is {} x {} x {}'.format(ds.RasterXSize,ds.RasterYSize,nband))
        #print('Projection is {}'.format(ds.GetProjection()))
        #print('Metadata=',ds.GetMetadata())
        #print('len(GetProjection)=',len(ds.GetProjection()))
        #print('GetProjection=',ds.GetProjection()) 
        #print('UNIT=',osr.SpatialReference(wkt=ds.GetProjection()).GetAttrValue('UNIT',0))
        #print('EPSG=',osr.SpatialReference(wkt=ds.GetProjection()).GetAttrValue('AUTHORITY',1))
        #dType = gdal.GetDataTypeName(ds.GetRasterBand(1).DataType)    
        #print('Band Type = {}'.format(dType))   
    
    #Fetching a Raster Band
    #im = []
    #for i in range(ds.RasterCount):        
        #band = ds.GetRasterBand(i+1)  #Get the first band
        #(min,max) = band.ComputeRasterMinMax(True)  # estimate(faster)
        #print ('min~max=',min,max)
        #stats = band.GetStatistics(0,1) # actual(slower)
        #print ('stat=',stats) # [min, max, mean, stdev]       
        #Array = band.ReadAsArray()   #Open as an array, buffer (x2 firsrt, x0.3 next)
        #print ('min=',np.min(Array),'nanmin=',np.nanmin(Array),'max=',np.max(Array),'nanmax=',np.nanmax(Array),'mean=',np.mean(Array),'nanmean=',np.nanmean(Array))
        
        #NDV = band.GetNoDataValue() #Get the No Data Value
        #print ('i=',i,'NDV=',NDV        )
        #Array = Array.astype(np.float); # buffer02 (x3)
        #Array[Array == NDV] = np.nan  #Convert No Data Points(0s) to nans, otherwise wrong statistics due to 0s
        #print ('min=',np.min(Array),'nanmin=',np.nanmin(Array),'max=',np.max(Array),'nanmax=',np.nanmax(Array),'mean=',np.mean(Array),'nanmean=',np.nanmean(Array))
        
        #Array[Array<0] = 0  # filter neg float16 -> 0
        #im.append(Array)#b1            

    #bands = [ds.GetRasterBand(i+1) for i in range(0, ds.RasterCount)]
    #im = np.array([band.ReadAsArray() for band in bands])

    if nband>2: # Check and remove an alpha channel in 3b or 4b image        
        im = ds.ReadAsArray() # R,G,B,N=im[:,:,0],im[:,:,1],im[:,:,2],im[:,:,3], buffer (x5), loading time (4s for 166MB)    
        im = np.transpose(im, [1,2,0]) # reorder dimensions, so that channels are last
        
        #for i in range(nband): print(i,'count_nonzero255=',np.count_nonzero(im[:,:,i]==255))            
        if nband>3 and (np.count_nonzero(im[:,:,3]) == np.count_nonzero(im[:,:,3]==255)):   # all nonzeros are 255 if alpha band
            ds = gdal.Translate('/vsimem/dummy.tif' , ds, format='GTiff', bandList=[1,2,3])    
            im = ds.ReadAsArray() # R,G,B,N=im[:,:,0],im[:,:,1],im[:,:,2],im[:,:,3], buffer (x5), loading time (4s for 166MB)    
            nband = ds.RasterCount                
            im = np.transpose(im, [1,2,0]) # reorder dimensions, so that channels are last
    elif nband == 2:    # Check and remove an alpha channel in 2b Matrice images
        ds = gdal.Translate('/vsimem/dummy.tif' , ds, format='GTiff', bandList=[1])    
        im = ds.ReadAsArray() #Get the first band
        nband = ds.RasterCount                
    else:
        im = ds.ReadAsArray() # R,G,B,N=im[:,:,0],im[:,:,1],im[:,:,2],im[:,:,3], buffer (x5), loading time (4s for 166MB)    
    
    #im = cv2.cvtColor(im,cv2.COLOR_BGRA2RGBA); # colorswap RGB2BGR for cv2.imshow        
    if len(im.shape)==2: im = im.reshape(im.shape[0],im.shape[1],1) # 1b img shape:(H,W)->(H,W,1) for im[:,:,i], ->(H,W) for ImageTk.PhotoImage(), Save_GeoTiff()
    print('im.shape=',im.shape) 
    
    #decompose 4b image bands to correctly assign Red band
    if len(im.shape)>2 and im.shape[2]==4:
        fbasename = file.split('/')[-1].split('.')[0]
        if fbasename.lower().find('bgrn')>=0 or fbasename.lower().find('rgbn')<0: # bgrn or non-rgbn in filename
            im = im[:,:,[2,1,0,3]]   # Red@b2 in PS2 image BGRN -> RGBN 
            print('Image changed: BGRN -> RGBN') 

    #decompose 8b image bands to correctly assign Red band
    if len(im.shape)>2 and im.shape[2]==8:
        fbasename = file.split('/')[-1].split('.')[0]
        if fbasename.lower().find('bgrn')>=0 or fbasename.lower().find('rgbn')<0: # bgrn or non-rgbn in filename
            im = im[:,:,[5,3,1,7]]   # Red@b2 in PS2 image bBgGYRrN -> RGBN 
            #print('Image changed: bBgGYRrN -> RGBN') 
    
    #cv2.imshow('im0',im[:,:,0]), cv2.imshow('im1',im[:,:,1]),cv2.imshow('im2',im[:,:,2]),cv2.waitKey()
    #cv2.imshow('im',cv2.cvtColor(im,cv2.COLOR_RGB2BGR)), cv2.waitKey(0)
        
    #if nband==1:
        #im.append([])   #b2 = empty
        #im.append([])   #b3 = empty 
    #if nband==3 or (nband==4 and dType=='Byte'): #if 8-bit RGB or RGBA image, NIR band = empty
        #im.append([])   #b4 = empty
        
    return im,ds

#GPS convertion from UTM to Lat\Lon
def Convert_UTM2LL(E,N,zone=None,pj=None):  
    #print(E,N,zone)    
    UTM = osr.SpatialReference()    
    if zone != None:
        UTM.SetWellKnownGeogCS("WGS84")   #set coord sys to handle lat/lon  
        if N > 0: Is_Northern = 1
        else:     Is_Northern = 0  
        UTM.SetUTM(zone, Is_Northern)        
    if pj != None:
        UTM.ImportFromWkt(pj)
    LL = UTM.CloneGeogCS() #clone only the coord sys      
    UTM2LL = osr.CoordinateTransformation(UTM, LL)  #create transform component        
    #Lon_deg,Lat_deg,Altitude = UTM2LL.TransformPoint(E, N)  # GDAL 2.0   
    Lat_deg,Lon_deg,Altitude = UTM2LL.TransformPoint(E, N)  # GDAL 3.0   
    return Lat_deg, Lon_deg

# Convert runtime in second to hh:mm:ss
def ConvertSec2hhmmss(sec):    
    hh = int(sec/(3600))
    mm = int((sec - hh*3600)/60)
    ss = sec - hh*3600 - mm*60
    return hh,mm,ss

# stretch the image (e.g. uint16) to displayable image (uint8)
def Stretch_Image(im0,stretch):
    if len(im0.shape)>2: nband = im0.shape[2]
    else:               nband = 1
    # convert unit16 to unit8 w percentile stretch
    #print('stretch=',stretch[0],stretch[1],stretch[2],stretch[3],stretch[4])
    option = stretch[0] # Stretch options: 0 for Percentile, 1 for min/max
    L_bound = stretch[1] # 2.0: Lower boundary for percentile stretch
    U_bound = stretch[2] # 98.0: Upper boundary for percentile stretch         
    minGL = stretch[3] # 0: min GL
    maxGL = stretch[4] # 255: max GL 
    im = np.copy(im0)   # create a new memory    
    del im0
    
    # apply stretch option: 0 for Percentile, 1 for min/max
    if option == 1: 
        im[im<minGL] = np.nan
        L_bound, U_bound = 0,100        
    if option >= 0:    
        # convert unit16 to unit8 w min/max stretch
        if nband==1:
            print('np.min(im)=',np.min(im),'np.max(im)=',np.max(im))
            if np.min(im)<0: im[im<0] = np.nan   # filter negative px to nan if float32 NDVI image
            minP,maxP=np.nanpercentile(im[:,:],(L_bound,U_bound)); 
            im_stretched = np.interp(im[:,:],(minP,maxP),(0,1))*255        
        else:
            for i in range(nband):
                minP,maxP=np.nanpercentile(im[:,:,i],(L_bound,U_bound)); 
                if i==0: im_R = np.interp(im[:,:,i],(minP,maxP),(0,1))*255
                if i==1: im_G = np.interp(im[:,:,i],(minP,maxP),(0,1))*255
                if i==2: im_B = np.interp(im[:,:,i],(minP,maxP),(0,1))*255
                if i==3: im_N = np.interp(im[:,:,i],(minP,maxP),(0,1))*255
                #print('im_R=',im_R.shape,minP,maxP)
            #cv2.imshow('imR',im_R), cv2.imshow('imG',im_G),cv2.imshow('imB',im_B),cv2.imshow('imN',im_N),cv2.waitKey()
            if nband==3:
                im_stretched = np.stack((im_R,im_G,im_B), axis=2)  #8-bit RGB img
            elif nband==4 and len(im[:,:,3])!=0:    # 16-bit img, buffer (x1)
                im_stretched = np.stack((im_R,im_G,im_B,im_N), axis=2)  #float64, 16-bit BGRN img (0~65,536)->(1~255)
            else: 
                messagebox.showerror(title='Error',message='nband=%d is invalid!'%nband)
    del im
    im_stretched = np.uint8(im_stretched)    # float64->uint8
    return im_stretched

#GPS convertion from Lat\Lon to UTM
def Convert_LL2UTM(LL, CRS=None):  
    #print('LL(%d)='%(len(LL)),LL)
    #Convert Lat\Lon to Lat_deg\Lon_deg        
    if len(LL)>2:   #format of 123,N,456,W            
        Lat = float(LL[0]) 
        Lon = float(LL[2])
        #print (abs(Lat),abs(Lon))
        if abs(Lat)>180:    #if in minutes, min => deg
            Lat_deg =  (Lat//100)+(Lat/100-(Lat//100))*100/60
            Lon_deg =  (Lon//100)+(Lon/100-(Lon//100))*100/60                                           
        if LL[1]=="S": Lat_deg = -Lat_deg
        if LL[3]=="W": Lon_deg = -Lon_deg                                    
    else:
        Lat_deg = float(LL[0]) 
        Lon_deg = float(LL[1])                        

    # convert LL to UTM    
    zone = 1+int((Lon_deg+180)/6)  #UTM Zone, 12=AZ
    if CRS==None:   # default CRS = WGS84
        # Datum WGS84 parameters 
        a = 6378137 #equatorial radius
        f = 1/298.257223563 #flattening
        k0 = 0.9996 #scale factor
        ee = 0.00669438 #eccentricity
        EE = 0.006739497
        LongOrgRad = radians((zone-1)*6-180+3)
        Lat_rad = Lat_deg*pi/180
        Lon_rad = Lon_deg*pi/180
        N = a/sqrt(1-ee*sin(Lat_rad)**2)
        T = tan(Lat_rad)**2
        C = EE*cos(Lat_rad)**2
        A = cos(Lat_rad)*(Lon_rad-LongOrgRad)
        M = a*((1-ee/4-3*ee**2/64-5*ee**3/256)*Lat_rad-(3*ee/8+3*ee**2/32+45*ee**3/1024)*sin(2*Lat_rad)+(15*ee**2/256+45*ee**3/1024)*sin(4*Lat_rad)-(35*ee**3/3072)*sin(6*Lat_rad))    
        #print (LongOrgRad,Lat_rad,Lon_rad,N,T,C,A,M)
        Northing = k0*(M+N*tan(Lat_rad)*(A**2/2+(5-T+9*C+4*C**2)*A**4/24+(61-58*T+T**2+600*C-330*EE)*A**6/720))
        Easting = k0*N*(A+(1-T+C)*A**3/6+(5-18*T+T**2+72*C-58*EE)*A**5/120)+500000  
    else:   #convert LL2UTM using gdal
        if Lat_deg > 0: Is_Northern = 1
        else:           Is_Northern = 0  
        UTM = osr.SpatialReference()    
        UTM.SetWellKnownGeogCS(CRS)   #set coord sys to handle lat/lon  
        UTM.SetUTM(zone, Is_Northern)        
        LL = UTM.CloneGeogCS() #clone only the coord sys      
        LL2UTM = osr.CoordinateTransformation(LL, UTM)  #create transform component    
        Easting,Northing,Altitude = LL2UTM.TransformPoint(Lon_deg, Lat_deg, 0)    
 
    return Lon_deg,Lat_deg,Easting,Northing

#Calcuate VIs
def CalculateVIs(R):  
    #print(R)     #R0(590), R1(730), R2(530), R3(550), R4(800), R5(670)
    #if R[0]!=0 and R[1]!=0 and R[2]!=0 and R[3]!=0 and R[4]!=0 and R[5]!=0: 
    if (R[4] + R[0]) != 0: NDVIA800 =(R[4] -R[0]) / (R[4] + R[0])
    else: NDVIA800 = ""
    if (R[1] + R[0]) != 0: NDARE =(R[1] -R[0]) / (R[1] + R[0])
    else: NDARE = ""
    if (R[4] + R[5]) != 0: NDVIR800 =(R[4] -R[5]) / (R[4] + R[5])
    else: NDVIR800 = ""
    if (R[1] + R[5]) != 0: NDREE =(R[1] -R[5]) / (R[1] + R[5])
    else: NDREE = ""
    if (R[4] + R[1]) != 0: NDRE =(R[4] -R[1]) / (R[4] + R[1])
    else: NDRE = ""
    if (R[4] + R[2]) != 0: NDVIG800 =(R[4] -R[2]) / (R[4] + R[2])
    else: NDVIG800 = ""
    PIR =R[0] -R[2]
    CI800 =(R[4] / R[0]) -1
    if NDVIR800!=0: CCCI= (NDRE / NDVIR800)
    else: CCCI = ""
    if R[4] -R[5] != 0: DATT= (R[4] -R[1]) / (R[4] -R[5])
    else: DATT = ""
    if (R[4] -R[0]) != 0: DATTA= (R[4] -R[1]) / (R[4] -R[0])
    else: DATTA = ""
    if NDVIA800!=0: CCCIA= NDRE / NDVIA800        
    else: CCCIA = ""
    #else: 
        #NDVIA800=NDARE=NDVIR800=NDREE=NDRE=NDVIG800=PIR=CI800=CCCI=DATT=DATTA=CCCIA=""
    return NDVIA800,NDARE,NDVIR800,NDREE,NDRE,NDVIG800,PIR,CI800,CCCI,DATT,DATTA,CCCIA

#Update ROIs from 2-\4-pt to 4-pt tROI when redefined ROI
def Update_ROIs(grid,ROI,rot=None):      
    mode = ROI[0][-1]    
    
    #Draw a grid if RxC > 1x1
    if mode != 'n':
        if rot == 0:            
            #print('ROI=',ROI,'grid=',grid)
            ROIs = Update_ROI(grid,ROI)
            #print('ROIs=',ROIs)
        else:
            rotGRID,a,b,r,min_theta, max_theta = Calculate_GRID_corners(ROI,rot) 
            i1,j1,i2,j2,i3,j3,i4,j4 = rotGRID
        
        # Calculate 4 corners of each subROI
        R,C = grid
        tROI = [[0]*9 for i in range(R*C)]    # 4 corners: NW,SE,NE,SW        
        #print('R,C=',R,C,'i1,j1,i2,j2,i3,j3,i4,j4=',i1,j1,i2,j2,i3,j3,i4,j4)        
        for j in range(R):        
            for i in range(C):        
                # define 4 corners of horizonal two lines of a grid 
                if i==0 and rot!=0:
                    ii1,jj1,ii3,jj3 = (i4-i1)*j/R+i1, (j4-j1)*j/R+j1, (i2-i3)*j/R+i3, (j2-j3)*j/R+j3
                    ii4,jj4,ii2,jj2 = (i4-i1)*(j+1)/R+i1, (j4-j1)*(j+1)/R+j1, (i2-i3)*(j+1)/R+i3, (j2-j3)*(j+1)/R+j3
                    i_dt = (ii3 - ii1)/C    # delta i of subROI in rotation
                    j_dt = (jj3 - jj1)/C    # delta j of subROI in rotation
                    #print('ii1,jj1,ii2,jj2,ii3,jj3,ii4,jj4=',ii1,jj1,ii2,jj2,ii3,jj3,ii4,jj4,'i_dt,j_dt=',i_dt,j_dt)        

                # assign 4 corners to tROI
                if rot == 0:
                    tROI[i+j*C][0:4] = ROIs[i+j*C][0:4]
                    tROI[i+j*C][4:8] = ROIs[i+j*C][2], ROIs[i+j*C][1], ROIs[i+j*C][0], ROIs[i+j*C][3]  # NE(i3,j3) SW(i4,j4)
                else:
                    tROI[i+j*C][0:4] = round(ii1+i_dt*i), round(jj1+j_dt*i), round(ii4+i_dt*(i+1)), round(jj4+j_dt*(i+1))   # NW(i1,j1) SE(i2,j2)
                    tROI[i+j*C][4:8] = round(ii1+i_dt*(i+1)), round(jj1+j_dt*(i+1)), round(ii4+i_dt*i), round(jj4+j_dt*i)  # NE(i3,j3) SW(i4,j4)
                tROI[i+j*C][8] = ROI[0][-1] # ROI shape: r (rectangle), c (circle)
    return tROI

#Update ROI to ROIs (2-corner) when redefined ROI
def Update_ROI(grid,ROI,imWH=None):      
    #if len(ROI)<grid[0]*grid[1]:
        #print('0Update_ROI: grid(RxC)=%d x %d=%d'%(grid[0],grid[1],grid[0]*grid[1]),'len(ROI)=',len(ROI),'len(ROI[0])=',len(ROI[0]))    
    #else: 
        #print('\n0Update_ROI: grid(RxC)=%d x %d=%d'%(grid[0],grid[1],grid[0]*grid[1]),'len(ROI)=',len(ROI),'len(ROI[0])=',len(ROI[0]))    

    # filter ROI if beyond image boundary
    if imWH!=None: 
        imW,imH = imWH
        n = len(ROI)-1
        if ROI[n][2]>=imW or ROI[n][3]>=imH:
            print('imW,imH=',imW,imH,',ROI[n]=',ROI[n])
            if ROI[0][0]>=imW: ROI[0][0] = 0 
            if ROI[0][1]>=imH: ROI[0][1] = 0    
            if ROI[n][2]>=imW: ROI[n][2] = imW-1   #640x480: col=0~640, W=640-0
            if ROI[n][3]>=imH: ROI[n][3] = imH-1        

    (X1,Y1),(X2,Y2),mode = ROI[0][:2],ROI[len(ROI)-1][2:4],ROI[0][-1]

    # reposition to (X1,Y1)<(X2,Y2) if reversed
    #print('ROI=',ROI,end='->')  
    if X1>X2: X1,X2 = X2,X1
    if Y1>Y2: Y1,Y2 = Y2,Y1
    #print(ROI)  
    
    # create ROIs from the outer ROI and GRID
    if mode==None: 
        mode = 'r' # bring back to 'r' after Read_Config
    ROIs = [[0]*5 for i in range(grid[0]*grid[1])]  # 2-corner ROIs
    for j in range(grid[0]):    #row
        for i in range(grid[1]):    #col            
            x2 = X1+round((i+1)*(X2-X1)/grid[1])
            y2 = Y1+round((j+1)*(Y2-Y1)/grid[0])
            if i == 0:
                if j==0:
                    ROIs[0] = [X1, Y1,                                                x2, y2, mode]
                else:
                    ROIs[j*grid[1]] =[ROIs[(j-1)*grid[1]][0], ROIs[(j-1)*grid[1]][3], x2, y2, mode]
            else:
                ROIs[j*grid[1]+i] = [ROIs[j*grid[1]+i-1][2], ROIs[j*grid[1]+i-1][1], x2, y2, mode]
    ##print('ROIs=',ROIs)    
    #print('1Update_ROI: grid(RxC)=%d x %d=%d'%(grid[0],grid[1],grid[0]*grid[1]),'len(ROIs)=',len(ROIs),'len(ROIs[0])=',len(ROIs[0]))    
    return ROIs

# Calculate 4-side boundary of GRID from tROI
def Calculate_GRID_boundary(ROIpx,imW,imH):
    # calculate 4-side boundary of tROI if rotated ROI
    tX1,tY1,tX2,tY2 = imW,imH,0,0
    minx,miny,maxx,maxy = [imW]*len(ROIpx), [imH]*len(ROIpx), [0]*len(ROIpx), [0]*len(ROIpx)
    #print('ROIpx[0]=',ROIpx[0])
    #print('ROIpx[0][0]=',ROIpx[0][0])
    #print('ROIpx[0][0][1]=',ROIpx[0][0][1])
    yy, xx = np.ogrid[:imH, :imW]  # a range of x-\y-coord, xx=[[0],...,[n]], yy=[[0,...,n]] 
    #print('len(xx[0]),len(yy)=',len(xx[0]),len(yy))   
    mask_tROI = xx+yy<0 # create a im-sized matrix with all False, [j][i]            
    for n in range(len(ROIpx)):
        for p in range(len(ROIpx[n])):
            minx[n],miny[n], maxx[n],maxy[n] = min(ROIpx[n][p][0],minx[n]), min(ROIpx[n][p][1],miny[n]), max(ROIpx[n][p][0],maxx[n]), max(ROIpx[n][p][1],maxy[n])
            mask_tROI[ROIpx[n][p][1]][ROIpx[n][p][0]] = True  # [j][i]
        tX1,tY1, tX2,tY2 = min(minx[n],tX1),min(miny[n],tY1),max(maxx[n],tX2),max(maxy[n],tY2)
        #print('minx[n],miny[n],maxx[n],maxy[n]=',minx[n],miny[n], maxx[n],maxy[n])   
    OuterBoundary = [tX1,tY1, tX2,tY2]
    InnerBoundary = [minx,miny,maxx,maxy]
    return mask_tROI, OuterBoundary, InnerBoundary

# Calculate 4-corner GRID from 2-corner GRID
def Calculate_GRID_corners(ROI,rot):
    # calculate max theta             
    i1 = ROI[0][0]
    j1 = ROI[0][1]    
    i2 = ROI[len(ROI)-1][2]
    j2 = ROI[len(ROI)-1][3]  
    if (i2 - i1) != 0:
        min_theta = degrees(atan((j1 - j2)/(i2 - i1)))  # NOTE: change y-values to negative in xy coord 
    else: min_theta = 0
    max_theta = 90 + min_theta

    # Calculate the other diagonal corners: NE(i3,j3) SW(i4,j4)
    if rot != None : # calculate 
        j1, j2 = -j1, -j2 # NOTE: change y-values to negative in xy coord 
        
        # calculate parameters to find a rotated pt (i3,j3)
        r =sqrt((i2-i1)**2 + (j2-j1)**2)/2  # radius
        a = i1 + (i2-i1)/2  # center x
        b = j1 + (j2-j1)/2  # center y
        s = tan(radians(rot))   # slope
        #print('r,a,b,s=',r,a,b,s)            
        
        # calculated parameters for quadratic formula
        D = j1 - s*i1 - b
        A = s**2 + 1
        B = 2*(D*s - a)
        C = D**2 + a**2 - r**2
        #print('D,A,B,C=',D,A,B,C)
        i3 = (-B + sqrt(B**2 - 4*A*C)) / (2*A)  # either of +/- frin quadratic formula
        j3 = sqrt(r**2 - (i3-a)**2) + b  # either of +/- from circular formula
        
        # calculate theta when crossing (xmax,b)
        xmax = a + r
        L = xmax - i1
        H = b - j1
        if L != 0: theta = degrees(atan(H/L))
        else: theta = 0
        #print('L=',L,'H=',H,'theta=',theta)
        # reverse the sign of sqrt if rotation angle is crossing (xmax,b)
        if rot < theta:                  
            j3 = -sqrt(r**2 - (i3-a)**2) + b  # either of +/- from circular formula
        
        i4 = a + (a - i3)
        j4 = b + (b - j3)

        j1, j2, j3, j4, b = -j1, -j2, -j3, -j4, -b # NOTE: change y-values back to positive in ij coord
        #print('i3,j3,i4,j4=',i3,j3,i4,j4)
    else:
        i3, j3, i4, j4 = i2, j1, i1, j2
    rotGRID = [i1,j1,i2,j2,i3,j3,i4,j4]
    return rotGRID, a,b,r, min_theta, max_theta

#Draw pixels in subROI
def Draw_ROIpx(ROIpx,canvas):    
    #print('len(tROI)=',len(tROI))
    for i in range (len(ROIpx)):    # 3D array: [n-th ROI of ROIs], [n-th px in ROI], [i or j value]        
        if   i==0: color = 'red'
        elif i==1: color = 'green'
        elif i==2: color = 'blue'
        else:      color = 'yellow'
        
        # draw px in 2nd subROI for verification
        if i==1:
            dt, zf, n = 1, 1, len(ROIpx[i])         
            #if n > 1000: 
                #zf = 100
                #n = int(n/zf)
            for j in range(n): 
                j = j*zf
                #print('i,j=',i,j,'len(ROIpx[i])',len(ROIpx[i]))
                if j>=len(ROIpx[i]): j = len(ROIpx[i])-1
                x, y = ROIpx[i][j][0], ROIpx[i][j][1]
                x1, y1, x2, y2 = x-dt, y-dt, x+dt, y+dt
                #print('x,y=',x,y,'x1,y1,x2,y2=',x1,y1,x2,y2)
                canvas.create_oval(x1,y1,x2,y2,outline=color)
    #print('len(ROIpx)=',len(ROIpx),ROIpx)
    #if len(ROIpx)>1:
        #print('ROIpx[1]=',ROIpx[1])
        #print('ROIpx[1][2]=',ROIpx[1][2])

#Draw ROI
def Draw_ROI(grid,ROIs,zf,canvas,rot=None):    
    (X1,Y1),(X2,Y2),mode = ROIs[0][:2],ROIs[len(ROIs)-1][2:4],ROIs[0][-1]
    #print('X1,Y1,X2,Y2,mode=',X1,Y1,X2,Y2,mode)
    
    #Draw the outer ROI
    #if mode == 'c':
        #canvas.create_oval(X1*zf,Y1*zf,X2*zf,Y2*zf, outline='red')    #oval ROI
    #elif mode == 'r' or mode == 'p' :
        #canvas.create_rectangle(X1*zf,Y1*zf,X2*zf,Y2*zf, outline='blue')   #rectangular ROI        

    #Draw a grid if RxC > 1x1
    if mode != 'n':
        if rot == 0:            
            ROIs = Update_ROI(grid,ROIs)    # update ROI to ROIs (2-corner) when redefined ROI
        rotGRID,a,b,r,min_theta, max_theta = Calculate_GRID_corners(ROIs,rot) 
        i1,j1,i2,j2,i3,j3,i4,j4 = rotGRID        
    
        # clear the previous ROI drawsing
        canvas.delete('all')
        canvas.create_image(0,0,anchor="nw",image=img_photo)                     
        
        if mode == 'c':
            canvas.create_oval(X1*zf,Y1*zf,X2*zf,Y2*zf, outline='blue')    #oval ROI        
        
        # Calculate 4 corners of each subROI
        R,C = grid
        tROI = [[0]*9 for i in range(R*C)]    # 4 corners: NW,SE,NE,SW
        #print('R,C=',R,C,'i1,j1,i2,j2,i3,j3,i4,j4=',i1,j1,i2,j2,i3,j3,i4,j4)        
        for j in range(R):        
            for i in range(C):        
                # define 4 corners of horizonal two lines of a grid 
                if i==0:
                    hi1,hj1,hi3,hj3 = (i4-i1)*j/R+i1, (j4-j1)*j/R+j1, (i2-i3)*j/R+i3, (j2-j3)*j/R+j3
                    hi4,hj4,hi2,hj2 = (i4-i1)*(j+1)/R+i1, (j4-j1)*(j+1)/R+j1, (i2-i3)*(j+1)/R+i3, (j2-j3)*(j+1)/R+j3
                    i_dt = (hi3 - hi1)/C    # delta i of subROI in rotation
                    j_dt = (hj3 - hj1)/C    # delta j of subROI in rotation
                    #print('hi1,hj1,hi2,hj2,hi3,hj3,hi4,hj4=',hi1,hj1,hi2,hj2,hi3,hj3,hi4,hj4,'i_dt,j_dt=',i_dt,j_dt)        

                    # draw the h-lines 
                    if mode != 'c':
                        canvas.create_line(hi1*zf,hj1*zf,hi3*zf,hj3*zf, fill='blue')   #h-grid in ROI 
                        canvas.create_line(hi4*zf,hj4*zf,hi2*zf,hj2*zf, fill='blue')   #h-grid in ROI                 
                
                # assign 4 corners to tROI
                if rot == 0:
                    tROI[i+j*C][0:4] = ROIs[i+j*C][0:4]
                    tROI[i+j*C][4:8] = ROIs[i+j*C][2], ROIs[i+j*C][1], ROIs[i+j*C][0], ROIs[i+j*C][3]  # NE(i3,j3) SW(i4,j4)
                else:
                    tROI[i+j*C][0:4] = round(hi1+i_dt*i), round(hj1+j_dt*i), round(hi4+i_dt*(i+1)), round(hj4+j_dt*(i+1))   # NW(i1,j1) SE(i2,j2)
                    tROI[i+j*C][4:8] = round(hi1+i_dt*(i+1)), round(hj1+j_dt*(i+1)), round(hi4+i_dt*i), round(hj4+j_dt*i)  # NE(i3,j3) SW(i4,j4)
                tROI[i+j*C][8] = ROIs[i+j][-1] # ROI shape: r (rectangle), c (circle)

                #draw the v-lines 
                if j==0 and mode != 'c':
                    vi1,vj1,vi4,vj4 = round((i3-i1)*i/C+i1), round(j1-(j1-j3)*i/C), round((i2-i4)*i/C+i4), round(j4-(j4-j2)*i/C)                
                    vi3,vj3,vi2,vj2 = round((i3-i1)*(i+1)/C+i1), round(j1-(j1-j3)*(i+1)/C), round((i2-i4)*(i+1)/C+i4), round(j4-(j4-j2)*(i+1)/C)                
                    canvas.create_line(vi1*zf,vj1*zf,vi4*zf,vj4*zf, fill='blue')
                    canvas.create_line(vi3*zf,vj3*zf,vi2*zf,vj2*zf, fill='blue')
                
                ## draw each subROI (for verification, not for use due to slow drawing)
                #x1,y1,x2,y2,x3,y3,x4,y4 = tROI[i+j*C][:8]                
                #canvas.create_line(x1*zf,y1*zf,x3*zf,y3*zf, fill='blue')   #h-grid in ROI 
                #canvas.create_line(x4*zf,y4*zf,x2*zf,y2*zf, fill='blue')   #h-grid in ROI 
                #canvas.create_line(x1*zf,y1*zf,x4*zf,y4*zf, fill='blue')   #v-grid in ROI 
                #canvas.create_line(x3*zf,y3*zf,x2*zf,y2*zf, fill='blue')   #v-grid in ROI                     
                
    #print('rotGRID=',rotGRID)
    #print('ROIs=',ROIs)
    #print('tROI=',tROI)
    return tROI

#Get a mask image from different sources (1b, 2b, udm)
def Get_Mask(mask_file):      
    #Check if a mask file is loaded
    if len(mask_file)==0: 
        messagebox.showwarning(title='Warning',message='Load a mask image first!')
        return
    im,ds = LoadGeoTiff(mask_file)
    print('ds.RasterXSize=',ds.RasterXSize,'ds.RasterYSize=',ds.RasterYSize,'ds.RasterCount=',ds.RasterCount)   
    if ds.RasterCount==2: #detect 2b Cloud&Shadow mask image:b1:0(fg),1(outBorder&cloud),b2:0(fg),1(outBorder&shadow)
        im_R = np.zeros((ds.RasterYSize,ds.RasterXSize),dtype=np.uint8) #(height,width)
        RGB = np.stack((im_R, im[1], im[0]), axis=2)  #8-bit RGB img, 16-bit BGRN img                       
        img = Image.fromarray(np.uint8(RGB),'RGB')                
        mask = np.array(img.getchannel(1))+np.array(img.getchannel(2))  #combine Shadow(b2) & Cloud(b1) mask images
        mask[mask==0] = 255 #0(fgnd)->255
        mask[mask<255] = 0  #b1:1(cloud)->0,b2:1(shadow)->0,b1+b2:2(clod&shadow)->0
    else:
        mask = cv2.imread(mask_file,cv2.IMREAD_UNCHANGED)
        if mask_file.find('border') != -1:    #detect 8-bit border file:0(outBorder),1(fg)
            mask[mask==1] = 255 #1(fgnd)->255 
        if mask_file.find('udm') != -1:    #detect 8-bit udm file:0(fg),1(outBorder),2(cloud),4(B),8(G),16(R),32(RE),64(N)
            mask[mask==0] = 255 #0(fgnd)->255 
            mask[mask<255] = 0  #1(nan)->0, 2(cloud)->0, 64(N missed)->0                 
    return mask

#Get a masked image
def Get_Masked(im,mask):      
    #print(im.shape,mask.shape)
    #if len(im.shape)==3 and im.shape[2]==3: im = cv2.cvtColor(im,cv2.COLOR_RGB2BGR)   #convert matrix(RGB)=>array(BRG)    
    #print('im=',im.shape,'mask=',mask.shape)
    im_masked = cv2.bitwise_and(im,im,mask=mask)
    #print(im.shape,mask.shape,im_masked.shape)
    #if len(im.shape)==3 and im.shape[2]==3: im_masked = cv2.cvtColor(im_masked,cv2.COLOR_BGR2RGB)   #convert array(BRG)=>matrix(RGB)
    ##convert blacks in background pixels to magents    
    if kb.is_pressed('shift'):
        # Method 1: fgnd=[0,0,0]=>[255,0,255]
        background = np.full(im.shape, [255,0,255], dtype=np.uint8) #[255,0,255]=magents
        im_bg = cv2.bitwise_and(background, background, mask=~mask) #get bgnd image #~mask(inverted mask)=cv2.bitwise_not(mask)    
        print('im_masked=',im_masked.shape,'im_bg=',im_bg)
        im_masked = cv2.bitwise_or(im_masked, im_bg)    #combine fg+bk    
    
        # Method 2:this also brings fgnd=[0,0,0]=>[255,0,255]
        #im_masked[np.where((im_masked==[0,0,0]).all(axis=2))] = [255,0,255] 
        
        # Method 3:
        #h,w,bpp = im_masked.shape
        #for py in range(0,h):
            #for px in range(0,w):
                ##print('im_masked[%d][%d]='%(py,px),im_masked[py][px])
                #if im_masked[py][px][0]==0 & im_masked[py][px][1]==0 & im_masked[py][px][2]==0:
                    #im_masked[py][px]=[255, 0, 255] #magenta
    return im_masked

# A function to create the output geotif image
def createOutputImage(in_ds, outFilename, out_bands=None, dtype=None):
    # Define the image driver which is the output file format (e.g., GeoTiff)
    driver = gdal.GetDriverByName( "GTiff" )
    # Check that this driver can create a new file.
    metadata = driver.GetMetadata()
    if metadata[gdal.DCAP_CREATE] != 'YES':
        print ('Driver GTIFF does not support Create()')
        sys.exit(-1)    

    print('in_bands=',in_ds.RasterCount,'out_bands=',out_bands)
    print('dtype=',dtype,'outFilename=',outFilename)   
    print('out_bands=',in_ds.RasterCount,'dtype=',in_ds.GetRasterBand(1).DataType)

    if dtype == None:
        # Create an img using CreateCopy() for exact copy
        out_ds = driver.CreateCopy( outFilename, in_ds, 0 )        
        return out_ds
    
    # Create an img using Create() for new geo trans\proj
    out_ds = driver.Create(outFilename, in_ds.RasterXSize, in_ds.RasterYSize, out_bands, dtype)
    # Set the spatial information to the output image.
    #out_ds.SetGeoTransform([444720, 30, 0, 3751320, 0, -30 ])   #[x-coord of TL px, px width, proj of y on x-axis, y-coord of TL px, proj of x on y-axis, px height]
    #proj = osr.SpatialReference()     # establish encoding
    #proj.ImportFromEPSG(3857)         # WGS84 lat/long 
    #proj.SetUTM( 11, 1 )
    #proj.SetWellKnownGeogCS( 'NAD27' )
    #proj.ImportFromWkt(in_ds.GetProjectionRef())
    #out_ds.SetProjection(proj.ExportToWkt())   # export coords to file
    out_ds.SetGeoTransform(in_ds.GetGeoTransform())
    out_ds.SetProjection(in_ds.GetProjection())
    return out_ds

def Generate_NDVI(imN,imR):
    ## Normalize PS2(BGRN) to TOA
    #from xml.dom import minidom
    #xmldoc = minidom.parse("20161218_101700_0e0d_3B_AnalyticMS_metadata.xml")
    #nodes = xmldoc.getElementsByTagName("ps:bandSpecificMetadata")    
    #coeffs = {}
    #for node in nodes: # XML parser refers to bands by numbers 1-4
        #bn = node.getElementsByTagName("ps:bandNumber")[0].firstChild.data
        #if bn in ['1', '2', '3', '4']:
            #i = int(bn)
            #value = node.getElementsByTagName("ps:reflectanceCoefficient")[0].firstChild.data
            #coeffs[i] = float(value) # Find TOA Ref coeffs
    #imR = imR * coeffs[3] # Multiply the band by TOA Ref coeff
    #imN = imN * coeffs[4] # Multiply the band by TOA Ref coeff
    
    # convert dtype -> float32
    imN = imN.astype(np.float32)
    imR = imR.astype(np.float32)
    #np.seterr(divide='ignore', invalid='ignore')    # Allow division by zero 
    
    # Calculate NDVI
    #print('imNDVI.shape=',imNDVI.shape,'imNDVI.dtype=',imNDVI.dtype)   
    #imNDVI = np.where((imR+imN)==0.,0,(imN-imR)/(imN+imR)) # filter 0s in numerator to NDVI=0
    imNDVI = np.where(imR==np.nan,np.nan,(imN-imR)/(imN+imR)) # filter nan in image to NDVI=nan
    #imNDVI[imNDVI<0] = np.nan  # filter negative NDVI pixels to nan (not calculated in mean, but changing color in display)
    #imNDVI = (imNDVI*255).astype(np.uint8)
    #print(np.nanmin(imNDVI),np.nanmax(imNDVI),np.nanmean(imNDVI))
    #cv2.imshow('N',imN),cv2.imshow('R',imR),cv2.imshow('imNDVI',imNDVI),cv2.waitKey(0)
    
    return imNDVI

def Calculate_NDVI(imN,imR):
    imNDVI = Generate_NDVI(imN,imR)
    ndvi,ndvi_min,ndvi_max = np.nanmean(imNDVI),np.nanmin(imNDVI),np.nanmax(imNDVI)
    del imNDVI
    #print(ndvi_min,ndvi_max,ndvi_mean)
    return ndvi,ndvi_min,ndvi_max

def Calculate_NDVI_old(in_ds):  #px-level calcuation, slower than Calculate_NDVI()
    # Create the output dataset
    outFilename = 'C:\\Users\\james.kim2\\Documents\\Research\\Python\\IMAP\\Data\\ndvi.tif'        
    out_bands = 1; dtype=gdal.GDT_Float32   # 1:uint8, 6:gdal.GDT_Float32
    out_ds = createOutputImage(in_ds, outFilename, out_bands, dtype)
    # Check the dataset was successfully created.
    if out_ds is None:
        print ('Could not create output image')
        sys.exit(-1)
        
    # Loop through each line in turn.
    for line in range(in_ds.RasterYSize):
        # Define variable for output line.
        outputLine = ''.encode()
            
        # Read in data for the current line from the image band representing the red wavelength
        red_scanline = in_ds.GetRasterBand(3).ReadRaster( 0, line, in_ds.RasterXSize, 1, in_ds.RasterXSize, 1, gdal.GDT_Float32 )
        # Unpack the line of data to be read as floating point data
        red_tuple = struct.unpack('f' * in_ds.RasterXSize, red_scanline)        
        
        # Read in data for the current line from the image band representing the NIR wavelength
        nir_scanline = in_ds.GetRasterBand(4).ReadRaster( 0, line, in_ds.RasterXSize, 1, in_ds.RasterXSize, 1, gdal.GDT_Float32 )
        # Unpack the line of data to be read as floating point data
        nir_tuple = struct.unpack('f' * in_ds.RasterXSize, nir_scanline)    
        
        # Loop through the columns within the image
        for i in range(len(red_tuple)):
            #Calculate the NDVI for the current pixel.
            ndvi_lower = (nir_tuple[i] + red_tuple[i])
            ndvi_upper = (nir_tuple[i] - red_tuple[i])
            #Be careful of zero divide
            if ndvi_lower == 0:  ndvi = 0
            elif ndvi_upper < 0: ndvi = 0
            else:                ndvi = ndvi_upper/(ndvi_lower*1.0)
            # Add the current pixel to the output line
            outputLine = outputLine + struct.pack('f', ndvi)
            
        # Write the completed line to the output image
        out_ds.GetRasterBand(1).WriteRaster(0, line, in_ds.RasterXSize, 1, outputLine, buf_xsize=in_ds.RasterXSize, buf_ysize=1, buf_type=gdal.GDT_Float32)                
    
    print('len(outputLine.split())',len(outputLine.split()))
    # Delete the output line following write
    del outputLine        
    del out_ds
    return 

def Save_GeoTiff(im,inFilename,outFilename,ROI=None, GSD=None):    
    # clip the image if ROI
    if ROI != None:         
        im= im[ROI[0][0]:ROI[0][1],ROI[1][0]:ROI[1][1]] # im[Y1:Y2,X1:X2]
        
    # Set spatial characteristics of the output object to mirror the input
    src = rasterio.open(inFilename)
    meta = src.meta
    #print('meta=',meta)
    #if meta['crs']==None: print('no crs')   #Non-Georeferencing
    
    #out_dtype = meta['dtype']
    out_dtype = im.dtype
    if len(im.shape)>2: out_count = im.shape[2] # 3b img
    else: out_count = 1
    if out_dtype == 'uint8': meta.update(dtype=out_dtype, nodata=None, count=out_count)
    else:                    meta.update(dtype=out_dtype, count=out_count)

    # clip the image if ROI by transfrom window 
    #print('ROI=',ROI,'h,w=',im.shape[0],im.shape[1],meta['height'],meta['width'])
    if im.shape[0]!=meta['height'] or im.shape[1]!=meta['width']:
        if GSD!=None: 
            gt = src.transform
            print('gt=',gt)
            gt = gt * gt.scale(GSD[0]/gt[0],GSD[1]/gt[4])  #gt[0] for GSDx, gt[4] for GSDy
            print('new gt=',gt)            
        else: gt = src.window_transform(ROI)
        meta.update(width=im.shape[1],height=im.shape[0],transform=gt)
    
    # Create the geotif file
    with rasterio.open(outFilename, 'w+', **meta) as dst:
        if len(im.shape)==2 or im.shape[2]==1: 
            print('im.shape=',im.shape,'%.3f (%.3f~%.3f)'%(np.nanmean(im),np.nanmin(im),np.nanmax(im)))
            if len(im.shape)>2 and im.shape[2]==1: im = im.reshape(im.shape[0],im.shape[1]) # (H,W,1)->(H,W) for write_band()
            print('out_dtype=',out_dtype,'im=',im.shape,im.dtype,'%.3f (%.3f~%.3f)'%(np.nanmean(im),np.nanmin(im),np.nanmax(im)))
            dst.write_band(1, im.astype(out_dtype))    # 1b img
            return im.astype(out_dtype)
        else:                
            #decompose 4b image bands to bring back to the original band order, e.g. RGBN->BGRN
            if len(im.shape)>2 and im.shape[2]==4:
                fbasename = inFilename.split('/')[-1].split('.')[0]
                if fbasename.lower().find('bgrn')>=0 or fbasename.lower().find('rgbn')<0: # bgrn or non-rgbn in filename
                    im = im[:,:,[2,1,0,3]]   # RGBN -> BGRN Red@b2 in PS2 image 
                    #print('Image changed: RGBN -> BGRN')             
            dst.write(np.moveaxis(im.squeeze(),-1,0)) # To make np array in (band,R,C), move axis with 4 entries to beginning and remove extra dimension)
    
# Create the output dataset    
def Create_GeoTiff(in_ds,outFilename,im):  
    mode = 3    #1:1b extract from in_bands, 2:in-bands (copy) 3:8-bit from imgscn, None=in-bands (identical)    

    #generate a conversion dictionary
    NP2GDAL = {
      "uint8": 1,
      "int8": 1,
      "uint16": 2,
      "int16": 3,
      "uint32": 4,
      "int32": 5,
      "float32": 6,
      "float64": 7,
      "complex64": 10,
      "complex128": 11,
    } 
    
    if mode == None:
        out_ds = createOutputImage(in_ds, outFilename)
    elif mode == 1:  
        out_bands = 1;       dtype = in_ds.GetRasterBand(1).DataType
        out_ds = createOutputImage(in_ds, outFilename, out_bands, dtype)
        # copy 1b from input bands to 1b output
        in_data = in_ds.GetRasterBand(1).ReadAsArray()   
        out_ds.WriteRaster(0,0,in_ds.RasterXSize,in_ds.RasterYSize,in_data.tostring(),in_ds.RasterXSize,in_ds.RasterYSize)
    elif mode == 2:
        out_bands = in_ds.RasterCount; dtype = in_ds.GetRasterBand(1).DataType
        out_ds = createOutputImage(in_ds, outFilename, out_bands, dtype)
        # copy pixel data from input band to bands of output
        in_data = in_ds.ReadAsArray(0,0,in_ds.RasterXSize,in_ds.RasterYSize)   
        out_ds.WriteRaster(0,0,in_ds.RasterXSize,in_ds.RasterYSize,in_data.tostring(),in_ds.RasterXSize,in_ds.RasterYSize)
    elif mode==3:        
        out_bands = len(im); dtype = NP2GDAL[im[0].dtype.name]        
        out_ds = createOutputImage(in_ds, outFilename, out_bands, dtype)
        # Loop through each line in turn.
        for k in range(out_bands):
            out_ds.GetRasterBand(k+1).WriteArray(im[k])   #GetRasterBand(1): 1st band    
    
    # Check the datasets was successfully created.
    if out_ds is None:
        print ('Could not create output image')
        sys.exit(-1)
        
    del out_ds
    return 

#Extract dir, image name from filename
def ExtractFilename(filename):    
    fdir = filename[:filename.find(filename.split('/')[-1])]        
    fname = filename.split('/')[-1]
    fbasename = fname.split('.')[0]
    ftype = fname.split('.')[-1]
    #print('fdir=',fdir,'fname=',fname,'fbasename=',fbasename,'ftype=',ftype)
    return fdir,fname,fbasename,ftype

#Find parent's directroy in n step
def FindParentDir(src_dir,n):    
    #print('src_dir=',src_dir)
    dst_dir = src_dir[:src_dir.find(src_dir.split('/')[-n])]            
    #print('dst_dir=',dst_dir)
    return dst_dir

#Read Avg raw files to extract data and write to a file (correted-GPS for NDVI sensor, each sensor in CSV Layer)
def Proc4Avg(fdir): 
    src_dir = tkinter.filedialog.askdirectory(parent=fdir,title='Open a folder to load raw txt files') 
    # exit if canceled to open a file
    if len(src_dir)==0:    
        return       
    
    start_time = time.time()
    src_dir = src_dir+'/'
    #dst_dir = src_dir[:src_dir.rfind(src_dir.split('/')[-2])]    #\Avenger\
    dst_dir = src_dir
    print('src_dir=',src_dir,'\ndst_dir=',dst_dir)
    
    ofilename = '%sAvg.csv'%(dst_dir)                        
    #Detect # of files and # of lines
    num_f = 0
    for filename in os.listdir(src_dir): #read files in the dir                
        if filename.endswith("GPS_RawData.txt"): continue #skip GPS_RawData.txt
        if filename.endswith(".dat"): filename_dat = filename
        if filename.endswith(".txt"):
            num_f = num_f + 1
            num_lines = sum(1 for line in open(src_dir + filename))
            print (num_f,': num_lines=',num_lines)
    #define array size
    lst = [0]*num_lines*(num_f)     #entire list of data lines
            
    #Detect # of column from heading and 2nd data line 
    i = 0
    num_col = [0]*num_f
    for filename in os.listdir(src_dir): #read files in the dir                  
        if filename.endswith("GPS_RawData.txt"): continue #skip GPS_RawData.txt
        if filename.endswith(".txt"):                
            print ('%d/%d'%(i+1,num_f),'filename=',filename)
            for j,line in enumerate(open(src_dir + filename)):
                if j == 0 or j == 2: 
                    lst[i*num_lines+j] = line.split(',')    
                    if j == 2: 
                        num_col[i] = max(len(lst[i*num_lines+0]), len(lst[i*num_lines+2]))                        
            i = i + 1
    
    #Store the data points into 3-D array
    i = 0
    for filename in os.listdir(src_dir): #read files in the dir                  
        if filename.endswith("GPS_RawData.txt"): continue #skip GPS_RawData.txt
        if filename.endswith(".txt"):                
            with open(src_dir+ filename) as fp:                    
                for j, line in enumerate(fp):                
                    lst[i*num_lines+j] = line.split(',') 
                    #print (i,j,len(lst[i*num_lines+j]),num_col[i],filename,end='->')
                    #detect EOL and delete linefeed
                    lst[i*num_lines+j][len(lst[i*num_lines+j])-1] = lst[i*num_lines+j][len(lst[i*num_lines+j])-1][:-1]
                    #print (i,j,lst[i*num_lines+j][len(lst[i*num_lines+j])-1])
                    #align col headings 
                    if len(lst[i*num_lines+j])<num_col[i]:
                        #add comma for the empty data
                        for kk in range(len(lst[i*num_lines+j]),num_col[i]):
                            lst[i*num_lines+j].append("")
                            #print (kk,lst[i*num_lines+j])
                        #print (len(lst[i*num_lines+j]))
            i = i + 1
    
    #row index of sensors       
    #filename_dat = LoadFile(self)
    CC2BB,yaw_off,a,ctr_off,arm_n,RS,Heading_n,GPS,IMU,CC,T,Tb,H,P,GPS_n,IMU_n,CC_n,T_n,Tb_n,H_n,P_n = Read_Platform4Avg(src_dir+ filename_dat)
    #print (IMU,IMU_n)
    #print(ctr_off,arm_n)

    #write the data combined from raw data files to a GPS-corrected csv file 
    LL = [0]*arm_n    
    CC_bands = [0]*6
    with open(ofilename,'w+') as fo:                                            
        #Combine data files into a single file per timestamp
        print4monitor = 0
        for j in range(num_lines):
            if j== 0: print (num_lines,end=" lines:")                
            if j%(int(num_lines/10)) == 0: print("-",end="")
            if j == 0: 
                fo.write('Easting,Northing,Easting_arm,Northing_arm,')   #insert Lat\Lon colomns before 1st file 
                for k in range(Heading_n): fo.write(lst[1*num_lines+j][k]+',')  #write the first common columns
                fo.write('Lat, Lon,')   # GPS coord.
                fo.write('VI1b (R2 R1),VI2b (R2 R3), R1b (590),R2b (730),R3b (530),')   #CC_B
                fo.write('VI1f (R2 R1),VI2f (R2 R3), R1f (550),R2f (800),R3f (670),')   #CC_F
                for k in range(H,H+H_n): fo.write(lst[2*num_lines+j][k-1]+',')    #H
                for k in range(IMU,IMU+IMU_n): fo.write(lst[3*num_lines+j][k]+',')    #IMU, NOTE: align col headings by skiping a bug '-a' of col heading in IMU_RawData.txt
                for k in range(T,T+T_n): fo.write(lst[4*num_lines+j][k-1]+',')    #T (N,F)
                for k in range(Tb,Tb+Tb_n): fo.write(lst[4*num_lines+j][k-1]+',')    #Tb (N,F)
                for k in range(P,P+P_n): fo.write(lst[5*num_lines+j][k-1]+',')    #P       
                fo.write('NDVIA800,NDARE,NDVIR800,NDREE,NDRE,NDVIG800,PIR,CI800,CCCI,DATT,DATTA,CCCIA')  #VIs
                fo.write('\n')    
            else: 
                #Read Lat/Lon & convert to UTM
                for g in range(GPS,GPS+GPS_n): LL[g-GPS] = lst[1*num_lines+j][g-1]
                if len(LL[0]) == 0: continue    #skip the iteration if no GPS
                Lon_deg0,Lat_deg0,Easting0,Northing0 = Convert_LL2UTM(LL)
                #write row per line
                yaw = radians(float(lst[3*num_lines+j][IMU-1])+yaw_off)   #deg2rad
                #if j>400 and j<4000: print(lst[3*num_lines+j],float(lst[3*num_lines+j][IMU-1]),yaw)

                #correct GPS center offset
                b = -ctr_off  
                Easting = Easting0 + (-b*cos(yaw))/100
                Northing = Northing0 + (b*sin(yaw))/100
                Lat_deg, Lon_deg = Convert_UTM2LL(Easting,Northing,12)
                for kk in range(arm_n):                        
                    
                    #Correct GPS heading offsets
                    b = ctr_off + RS*(kk-arm_n/2 +1/2)*2.54  #inch2cm    
                    Easting,Northing = Correct_GPSoffset(Easting0,Northing0,a/100,b/100,yaw)
                    #print(j,kk,'b (left) =,',b,'E=',Easting,'N=',Northing) 
                    
                    fo.write(str(Easting0)+','+str(Northing0)+','+str(Easting)+','+str(Northing)+',') #insert UTM colomns before 1st file                         
                    for k in range(Heading_n): fo.write(lst[1*num_lines+j][k]+',')  #write the first common columns  
                    fo.write(str(Lat_deg0)+','+str(Lon_deg0)+',') #insert Lat\Lon colomns after 1st file
                    n = 0
                    CC_flag = 0
                    for k in range(CC+kk*CC_n,CC+(kk+1)*CC_n): 
                        fo.write(lst[0*num_lines+j][k-1]+',')    #CC (B,F)
                        #Store reflectance of 6 bands
                        if (k>=CC+kk*CC_n+2 and k<=CC+kk*CC_n+4) or (k>=CC+kk*CC_n+7 and k<=CC+kk*CC_n+9) : 
                            #print (len(lst[0*num_lines+j][k-1]),lst[0*num_lines+j][k-1])
                            if len(lst[0*num_lines+j][k-1])==0 or float(lst[0*num_lines+j][k-1])==0: CC_flag = 1
                            else: CC_bands[n] = float(lst[0*num_lines+j][k-1])
                            n = n + 1
                    for k in range(H+kk,H+kk+H_n): fo.write(lst[2*num_lines+j][k-1]+',')    #H
                    #print(IMU,kk,IMU_n,':',IMU+kk,IMU+kk+IMU_n,'num_lines=',num_lines,j,k)
                    for k in range(IMU,IMU+IMU_n): fo.write(lst[3*num_lines+j][k-1]+',')    #IMU
                    for k in range(T+kk*T_n,T+(kk+1)*T_n): fo.write(lst[4*num_lines+j][k-1]+',')    #T (N,F)
                    for k in range(Tb+kk*Tb_n,Tb+(kk+1)*Tb_n): fo.write(lst[4*num_lines+j][k-1]+',')    #Tb (N,F)
                    for k in range(P+kk,P+kk+P_n): fo.write(lst[5*num_lines+j][k-1]+',')    #P              
                    #Calculate VIs
                    if CC_flag == 0:
                        NDVIA800,NDARE,NDVIR800,NDREE,NDRE,NDVIG800,PIR,CI800,CCCI,DATT,DATTA,CCCIA = CalculateVIs(CC_bands)
                        #print (NDVIA800,NDARE,NDVIR800,NDREE,NDRE,NDVIG800,PIR,CI800,CCCI,DATT,DATTA,CCCIA)
                        fo.write(str(NDVIA800)+','+str(NDARE)+','+str(NDVIR800)+','+str(NDREE)+','+str(NDRE)+','+str(NDVIG800)+','+str(PIR)+','+str(CI800)+','+str(CCCI)+','+str(DATT)+','+str(DATTA)+','+str(CCCIA)+',')    #P              
                    else:                            
                        fo.write(',,,,,,,,,,,,')    #P              
                    fo.write('\n')                            
    
    #write the data combined from raw data files to a csv file
    #with open(ofilename,'w+') as fo:                                            
        ##Combine data files into a single file per timestamp
        #print4monitor = 0
        #for j in range(num_lines):
            #print ("%d/%d "%(j,num_lines),end="")
            #for i in range(num_f):
                ##insert Lat\Lon colomns before 1st file 
                #if i==0: 
                    #if j == 0: fo.write('Lat,Lon,Northing,Easting,')
                    #else: 
                        ##Read Lat/Lon & convert to UTM
                        #LL = [0]*4                            
                        #for g in range(11,15): LL[g-11] = lst[1*num_lines+j][g]
                        #Lon_deg,Lat_deg,Easting,Northing = Convert_LL2UTM(LL)
                        ##print(Lat_deg,Lon_deg,'N=',Northing,'E=',Easting)
                        ##print(Convert_UTM2LL(Easting, Northing, 12))
                        #fo.write(str(Lat_deg)+','+str(Lon_deg)+','+str(Northing)+','+str(Easting)+',') 
                #elif i == 1: continue   #jump to the top of the loop
                #for k in range(len(lst[i*num_lines+j])):
                    #if k == len(lst[i*num_lines+j])-1: #skip EOL linefeed to add all data per timestamp until the last file
                        ##add comma until detecting the end of data line
                        #if print4monitor: 
                            #print(lst[i*num_lines+j][k][:-1],end="")
                            #if i+1<num_f: print(',',end="")
                        #if i==3 and j>0: pass #NOTE: align col headings by skiping a bug of 2nd col heading in IMU_RawData.txt
                        #else:
                            #fo.write(lst[i*num_lines+j][k][:-1])
                            #if i+1<num_f: fo.write(',')                                
                    #else:
                        #if print4monitor: print(lst[i*num_lines+j][k],',', end =" ")                            
                        #if i==3 and j==0 and k==9: pass #NOTE: align col headings by skiping a bug of 2nd col heading in IMU_RawData.txt
                        #else:
                            #if i>0 and k<9: pass    #skip the common heading cols in the rest of files
                            #else:
                                #fo.write(lst[i*num_lines+j][k]+',')
                        
            #print('')   #create a linefeed, cf)'\n' creates additional linefeed
            #fo.write('\n')        
    hh,mm,ss = ConvertSec2hhmmss(float(time.time()-start_time))
    print("Done! (run time = %02d:%02d:%02.1f seconds)"%(hh,mm,ss))

#Frequency-domain image filters
def FFT(im):
    import scipy as sp
    rows, cols = im.shape[0], im.shape[1]    
    #rows = cv2.getOptimalDFTSize(rows)     #find a new row for the fastest when array size is 2^n, but resized
    #cols = cv2.getOptimalDFTSize(cols)     #find a new col for the fastest when array size is 2^n, but resized
    print('rows,cols =',im.shape[0],im.shape[1],'=>',rows,cols, end=' ')
    im_fft = np.fft.fft2(im,[rows,cols])  
    im_fshift = np.fft.fftshift(im_fft)     #shift DC component (0Hz) at TopLeft => Center
    im_log = 20*np.log(np.abs(im_fshift))   #abs to extract real parts and log the large number

    #LPF: create a mask first, center square is 1, remaining all 0 (bgnd masking)
    mask = np.zeros((rows,cols),np.uint8)        
    mask[round(rows/2)-30:round(rows/2)+30, round(cols/2)-30:round(cols/2)+30] = 1  #LPF w 60x60 window        
    im_LPF = im_fshift*mask    # apply mask
    
    #HPF: FFT image's center square is 0(bgnd masking)    
    im_fshift[round(rows/2)-30:round(rows/2)+30, round(cols/2)-30:round(cols/2)+30] = 0    #HPF w 60x60 window
    im_HPF = im_fshift
    
    filters = [im_LPF, im_HPF]
    filter_name = ['Input Image','FFT','LPF', 'HPF','Guassian','Median','Sobel','Laplacian']    
    f_ishift = [np.fft.ifftshift(x) for x in filters]   #inverse FFT
    Im_fft = [np.fft.ifft2(y) for y in f_ishift]        #shift DC component (0Hz) back at Center => TopLeft
    Im = [np.abs(z) for z in Im_fft]
    Im.insert(0,im), Im.insert(1,im_log)
    Im.append(sp.ndimage.gaussian_filter(im,1))
    Im.append(sp.ndimage.median_filter(im,1))
    Im.append(sp.ndimage.sobel(im,1))   
    Im.append(sp.ndimage.laplace(im))   

    for i in range(8):
        plt.subplot(4,2,i+1),plt.imshow(Im[i],cmap = 'gray')
        plt.title(filter_name[i]), plt.xticks([]), plt.yticks([])    
    plt.show()        

    return im_log 

def About():
    messagebox.showinfo(title='About',message='IMAP ver. 1')
    #ret = messagebox.askyesno(title='Config',message='Change?')
    #if ret > 0:
        #pass
    return

class Dialog_BandMath(simpledialog.Dialog):
    def __init__(self, parent, eqBM, eqBM_Ent, Bands_Math, fdir): 
        self.Bands = Bands_Math
        self.eqBM = eqBM
        self.eqBM_Ent = eqBM_Ent
        self.fields = '(b1-b2)/(b1+b2)','(B-R)/(B+R)', '(G-R)/(G+R)', ''
        self.Dlg_BM = tk.Toplevel()  #create another window using Toplevel()
        self.Dlg_BM.title("Band Math")
        #Dlg_BM.geometry("%dx%d%+d%+d" % (250, 150, 100, 100))        
        self.fdir = fdir
        print('self.fdir=',self.fdir)
        
        
        # create a widget for filename of CH1
        self.entBand = []
        for i,band in enumerate(self.Bands):
            frm = tk.Frame(self.Dlg_BM)
            frm.pack(side=tk.TOP, fill=tk.X, padx=5, pady=0)
            #Add a button to load a file
            btn = tk.Button(frm,text="b%d:"%(i+1),command=(lambda ch=i: self.btn_LoadFile(ch,self.fdir)))
            btn.pack(side=tk.LEFT)
            self.entBand.append(tk.Entry(frm, width = 80))
            self.entBand[i].pack(side=tk.LEFT)        
            self.entBand[i].insert(0,band)    #initialize an entry value      

        # create a widget for Radiobuttons
        self.var = tk.IntVar()
        self.var.set(self.eqBM) # initialize
        for i,field in enumerate(self.fields):
            frm = tk.Frame(self.Dlg_BM)
            frm.pack(side=tk.TOP, fill=tk.X, padx=5, pady=0)
            self.RdBtn = tk.Radiobutton(frm, text=field, variable=self.var,value=i).pack(side=tk.LEFT,anchor=tk.NW)
        
        # create a widget for Entry
        self.ent = tk.Entry(frm, width = 14)
        self.ent.pack(side=tk.LEFT)        
        self.ent.insert(0,self.eqBM_Ent)    #initialize an entry value

        # create a widget for Ok\Cabcek Buttons
        self.OkBtn = tk.Button(self.Dlg_BM,text='Ok', width=10,command=self.OkHandler)    
        self.OkBtn.pack(side=tk.LEFT, padx=20, pady=5)   #create a pack        
        self.CancelBtn = tk.Button(self.Dlg_BM,text='Cancel', width=10,command=self.CancelHandler)    
        self.CancelBtn.pack(side=tk.LEFT, padx=20, pady=5)   #create a pack        
        tk.simpledialog.Dialog.wait_window(self.Dlg_BM)    
    
    def btn_LoadFile(self,ch,fdir):
        self.file = LoadFile(fdir,'img')  
        self.fdir,fname,fbasename,ftype = ExtractFilename(self.file)
        print('self.file=',self.file)
        self.Bands[ch] = self.file
        self.entBand[ch].delete(0,tk.END)   # clear the entry
        self.entBand[ch].insert(0,self.Bands[ch])    # replace the entry value  
        self.Dlg_BM.lift()  # bring back the Dlg in front
        #self.Dlg_BM.attributes("-topmost",True)  # bring always the Dlg in front
        
    def OkHandler(self):            
        print('self.ent.get()=',self.ent.get(),'self.RdBtn=',self.RdBtn,'self.var.get()=',self.var.get())        
        self.eqBM = self.var.get()
        if self.eqBM_Ent != self.ent.get():            
            self.eqBM_Ent = self.ent.get()
            self.eqBM = 2
        self.cancel = False
        self.Dlg_BM.destroy()
        
    def CancelHandler(self):
        self.cancel = True
        self.Dlg_BM.destroy()
        

class Dialog_Merge(simpledialog.Dialog):
    def __init__(self, parent, Bands_Merge, fdir): 
        self.Bands = Bands_Merge
        self.Dlg_Mg = tk.Toplevel()  #create another window using Toplevel()
        self.Dlg_Mg.title("Band Merge")
        self.fdir = fdir
        print('self.fdir=',self.fdir)
        
        
        # create a widget for filename of CH1
        self.entBand = []
        for i,band in enumerate(self.Bands):
            frm = tk.Frame(self.Dlg_Mg)
            frm.pack(side=tk.TOP, fill=tk.X, padx=5, pady=0)
            #Add a button to load a file
            btn = tk.Button(frm,text="b%d:"%(i+1),command=(lambda ch=i: self.btn_LoadFile(ch,self.fdir)))
            btn.pack(side=tk.LEFT)
            self.entBand.append(tk.Entry(frm, width = 80))
            self.entBand[i].pack(side=tk.LEFT)        
            self.entBand[i].insert(0,band)    #initialize an entry value      

        # create a widget for Ok\Cabcek Buttons
        self.OkBtn = tk.Button(self.Dlg_Mg,text='Ok', width=10,command=self.OkHandler)    
        self.OkBtn.pack(side=tk.LEFT, padx=20, pady=5)   #create a pack        
        self.CancelBtn = tk.Button(self.Dlg_Mg,text='Cancel', width=10,command=self.CancelHandler)    
        self.CancelBtn.pack(side=tk.LEFT, padx=20, pady=5)   #create a pack        
        tk.simpledialog.Dialog.wait_window(self.Dlg_Mg)    
    
    def btn_LoadFile(self,ch,fdir):
        self.file = LoadFile(fdir,'img')  
        self.fdir,fname,fbasename,ftype = ExtractFilename(self.file)
        print('self.file=',self.file)
        self.Bands[ch] = self.file
        self.entBand[ch].delete(0,tk.END)   # clear the entry
        self.entBand[ch].insert(0,self.Bands[ch])    # replace the entry value  
        self.Dlg_Mg.lift()  # bring back the Dlg in front
        #self.Dlg_Mg.attributes("-topmost",True)  # bring always the Dlg in front
        
    def OkHandler(self):            
        self.cancel = False
        self.Dlg_Mg.destroy()
        
    def CancelHandler(self):
        self.cancel = True
        self.Dlg_Mg.destroy()
        
class Dialog_CalibTarget(simpledialog.Dialog):
    def __init__(self, parent, iCal, iCal_Ent): 
        self.iCal = iCal
        self.iCal_Ent = iCal_Ent
        self.fields = '18%', '48%', '% (R,G,B,N):'
        self.Dlg_TargetRef = tk.Toplevel()  #create another window using Toplevel()
        self.Dlg_TargetRef.title("Calibration Target")
        #Dlg_TargetRef.geometry("%dx%d%+d%+d" % (250, 150, 100, 100))        
        
        # create a widget of Checkbutton for ROI segmentation option
        self.Seg_var = tk.IntVar()
        self.Seg_var.set(0) # initialize to skip ROI seg as default
        self.CkBtn = tk.Checkbutton(self.Dlg_TargetRef, text='ROI Seg', variable=self.Seg_var).pack(side=tk.TOP,anchor=tk.NW,padx=5)

        # create a widget for Radiobuttons
        self.var = tk.IntVar()
        self.var.set(self.iCal) # initialize 
        for i,field in enumerate(self.fields):
            frm = tk.Frame(self.Dlg_TargetRef)
            frm.pack(side=tk.TOP, fill=tk.X, padx=5, pady=0)
            self.RdBtn = tk.Radiobutton(frm, text=field, variable=self.var,value=i).pack(side=tk.LEFT,anchor=tk.NW)
        
        # create a widget for Entry
        self.ent = tk.Entry(frm, width = 14)
        self.ent.pack(side=tk.LEFT)        
        self.ent.insert(0,self.iCal_Ent)    #initialize an entry value
        #lab = tk.Label(frm, text='%')
        #lab.pack(side=tk.LEFT)

        # create a widget for OK\Cancel Buttons
        self.OkBtn = tk.Button(self.Dlg_TargetRef,text='Ok', width=10,command=self.OkHandler)    
        self.OkBtn.pack(side=tk.LEFT, padx=20, pady=5)   #create a pack        
        self.CancelBtn = tk.Button(self.Dlg_TargetRef,text='Cancel', width=10,command=self.CancelHandler)    
        self.CancelBtn.pack(side=tk.LEFT, padx=20, pady=5)   #create a pack        
        tk.simpledialog.Dialog.wait_window(self.Dlg_TargetRef)    

    def OkHandler(self):            
        print('self.ent.get()=',self.ent.get(),'self.RdBtn=',self.RdBtn,'self.var.get()=',self.var.get())        
        print('self.CkBtn=',self.CkBtn,'self.Seb_var.get()=',self.Seg_var.get())
        self.iCal_Seg = self.Seg_var.get()
        self.iCal = self.var.get()
        if self.iCal_Ent != self.ent.get():            
            self.iCal_Ent = self.ent.get()
            self.iCal = 2
        self.cancel = False
        self.Dlg_TargetRef.destroy()
        
    def CancelHandler(self):
        self.cancel = True
        self.Dlg_TargetRef.destroy()

class ScrollableFrame(tk.Frame):
    def __init__(self, parent, min_canvas_size):        
        tk.Frame.__init__(self, parent)
        #print('min_canvas_size=',min_canvas_size)

        # create scrollbars
        self.sbarV = tk.Scrollbar(self, orient = tk.VERTICAL)
        self.sbarV.pack(fill = tk.Y, side = tk.RIGHT, expand = tk.FALSE)
        self.sbarH = tk.Scrollbar(self, orient = tk.HORIZONTAL)
        self.sbarH.pack(fill = tk.X, side = tk.BOTTOM, expand = tk.FALSE)

        #Create a canvas object and associate the scrollbars with it
        self.canvas = tk.Canvas(self, bd = 0, highlightthickness = 0, yscrollcommand = self.sbarV.set, xscrollcommand = self.sbarH.set)
        self.canvas.pack(side = tk.LEFT, fill = tk.BOTH, expand = tk.TRUE)

        #Associate scrollbars with canvas view
        self.sbarV.config(command = self.canvas.yview)
        self.sbarH.config(command = self.canvas.xview)

        #? set the view to 0,0 at initialization 
        self.canvas.xview_moveto(0)
        self.canvas.yview_moveto(0)
        
    #Read the mouse wheel with the Tkinter GUI toolkit
    def MouseWheel(self, event):
        self.canvas.yview_scroll(-1*int(event.delta/120), "units") 

class WindowScrollbar:
    def __init__(self,root):
        self.frame = root.master
        
        # scroll Bar x For width
        self.sbarH=tk.Scrollbar(self.frame, orient='horizontal')
        self.sbarH.pack(side='bottom', fill='x')
        
         # Scroll Bar y For Height
        self.sbarV = tk.Scrollbar(self.frame, orient='vertical')
        self.sbarV.pack(side='right', fill='y')  
        
        #Create a canvas object and associate the scrollbars with it
        #self.canvas = tk.Canvas(self, bd = 0, highlightthickness = 0, yscrollcommand = self.sbarV.set, xscrollcommand = self.sbarH.set)
        #self.canvas.pack(side = tk.LEFT, fill = tk.BOTH, expand = tk.TRUE)

        #self.sbarH.config(command=self.canvas.xview)
        #self.sbarV.config(command=self.canvas.yview)

class Dialog_Input4Layer(simpledialog.Dialog):
    def __init__(self,parent, title=None, entry_to_clear=None, var1=None, var2=None,var3=None):     
        self.col_lst = var1
        self.cmap = var2
        self.fbasename = var3
        tkinter.simpledialog.Dialog.__init__(self, parent, title) 
    def body(self, parent):
        self.lb0 = tk.Label(parent, text="Index range : ")
        self.lb0.grid(row=0,sticky=tk.W)
        self.e01 = tk.Entry(parent, width=4)
        self.e01.grid(row=0,column=1,sticky=tk.W)
        self.e01.insert(0,'-.2')
        self.lb1 = tk.Label(parent, text="-")
        self.lb1.grid(row=0,column=2)
        self.e02 = tk.Entry(parent, width=4)
        self.e02.grid(row=0,column=3,sticky=tk.W)
        self.e02.insert(0,'0.8')
        self.var1 = tk.IntVar()
        tk.Checkbutton(parent, text='Strecth', variable=self.var1).grid(row=0,column=4,sticky=tk.W)  
        self.var1.set(0)
        self.e1 = tk.Entry(parent, width=4)
        self.e1.grid(row=0,column=5,sticky=tk.E)
        self.e1.insert(0,'2')
        tk.Label(parent, text="-").grid(row=0,column=6)
        self.e2 = tk.Entry(parent, width=4)
        self.e2.grid(row=0,column=7,sticky=tk.W)
        self.e2.insert(0,'98')
        tk.Label(parent, text="%").grid(row=0,column=8,sticky=tk.W)

        tk.Label(parent, text="Marker size : ").grid(row=1,sticky=tk.W)
        self.e3 = tk.Entry(parent, width=4)
        self.e3.grid(row=1,column=1,sticky=tk.W)
        self.e3.insert(0,'5')
        self.markerSize = 0
        self.var2 = tk.IntVar()
        tk.Checkbutton(parent, text='Connect markers', variable=self.var2).grid(row=1,column=4,columnspan=5,sticky=tk.W)  
        self.var2.set(0)

        # options for GT Plot Map
        self.var3 = tk.IntVar()
        self.CB4Pmap = tk.Checkbutton(parent, text='Plot Mapping', variable=self.var3)
        self.CB4Pmap.grid(row=2,column=0,columnspan=2,sticky=tk.W)  
        self.var3.set(0)
        # add an ComboBox
        self.lb2 = tk.Label(parent, text="R1C1:")
        self.lb2.grid(row=2,column=3,sticky=tk.E)
        self.dic_R1C1 = {'NW':0, 'NE':1, 'SE':2, 'SW':3}
        self.comboboxR1C1 = ttk.Combobox(parent, value = list(self.dic_R1C1.keys()), width=4)
        self.comboboxR1C1.grid(row=2,column=4,sticky=tk.W)
        self.comboboxR1C1.current(0)        
        self.var4 = tk.IntVar()        
        self.CB4BiDir = tk.Checkbutton(parent, text='Bi-directional', variable=self.var4)
        self.CB4BiDir.grid(row=2,column=5,columnspan=4,sticky=tk.W)  
        self.var4.set(1)
        #self.lb3 = tk.Label(parent, text="row:")
        #self.lb3.grid(row=2,column=1,sticky=tk.E)
        #self.e4 = tk.Entry(parent, width=5)
        #self.e4.grid(row=2,column=2,sticky=tk.E)
        #self.e4.insert(0,'185')

        ## disable Index Range if Stretch is checked
        #self.lb0.config(state=tk.DISABLED)
        #self.e01.config(state=tk.DISABLED)
        #self.lb1.config(state=tk.DISABLED)
        #self.e02.config(state=tk.DISABLED)
        
        # disable Plot Mapping if not GT
        if self.fbasename.lower().find('gt')<0:            
            self.CB4Pmap.config(state=tk.DISABLED)
            self.comboboxR1C1.config(state=tk.DISABLED)
            self.lb2.config(state=tk.DISABLED)
            self.CB4BiDir.config(state=tk.DISABLED)            
            #self.lb3.config(state=tk.DISABLED)
            #self.e4.config(state=tk.DISABLED)
        
        ## add Checkbuttons
        #cb = [0]*10
        #for i in range(10):
            #cb[i] = tk.Checkbutton(parent, text="cb%d"%(i))
            #cb[i].grid(row=1, column=i, sticky=tk.W)        

        # add an ComboBox
        #print('col_lst=',self.col_lst)
        self.col_VI = []
        self.col_NDVI = 0
        for i in range(len(self.col_lst)):
            #print(i,self.col_lst[i],self.col_lst[i].find('VI'))
            if (self.col_lst[i].lower().find('vi')>=0 or self.col_lst[i].lower().find('honey')>=0 or
            self.col_lst[i].lower().find('yaw')>=0 or self.col_lst[i].lower().find('pitch')>=0 or self.col_lst[i].lower().find('roll')>=0 or
            self.col_lst[i].lower().find('irt')>=0 or self.col_lst[i].lower().find('pep')>=0 or self.col_lst[i].lower().find('ndre')>=0 or
            self.col_lst[i].lower().find('pir')>=0 or self.col_lst[i].lower().find('ci')>=0 or 
            self.col_lst[i].lower()=='nir' or self.col_lst[i].lower()=='re' or self.col_lst[i].lower()=='r'):
                self.col_VI.append(i)           
                if self.col_lst[i].lower().find('ndvi') == 0 and self.col_NDVI==0:
                    self.col_NDVI = len(self.col_VI) - 1
                
        #print('self.col_VI=',self.col_VI)
        dic_VI = {self.col_lst[self.col_VI[i]]:self.col_VI[i] for i in range(len(self.col_VI))}
        print('dic_VI=',dic_VI)
        #print('list(dic_VI.keys())=',list(dic_VI.keys()))
        self.combobox = ttk.Combobox(parent, value = list(dic_VI.keys()))
        self.combobox.current(self.col_NDVI)        
        self.combobox.grid(row=3,columnspan=4)

        b = tk.Button(parent, text='Choose Color', fg="darkgreen", command=self.callback).grid(row=3,column=4,columnspan=3,sticky=tk.W)   
        

    def apply(self):
        if self.e01.get()=="": self.pct1 = 0
        else: self.idxL = float(self.e01.get()) # 0
        if self.e02.get()=="": self.pct2 = 0
        else: self.idxH = float(self.e02.get()) # 1        
        if self.e1.get()=="": self.pct1 = 0
        else: self.pct1 = float(self.e1.get()) # 2%
        if self.e2.get()=="": self.pct2 = 0
        else: self.pct2 = float(self.e2.get()) # 98%        
        if self.e3.get()=="": self.markerSize = 0   # marker size
        else: self.markerSize = int(self.e3.get()) # 5
        #if self.e4.get()=="": self.markerSize = 0   # starting at n-th row in csv
        #else: self.RowCSV = int(self.e4.get()) # 5
        self.col = self.col_VI[self.combobox.current()]
        self.stretch = self.var1.get()        
        self.connect = self.var2.get()        
        self.plotmap = self.var3.get() 
        self.bidir = self.var4.get() 
        self.R1C1 = list(self.dic_R1C1)[self.comboboxR1C1.current()]
        #print ('idx=',self.idxL,self.idxH,"pct=",self.pct1,self.pct2,'stretch=',self.stretch,'col=',self.col, 'connect=',self.connect,'cmap=',self.cmap, 'ms=',self.markerSize)
        #if self.fbasename.lower().find('gt')>=0:
            #print ('plotmap=',self.plotmap,'R1C1=',self.R1C1,'bidir=',self.bidir) # 'RowCSV=',RowCSV

    def callback(self):
        from tkinter.colorchooser import askcolor 
        cmap, web_color  = askcolor(color="#6A9662", title = "Color Chooser") 
        if cmap[0] is None: cmap = [255,0,0]
        self.cmap = cmap
        print ('cmap=',cmap[0],web_color, 'self.cmap=',self.cmap)
    
    #def cancel(self, event=None):  #NOTE: fix to skip cancel when clicking ok
        ## put focus back to the parent window
        #self.pct1 = 0        
        #self.col = self.combobox.current() # 34
        #self.parent.focus_set()
        #self.destroy()

class Window(tk.Frame):
    def __init__(self):     
        #create a window of main tk object        
        self.root = tk.Tk() 
        
        # window size to fit the screen
        self.scr_w = self.root.winfo_screenwidth()
        self.scr_h = self.root.winfo_screenheight()
        #print ('screen size=',self.scr_w,'x ',self.scr_h)
        self.root.geometry('%dx%d+0+0'%(self.scr_w,self.scr_h))
        
        # init Frame
        tk.Frame.__init__(self, self.root)
        min_canvas_size = (500, 500)

        # define outside margins of the image frame
        margin_w = 400
        margin_h = 190

        # create an image display frame (gray window)
        self.frame_w = self.scr_w-margin_w
        self.frame_h = self.scr_h-margin_h
        
        # create a paned window
        pWin = ttk.PanedWindow(self.root, orient = tk.HORIZONTAL)
        pWin.pack(fill=tk.BOTH, expand=True)
    
        ## create a label frame
        #tk.LabelFrame(self.root, height=480, width=640, text='My Frame').pack()
        frame1 = ttk.Frame(pWin, width=margin_w, height=self.frame_h, relief=tk.SUNKEN)
        self.frame2 = ttk.Frame(pWin, width=margin_w, height=self.frame_h)
        pWin.add(frame1, weight=4)  # keep the ratio
        pWin.add(self.frame2)
        
        # create window screenbars
        #WindowScrollbar(self.root)
        
        # create frame screenbars
        self.frame = ScrollableFrame(frame1, min_canvas_size)
        #self.frame = ScrollableFrame(self.root, min_canvas_size)
        self.frame.pack(fill=tk.BOTH, expand=tk.YES)        
        
        # create a resizable canvas info of the main frame
        #myframe = tk.Frame(self.root)
        #myframe.pack(fill=tk.BOTH, expand=tk.YES)
        #mycanvas = ResizingCanvas(frame,height=self.scr_h,width=self.scr_w, highlightthickness=0)
        #mycanvas.pack(fill=tk.BOTH, expand=tk.YES)        
        # add some widgets to the canvas
        #mycanvas.create_rectangle(350, 25, 370, 75, fill="blue")            
        #mycanvas.create_window(0, 0, window=frame, height=self.scr_h,width=self.scr_w)   # ERROR: image position, partly cut, scrl bars        
        # tag all of the drawn widgets
        #mycanvas.addtag_all("all")
        
        # initiate window to add menu
        self.init_window()              
        #self.__add_plot()
        
        # read config
        self.Read_Config()        

        # initialize variables
        self.btn_ROI_var = 0        
        self.ch = 'R'   #1-ch img, R,G,B,RGB,HSV,LAB,NDVI,Heatmap,Gray,C&S,RGBN
        self.Bands_Math = ['band1','band2'] # two Bands for Band Math
        self.Bands_Merge = ['band1','band2','band3','band4'] # four Bands for Band Merge
        self.eqBM = 0 #0:(B-R)/(B+R), 1:(G-R)/(G+R), 2:arbitrary
        self.eqBM_Ent = '(B-G)/(B+G)' #2:arbitrary 
        self.iCal = 0 #0:8%, 1:48%, 2:arbitrary
        self.iCal_Ent = '10' #2:arbitrary 
        self.mask_file = [] #mask file name        
        self.zf = 1    #zoom factor
        self.pre_zf = 1 #previous zoom factor
        self.rot = 0    #zoom factor
        self.ctrl_btn = 0   #ctrl key unpressed
        self.shft_btn = 0   #shift key unpressed
        self.alt_btn = 0   #alt key unpressed
        self.MouseBtn2 = False #mouse button-2 (middle) key unpressed
        self.georef = 0
        self.flag_im = 0   # flag for im to avoid reopening Read_Config()
        self.flag_im_base = 0   # flag of im_base for shapefile display
        self.flag_segStats = 0  # flag to skip Seg Stats 
        self.flag_ResetROI = 0  # flag to reset ROI
        self.flag_Stretch = 0   # flag to stretch to min/max
        self.flag_zm = True   # flag for zooming to resize only when changed
        self.flag_BatchCal = 0   # flag for calibration in Batch procwess
        self.flag_BatchClip2ROI = 1   # flag for saving images in ROI only
        self.flag_BatchSaveImgHeatmap = 1   # flag for saving Heatmap image in Batch process
        self.flag_BatchSaveImgHeatmapSeged = 1   # flag for saving HeatmapSeged image in Batch process
        self.flag_BatchSaveImgNDVI = 0   # flag for saving NDVI image in Batch process
        self.flag_BatchSaveImgMask = 0   # flag for saving Mask image in Batch process
        self.GRID_CntP = 0 # GRID foreground count in %
        self.max_y = 0  # max height of 3b image in histogram
        self.btn_cmap = 'RGB'
        self.cmap = 'RdYlGn'    # default color map for heatmap
        self.maskSrc = 'RGB'    # default mask source in batch process
        self.filterSrc = 'None' # default filtering source in batch process
        self.B = -1 # 1b selection: -1(none), 0(R), 1(G), 2(B)
        self.c_REFr = [8,10,9,45]   # default reflectance % of a taget check plot
        self.cropID ='' # default cropID is black unless ROI file opened
        self.pre_ROI0 = 0   # previous ROI[0] to detect tROI change
        self.ROI_rot = 0 # default angle for ROI rotation
        #self.seg = False # flag for Seg to skip histogram Update_imScn

    #Creation of init_window
    def init_window(self):              
        self.master.title("IMAP")        #change the title of our master widget        
        self.pack(fill=tk.BOTH, expand=1)   #allow the widget to take the full space of the root window        
        Menubar = tk.Menu(self.master)     #create a menu instance
        self.master.config(menu=Menubar)

        #create File object 
        self.FileMenu = tk.Menu(Menubar,tearoff=0)    
        Menubar.add_cascade(label='File',menu=self.FileMenu) #added "File" to main menu 
        # adds commands to the menu option
        self.FileMenu.add_command(label='Open Image',command=self.Load_Image)
        self.FileMenu.add_command(label='Save As',command=self.Save_File)        
        self.FileMenu.add_separator()
        self.FileMenu.add_command(label='Exit',command=self.client_exit)
        self.FileMenu.entryconfig('Save As',state='disabled') #disable before loading an image
        
        #Assign variables for checkbuttons
        self.RGB_var = tk.BooleanVar()
        self.RGB_var.set(True)
        self.HSV_var = tk.BooleanVar()
        self.LAB_var = tk.BooleanVar()        
        self.NDVI_var = tk.BooleanVar()            
        self.Heatmap_var = tk.BooleanVar()    
        
        #create Color View object
        #self.ViewMenu = tk.Menu(Menubar,tearoff=0)
        #Menubar.add_cascade(label='View',menu=self.ViewMenu)
        #Create submenu in Menu\View\        
        #self.ViewMenu.add_checkbutton(label='NDVI',command=self.View_NDVI,variable=self.NDVI_var)        
        #for i in range(1): self.ViewMenu.entryconfig(i,state='disabled') #disable before loading an image
   
        #create Layer object
        self.LayerMenu = tk.Menu(Menubar,tearoff=0)
        Menubar.add_cascade(label='Layer',menu=self.LayerMenu)
        self.LayerMenu.add_command(label='Add CSV',command=self.Layer_CSV)        
        self.LayerMenu.add_command(label='Add Shapeflie',command=self.Layer_SHP)        
        self.LayerMenu.add_command(label='Convert CSV->SHP',command=self.Proc_Convert_CSV2SHP)        
        self.LayerMenu.add_command(label='Convert ROI->SHP',command=self.Proc_Convert_ROI2SHP)        
        
        #create Filter object
        self.FilterMenu = tk.Menu(Menubar,tearoff=0)
        Menubar.add_cascade(label='Filter',menu=self.FilterMenu)
        self.FilterMenu.add_command(label='Erode',command=self.Filter_Erode)        
        self.FilterMenu.add_command(label='Dilate',command=self.Filter_Dilate)
        #Create submenu in Menu\Seg\Opening,Closing
        self.subOpeningMenu = tk.Menu(self.FilterMenu,tearoff=0)    
        self.FilterMenu.add_cascade(label='Opening',menu=self.subOpeningMenu)   
        self.subOpeningMenu.add_command(label='2x2',command=self.Filter_Opening)
        self.subOpeningMenu.add_command(label='5x5',command=self.Filter_Opening5)
        self.subClosingMenu = tk.Menu(self.FilterMenu,tearoff=0)    
        self.FilterMenu.add_cascade(label='Closing',menu=self.subClosingMenu)
        self.subClosingMenu.add_command(label='2x2',command=self.Filter_Closing)
        self.subClosingMenu.add_command(label='5x5',command=self.Filter_Closing5)   
        #self.FilterMenu.add_command(label='Band Math',command=self.BandMath)
        #self.FilterMenu.add_command(label='Threshold',command=self.Threshold) 
        self.FilterMenu.add_command(label='FFT',command=self.Filter_FFT)
        self.FilterMenu.add_command(label='Gradient',command=self.Filter_Gradient)
        self.FilterMenu.add_command(label='TopHat',command=self.Filter_TopHat)        
        self.FilterMenu.add_command(label='Mean',command=self.Filter_Mean)
        self.FilterMenu.add_command(label='Gaus',command=self.Filter_Gaus)
        self.FilterMenu.add_command(label='Otsu',command=self.Filter_Otsu)
        self.FilterMenu.add_command(label='GausFiltered',command=self.Filter_GausFiltered)        
        for i in range(6): self.FilterMenu.entryconfig(i,state='disabled') #disable before loading an image

        #create Processs object
        self.ProcMenu = tk.Menu(Menubar,tearoff=0)
        Menubar.add_cascade(label='Processs',menu=self.ProcMenu)
        self.ProcMenu.add_command(label='Calibrate',command=self.Proc_Calib)
        self.ProcMenu.add_command(label='Segment',command=self.Proc_Seg)        
        self.ProcMenu.add_command(label='Mask',command=self.Proc_LoadMask)        
        self.ProcMenu.add_command(label='Create NDVI',command=self.Proc_Create_NDVI)        
        for i in range(4): self.ProcMenu.entryconfig(i,state='disabled') #disable before loading an image
        
        #create Raster object
        self.RasterMenu = tk.Menu(Menubar,tearoff=0)
        Menubar.add_cascade(label='Raster',menu=self.RasterMenu)
        self.RasterMenu.add_command(label='Band Math',command=self.Raster_BandMath)        
        self.RasterMenu.add_command(label='Merge',command=self.Raster_Merge)
        for i in range(2): self.RasterMenu.entryconfig(i,state='disabled') #disable before loading an image
        
        #create Batch object
        self.BatchMenu = tk.Menu(Menubar,tearoff=0)
        Menubar.add_cascade(label='Batch',menu=self.BatchMenu)
        self.BatchMenu.add_command(label='Validate',command=self.Batch_Valid)
        self.BatchMenu.add_command(label='Calibrate',command=self.Batch_Calib)
        self.BatchMenu.add_command(label='Segment',command=self.Batch_Seg)
        self.BatchMenu.add_command(label='Avenger',command=self.Batch_Avg)
        self.BatchMenu.add_command(label='Pic2Video',command=self.Batch_Pic2Video)
        self.BatchMenu.add_command(label='FM Calendar',command=self.Batch_FmCalendar)        
        
        #create Setup object
        SetupMenu = tk.Menu(Menubar,tearoff=0)
        Menubar.add_cascade(label='Setup',menu=SetupMenu)
        SetupMenu.add_command(label='Read AM',command=self.Read_AM)   #Read csv to extract AM data time-matching
        SetupMenu.add_command(label='Load Config',command=self.Load_Config)
        SetupMenu.add_command(label='Save Config',command=self.Save_Config)
        SetupMenu.add_command(label='Load ROI',command=self.Load_ROI)
        SetupMenu.add_command(label='Save ROI',command=self.Save_ROI)
    
        #create Help object
        HelpMenu = tk.Menu(Menubar,tearoff=0)
        Menubar.add_cascade(label='Help',menu=HelpMenu)
        HelpMenu.add_command(label='About',command=About)
      
        #create Histogram graph
        self.init_histogram()

        #create buttons & sliders for segmentation   
        self.init_buttons_sliders()

        # create Text widgets (declair only once here due to the memory buildup)
        self.t = tk.Text(self.canvas.get_tk_widget(),relief=tk.FLAT)    # Text in histogram
        self.ts = tk.Text(self.canvas.get_tk_widget(),relief=tk.FLAT)   # Text in UpdateMouse_Image for statistics
        self.t0 = tk.Text(self.canvas.get_tk_widget(),relief=tk.FLAT)   # Text in UpdateMouse_Image for ROI
        self.t1 = tk.Text(self.canvas.get_tk_widget(),relief=tk.FLAT)   # Text in UpdateMouse_Image for RGB
        self.t2 = tk.Text(self.canvas.get_tk_widget(),relief=tk.FLAT)   # Text in UpdateMouse_Image for GPS
        self.text = tk.Text(self.canvas.get_tk_widget(),relief=tk.FLAT)   # Text in UpdateMouse_Image
        
        # suppress the creation of aux.xml file when using gdal library
        gdal.SetConfigOption("GDAL_PAM_ENABLED", "NO")        

    def __add_plot(self):
        pass
        # create a rectangle
        #self.frame.canvas.create_polygon(10, 10, 10, 150, 200, 150, 200, 10,  fill="gray", outline="black")
        #self.frame.canvas.create_oval(10, 10, 150, 150,  fill='#00ff00', outline="red")
    
    def Display_Image(self, im, seg=None):
        #print('im.shape',im.shape,im.dtype)
        
        # resize if zoomed
        if self.zf!=1 and self.flag_zm: 
            im = cv2.resize(im, (int(self.im.shape[1]*self.zf), int(self.im.shape[0]*self.zf)))        

        #global  imScn   # display image numpy (np.cv2) displayed on screen
                
        #if seg==None or seg=='add_layer': imScn = im    # to avoid histogram change, but causes seged px not updated
        imScn = im
        #print('imScn.dtype=',imScn.dtype,'imScn.shape=',imScn.shape)
            
        global img_photo     #must be here to clear img_photo memory buildup
        # convert imScn to 3b img
        if len(imScn.shape)>2 and imScn.shape[2]>3:
            img_photo = ImageTk.PhotoImage(Image.fromarray(cv2.cvtColor(imScn, cv2.COLOR_BGRA2BGR)))   # buffer (x3), processing time (1s for 166MB)
        else:
            if len(imScn.shape)>2 and imScn.shape[2]==1:
                imScn = imScn.reshape(imScn.shape[0],imScn.shape[1])   # (H,W,1)->(H,W) for ImageTk.PhotoImage()
            img_photo = ImageTk.PhotoImage(Image.fromarray(imScn))   # (file=self.src_file) buffer (x3), processing time (1s for 166MB)
        self.imScn = imScn
        
        #Display image
        #img_label = tk.Label(self.frame.canvas, image = img_photo).place(x=0,y=0) # buffer    
        self.frame.canvas.delete('all')    # clear the canvas, e.g., removing old ROI drawing
        self.frame.canvas.create_image(0,0,anchor="nw",image=img_photo) # buffer (x2), processing time (1s for 166MB)
        
        #Update the scrollbar length    
        self.frame.canvas.config(scrollregion=(0,0,int(self.imW*self.zf),int(self.imH*self.zf)))    
        
        #Update ROI if btn_ROI is selected
        if(self.btn_ROI_var == 1): 
            self.btn_ROI()    
            
        #update the histogram
        if self.flag_im and seg == None: #if not segmentation            
            #print('Image file size = %.1f MB'%(os.stat(self.src_file).st_size/1024/1024))
            self.update_Menu(self.ch)  
            #if os.stat(self.src_file).st_size/1024/1024 > 100:  # skip histogram if file size > 100MB
                #self.hist.clear()
                #self.canvas.draw()
            #else:
            self.Histogram() #do not update Histogram if doing segmentation    
        
        #update FileMenu
        self.FileMenu.entryconfig('Save As',state='normal') 
        
        #Read the mouse position in the image canvas
        self.frame.canvas.bind('<Motion>',self.MouseMotion_Image) 
        #Read mouse click start in the image canvas for Panning
        self.frame.canvas.bind("<ButtonPress-1>", self.MouseClickStart)
        #Read mouse click & drag in the image canvas for ROI drawing
        self.frame.canvas.bind('<B1-Motion>',self.MouseClick_Image)     
        #Read Ctrl + mouse click & drag in the image canvas for Panning
        self.frame.canvas.bind('<Control-B1-Motion>',self.CtrlMouseClick_Image)     
        #Read Shift + mouse click in the image canvas for oval ROI
        self.frame.canvas.bind('<Shift-B1-Motion>',self.ShiftMouseClick_Image)     
        #Read Alt + mouse click in the image canvas for curtain
        self.frame.canvas.bind('<Alt-B1-Motion>',self.AltMouseClick_Image)     
        #Read mouse middle key click in the image canvas for enclosed circle of ROI rotation
        self.frame.canvas.bind('<ButtonPress-2>',self.MouseBtn2Click_Image)     
        #Read the mouse unclick in the image canvas
        self.frame.canvas.bind('<ButtonRelease-1>',self.MouseUnclick_Image) 

    def Histogram(self):     
        # image selected for masking
        im = self.imScn_base    
        #print('im.dtype=',im.dtype,'len(im.shape)=',len(im.shape),'im.shape=',im.shape,'self.ch=',self.ch)
        #print('self.ch=',self.ch,'self.B=',self.B,'self.btn_cmap=',self.btn_cmap)
        
        #plt.hist(im.ravel(),256,[0,256]); plt.show()   #for 1-ch im        
        
        #Update the histogram and thresholding lines
        self.hist.clear()
        #print('max_y=',self.max_y,'self.scr_h-500=',self.scr_h-500)
        if (self.im_nband == 1 and self.ch!='Heatmap') or self.ch=='NDVI': self.btn_cmap = 'R'
        for i,col in enumerate(self.btn_cmap.lower()): # NOTE: .plot() takes color in lower case!
            if len(self.btn_cmap)==1: i = 'rgby'.find(self.btn_cmap.lower())    # detect 1b image & assign the selected channel

            #plot histogram graphs of the image
            self.hist.plot(cv2.calcHist([im],[i],None,[256],[1,256]),color=col) # 1: exclude saturated 0s
 
            #plot v-lines of thresholding sliders
            #print('col=',col,'max_y=',self.max_y,'self.bdry[%d]='%i,self.bdry[i][0],self.bdry[i][1],'self.sld.get()=',self.sldL[i].get(),self.sldU[i].get())
            if col=='HSV' or col=='LAB': col='RGB'
            self.hist.plot([self.bdry[i][0],self.bdry[i][0]],[0, self.max_y], color=col, linestyle='-', linewidth=.5)
            self.hist.plot([self.bdry[i][1],self.bdry[i][1]],[0, self.max_y], color=col, linestyle='-', linewidth=.5)
        self.canvas.draw()    
        
    def Img_Seg(self,flag_segStats=None,filterOpen=None,flag_batch=None,mask=None):     
        self.imH, self.imW = self.imScn_base.shape[:2]
        imH,imW = self.imH, self.imW
        ROI = self.ROI
                    
        # update ROIpx if changed
        #print('len(self.ROI)=',len(self.ROI),'len(self.ROIpx)=',len(self.ROIpx))
        #print('pre_ROI0=',self.pre_ROI0,'ROI[0]=',ROI[0])
        if self.ROI_rot==0: 
            self.GRID_CntIn = (ROI[-1][2]-ROI[0][0]) * (ROI[-1][3]-ROI[0][1])
        if self.ROI_rot!=0 and (len(self.ROI) != len(self.ROIpx) or self.pre_ROI0 != ROI[0]): 
            self.ROIpx = [[] for i in range(len(ROI))] # pixels (i,j) in tROI
            self.GRID_CntIn = 0
            for i in range (len(ROI)):
                self.ROIpx[i] = Extract_tROIpx(ROI[i])   # 3D array: [n-th ROI of ROIs], [n-th px in ROI], [i or j value] 
                #print('i=',i,'len(self.ROIpx[i]=',len(self.ROIpx[i]))
                self.GRID_CntIn = self.GRID_CntIn + len(self.ROIpx[i])    # total # of pixels in GRID            
            #print('self.GRID_CntIn=',self.GRID_CntIn)
            self.pre_ROI0 = ROI[0]            
        #if len(ROIpx)>1:
            #print('ROIpx[1]=',ROIpx[1])
            #print('ROIpx[1][2]=',ROIpx[1][2])
        
        # set the default status of buttons
        self.MASK_var.set(0)
        self.SEG_var.set(1)        
        
        # image selected for masking
        if len(self.btn_cmap)==1 and self.im_nband>1: # 1b image selected from 3b or 4b image
            #print('rgby'.find(self.btn_cmap.lower()),'self.im_nband=',self.im_nband,'self.im_stretched.shape=',self.im_stretched.shape)   
            if self.ch=='RGBN' or self.ch=='RGB' or self.ch=='HSV' or self.ch=='LAB': 
                im = self.imScn_base[:,:,'rgby'.find(self.btn_cmap.lower())]
                #im = self.im[:,:,'rgby'.find(self.btn_cmap.lower())]
            else: 
                im = self.imScn_base  # NDVI, RGBN
                #im = self.im  # NDVI, RGBN
        else:          
            im = self.imScn_base
            #im = self.im
        
        #cv2.imshow('im',im), cv2.waitKey(0)
        if len(im.shape)>2: nband = im.shape[2]
        else:               nband = 1
        #print ('im.dtype=',im.dtype,'im.shape=',im.shape,'self.ch=',self.ch,'len(ROI)=',len(ROI))
        #print ('im (min,max)=',np.nanmin(self.im),np.nanmax(self.im))
        
        # define thresholding values
        if nband>1: #if color image         
            if nband==3: #if 3b image         
                loB = np.array([self.sldL[0].get(),self.sldL[1].get(),self.sldL[2].get()])
                upB = np.array([self.sldU[0].get(),self.sldU[1].get(),self.sldU[2].get()])   
            elif nband>3: #if 4b image         
                loB = np.array([self.sldL[0].get(),self.sldL[1].get(),self.sldL[2].get(),self.sldL[3].get()])
                upB = np.array([self.sldU[0].get(),self.sldU[1].get(),self.sldU[2].get(),self.sldU[3].get()])   
            #else: messagebox.showwarning(title='Warning',message='The image is +4 bands!')
        else:               #if 1ch image
            loB = np.array([self.sldL['rgby'.find(self.btn_cmap.lower())].get()])
            upB = np.array([self.sldU['rgby'.find(self.btn_cmap.lower())].get()])
        #print('loB=',loB,'upB=',upB,'self.ch=',self.ch,'nband=',nband)        
                
        # find 4-side boundary of standing ROI
        (X1,Y1),(X2,Y2) = (ROI[0][:2], ROI[len(ROI)-1][2:4])    # 2-corner boundary of tROI        
        
        # calculate 4-side boundary of tROI if rotated ROI
        tX1,tY1,tX2,tY2 = X1,Y1,X2,Y2
        if self.ROI_rot!=0: 
            mask_tROI, OuterBoundary, InnerBoundary = Calculate_GRID_boundary(self.ROIpx,imW,imH)
            tX1,tY1, tX2,tY2 = OuterBoundary
            minx,miny,maxx,maxy = InnerBoundary
            #print('OuterBoundary=',OuterBoundary)
            #print('minx,miny,maxx,maxy=',minx,miny,maxx,maxy)
            GRID_CntTotal= (tX2-tX1)*(tY2-tY1)
        else: 
            GRID_CntTotal= self.GRID_CntIn
        
        GRID_CntOut= GRID_CntTotal-self.GRID_CntIn            
        #print('GRID_CntTotal=',GRID_CntTotal,'self.GRID_CntIn=',self.GRID_CntIn,'GRID_CntOut=',GRID_CntOut)
        #print('X1,Y1,X2,Y2=',X1,Y1,X2,Y2,'tX1,tY1,tX2,tY2=',tX1,tY1,tX2,tY2)          
        self.tX1, self.tY1, self.tX2, self.tY2 = tX1, tY1, tX2, tY2
                      
        # create grid_mask (grid size) & im_mask (image size)
        if self.ROI_var.get()==1:  
            # create a mask image segmenting px in grid 
            if type(mask)!=type(None): 
                if self.ROI_rot == 0: grid_mask = mask[Y1:Y2,X1:X2]
                else:                 
                    im_tROI = np.copy(im[tY1:tY2,tX1:tX2])
                    im_tROI[~mask_tROI[tY1:tY2,tX1:tX2]] = loB   # create a grid-masked im w tROI in FG & else (True px) in BG to threshold, [loB[0],loB[1],loB[2]]
                    #cv2.imshow('im_tROI',im_tROI), cv2.waitKey(0)    
                    grid_mask = cv2.inRange(im_tROI,loB,upB) # crop a grid-FG image seged by threshold, [0:rows,0:cols]                           
                    del im_tROI                                        
            else:
                #print ('len(ROI)=',len(ROI),'ROI[0]=',ROI[0])
                if ROI[0][-1] == 'c':                  
                    r1 = (X2-X1)/2.
                    r2 = (Y2-Y1)/2.
                    xc,yc = X1+int(r1),Y1+int(r2)    # center (xc, yc) of the oval
    
                    yy, xx = np.ogrid[-yc:imH-yc, -xc:imW-xc]  # a range of xy-coord with center shifted to (0,0), xx=[[0],...,[n]], yy=[[0,...,n]] 
                    #print('len(xx[0]),len(yy)=',len(xx[0]),len(yy))
                    mask_oval = xx*xx/(r1*r1) + yy*yy/(r2*r2) <= 1   # 2D array of mask image in True (in grid) or False (outside grid)
                    #print('len(mask_oval)=',len(mask_oval),'len(mask_oval[0])=',len(mask_oval[0]))
                    im_oval = np.copy(im)
                    im_oval[~mask_oval] = loB   # create a grid-masked im w oval in FG & else (True px) in BG to threshold, [loB[0],loB[1],loB[2]]
                    #cv2.imshow('im_oval',im_oval), cv2.waitKey(0)    
                    #from scipy.ndimage.filters import generic_filter as gf
                    #circular_mean = gf(data, np.mean, footprint=im_oval)

                    grid_mask = cv2.inRange(im_oval[Y1:Y2,X1:X2],loB,upB) # crop a grid-FG image seged by threshold, [0:rows,0:cols] 
                    print('grid_mask=',grid_mask.shape,grid_mask.dtype)
                    #cv2.imshow('grid_mask',grid_mask), cv2.waitKey(0)       
    
                    #for ROI metrics calculation
                    im_oval[~mask_oval] = 0   #[loB[0],loB[1],loB[2]]
                    grid_mask0 = cv2.inRange(im_oval[Y1:Y2,X1:X2],loB,upB) # create a grid-masked im w tROI in FG & else (True px) in BG to 0 for multi-GRIDs
                    del im_oval
                    #print('grid_mask0=',grid_mask0.shape,grid_mask0.dtype)
                    #cv2.imshow('grid_mask0',grid_mask0), cv2.waitKey(0)       
                else:     
                    if self.ROI_rot==0 or self.ROI_rot==None:
                        grid_mask = cv2.inRange(im[Y1:Y2,X1:X2],loB,upB) #im[0:rows,0:cols], masked binary ROI image in black bg & white fg           
                    else:
                        im_tROI = np.copy(im[tY1:tY2,tX1:tX2])
                        im_tROI[~mask_tROI[tY1:tY2,tX1:tX2]] = loB   # create a grid-masked im w tROI in FG & else (True px) in BG to threshold, [loB[0],loB[1],loB[2]]
                        #cv2.imshow('im_tROI',im_tROI), cv2.waitKey(0)    
                        grid_mask = cv2.inRange(im_tROI,loB,upB) # crop a grid-FG image seged by threshold, [0:rows,0:cols]                           
                        del im_tROI
            
            #Opening (erode+dilate) to remove noise blobs
            if filterOpen!=None: grid_mask = cv2.morphologyEx(grid_mask,cv2.MORPH_OPEN,np.ones((filterOpen,filterOpen)))            

            #cv2.imshow('0grid_mask',grid_mask), cv2.waitKey(0)        
            grid_mask.shape = (grid_mask.shape[0],grid_mask.shape[1],1)                        
            #print('grid_mask=',grid_mask.shape, 'self.GRID=',self.GRID)

            # create a blank im_mask with insert of the grid-seged image
            im_mask = np.invert(np.zeros((imH,imW,1),dtype=np.uint8))    #height(row) x width(col), white image            
            # insert the grid_mask image into im_mask
            im_mask[tY1:tY2,tX1:tX2] = grid_mask
            #cv2.imshow('im_mask',im_mask), cv2.waitKey(0)        
        else:             
            if type(mask)!=type(None): im_mask = mask
            else: 
                #print('loB,upB=',loB,upB)
                im_mask= cv2.inRange(im,loB,upB) 
                    
            #Opening (erode+dilate) to remove noise blobs
            if filterOpen!=None: im_mask = cv2.morphologyEx(im_mask,cv2.MORPH_OPEN,np.ones((filterOpen,filterOpen)))
        #cv2.imshow('im_mask',im_mask), cv2.waitKey(0)

        #print('im.shape=',im.shape,'im_mask.shape=',im_mask.shape)
        
        # Display the masked image
        #print('im.shape=',im.shape,'im_mask.shape=',im_mask.shape)    
        im_masked = cv2.bitwise_and(im,im,mask=im_mask) # RGBN, buffer (x2), time (2s for 166MB)        
        #cv2.imshow('im_masked',im_masked), cv2.waitKey(0)
        #if flag_batch==None: # skip if batch
        self.Display_Image(im_masked,seg=True)   # processing time: 3s for 166MB img        
        #else: self.imScn_base = im_mask   # return to Batch_Seg for heatmap masking
        #self.seg = False   #reset to default when seg is done

        # Pixel calculations from the raw image 
        GRID_CntP = 0
        GRID_Cnt = 0
        GRID_ave = []
        GRID_std = []
        GRID_HSV_ave = []
        ROI_Cnt = []
        ROI_CntP = []
        ROI_CntIn = [[] for i in range(len(ROI))]
        ROI_ave = []
        ROI_std = []
        ROI_ndvi = []
        ROI_HSV_ave = []
        if flag_segStats:  
            self.flag_segStats = 1
            self.SegStats_var.set(1)
            ## calcuate image metrics using the original image, self.im
            #dtype = str(self.im.dtype)
            #bitDepth = int(dtype[dtype.find('t')+1:])
            #print('self.im.dtype=',self.im.dtype,bitDepth)
            if self.rot==0:im = self.im    # use the raw image (self.im)
            else: 
                im = imu.rotate_bound(self.im,self.rot) # use the raw rotated image, not self.imScn_base
            
            #if bitDepth==8: # if not 8-bit image, use imScn_base stretched from 12-\16-bit image
                #del im
                #im = self.im
            
            if self.ROI_var.get()==1:  # calculate ROIs stats
                del im_masked
                
                # calculate a GRID stats (as a outer ROI)
                self.GRID_H, self.GRID_W = grid_mask.shape[0:2]
                if self.ROI_rot == 0: im_GRID= im[Y1:Y2,X1:X2] # croped 2-corner grid image, im[0:rows,0:cols] 
                else:                 
                    im_GRID= np.copy(im[tY1-1:tY2-1,tX1-1:tX2-1]) # croped 4-corner grid image, im[0:rows,0:cols] 
                    # change 4-piece BG triangles of grid to zero                
                    im_GRID[~mask_tROI[tY1-1:tY2-1,tX1-1:tX2-1]] = 0
                    #cv2.imshow('im_GRID',im_GRID)
                #cv2.imshow('grid_mask',grid_mask), cv2.imshow('im_GRID',im_GRID)
                im_GRID_masked = cv2.bitwise_and(im_GRID,im_GRID,mask=grid_mask) #roi image including masked pixels in 0 (black)                    
                #print('im_GRID_masked=',im_GRID_masked.shape,im_GRID_masked.dtype)
                if len(im_GRID_masked.shape)==2: im_GRID_masked = im_GRID_masked.reshape(im_GRID_masked.shape[0],im_GRID_masked.shape[1],1) # 1b img shape:(H,W)->(H,W,1) for im[:,:,i], ->(H,W) for ImageTk.PhotoImage(), Save_GeoTiff()                
                #cv2.imshow('grid_mask',grid_mask), cv2.imshow('im_GRID',im_GRID), cv2.imshow('im_GRID_masked',im_GRID_masked), cv2.waitKey(0)
                #cv2.imshow('im_GRID_masked',im_GRID_masked), cv2.waitKey(0)
                del im_GRID                                   
                for i in range(nband):  
                    ret = im_GRID_masked[:,:,i][im_GRID_masked[:,:,i]!=0]
                    #print('i=',i,'ret=',ret.shape,ret)
                    if len(ret)==0: ret = [0]
                    GRID_ave.append(np.nanmean(ret))
                    GRID_std.append(np.nanstd(ret))
                    #print('i=',i,'GRID_ave=',GRID_ave,'np.mean(ret)=',np.mean(ret))
                    
                    #if np.all(np.all(im_GRID_masked[:,:,i]!=im_GRID_masked[:,:,i]):    # detect nan, nan never equal to itself
                    ret = im_GRID_masked[:,:,i][im_GRID_masked[:,:,i]!=np.nan]
                    if len(ret)==0: ret = [0]                                        
                    if np.isnan(GRID_ave[i]):
                        GRID_ave[i] = np.nanmean(ret)# filter nan
                    if np.isnan(GRID_std[i]): 
                        GRID_std[i] = np.nanstd(ret) # filter nan
                    #print('i=',i,'GRID_ave[i]=',GRID_ave[i],'np.nanmean(ret)=',np.nanmean(ret),'np.mean(ret)=',np.mean(ret))
                                        
                    if flag_batch and i<3 and nband>=3: # 3b HSV image
                        im_GRID_HSV_masked = cv2.cvtColor(im_GRID_masked[:,:,:3],cv2.COLOR_RGB2HSV_FULL) #baseimage of imScn before segmentation
                        ret = im_GRID_HSV_masked[:,:,i][im_GRID_HSV_masked[:,:,i]!=0]
                        if len(ret)==0: ret = [0]                                            
                        GRID_HSV_ave.append(np.nanmean(ret))
                        del im_GRID_HSV_masked
                #GRID_Cnt = np.count_nonzero(im_GRID_masked[:,:,0])
                GRID_Cnt = np.count_nonzero(grid_mask) -  GRID_CntOut
                #print('GRID_Cnt',GRID_Cnt,'np.count_nonzero(grid_mask)=',np.count_nonzero(grid_mask),'self.GRID_CntIn=',self.GRID_CntIn,'GRID_CntOut=',GRID_CntOut)
                #print('tX2,tX1,tY2,tY1=',tX2,tX1,tY2,tY1,'(tX2-tX1)=',(tX2-tX1),'(tY2-tY1)=',(tY2-tY1),'(tX2-tX1)*(tY2-tY1)=',(tX2-tX1)*(tY2-tY1))                #if nband==4:    # 4b image, RGBN
                    #self.GRID_ndvi,ndvi_min,ndvi_max = Calculate_NDVI(im_GRID_masked[:,:,3],im_GRID_masked[:,:,0])
                    #print('self.GRID_ndvi=',self.GRID_ndvi)
                del im_GRID_masked
                #print('GRID_ave=',GRID_ave,'GRID_std=',GRID_std)
                
                # calculate ROI stats (as inner subROIs)
                #print('# of ROI=',len(ROI))                                
                for r in range(len(ROI)):    
                    # display status bars of processing
                    if len(ROI)>=10 and r== 0: print (len(ROI),end=" ROIs:")                
                    if len(ROI)>=10 and r%(int(len(ROI)/10)) == 0: print("-",end="")
                    if len(ROI)>=10 and r== len(ROI)-1: print (' ')
                
                    #create roi_mask of subROI (im16_ROI) to be masked       
                    im16_ROI = 0               
                    ROI_CntIn[r], ROI_CntOut = 0,0
                    if self.ROI_var.get()==1:   
                        if self.ROI_rot == 0: 
                            roi_mask = grid_mask[ROI[r][1]-Y1:ROI[r][3]-Y1,ROI[r][0]-X1:ROI[r][2]-X1]
                            im_ROI= im[ROI[r][1]:ROI[r][3],ROI[r][0]:ROI[r][2]] #im[0:rows,0:cols] 
                            ROI_CntTotal =  (ROI[r][3]-ROI[r][1])*(ROI[r][2]-ROI[r][0])
                            ROI_CntIn[r] = ROI_CntTotal
                        else:
                            # create a mask image by assgining tROIpx-FG coord to mask-FG
                            yy, xx = np.ogrid[:maxy[r]-miny[r]+1, :maxx[r]-minx[r]+1]  # a range of xy-coord of grid shifted to (0,0), xx=[[0],...,[n]], yy=[[0,...,n]] 
                            mask_subtROI = xx+yy<0 # create a subtROI-sized matrix with all False, [j][i]                             
                            for j in range(len(self.ROIpx[r])):                             
                                #print('r,j=',r,j,self.ROIpx[r][j][0],self.ROIpx[r][j][1],mask_subtROI[self.ROIpx[r][j][0]][self.ROIpx[r][j][1]])
                                mask_subtROI[self.ROIpx[r][j][1]-miny[r]][self.ROIpx[r][j][0]-minx[r]] = True  # assign True to FG px, [j][i]
                                if mask_subtROI[self.ROIpx[r][j][1]-miny[r]][self.ROIpx[r][j][0]-minx[r]] == True:
                                    ROI_CntIn[r] = ROI_CntIn[r] + 1
                            ROI_CntTotal = (maxx[r]-minx[r]+1)*(maxy[r]-miny[r]+1)                            
                            #print('i=',i,'ROI_CntIn[r]=',ROI_CntIn[r],'ROI_CntTotal=',ROI_CntTotal,'ROI_CntOut=',ROI_CntOut)
                            
                            #print('minx[r],miny[r],maxx[r],maxy[r]=',minx[r],miny[r],maxx[r],maxy[r])                            
                            im_subtROI = np.copy(im[miny[r]:maxy[r]+1,minx[r]:maxx[r]+1]) 
                            im_subtROI[~mask_subtROI] = loB   # create a grid-masked im w tROI in FG & else (True px) in BG to threshold, [loB[0],loB[1],loB[2]]
                            #print('r=',r,'len(im_subtROI)=',len(im_subtROI),'len(mask_subtROI)=',len(mask_subtROI),'len(mask_subtROI[0])=',len(mask_subtROI[0]))
                            roi_mask = cv2.inRange(im_subtROI,loB,upB) # crop a grid-FG image seged by threshold, [0:rows,0:cols]                           
                            #roi_mask = cv2.inRange(im_subtROI[miny[r]:maxy[r],minx[r]:maxx[r]],loB,upB) # crop a grid-FG image seged by threshold, [0:rows,0:cols]                           
                            #cv2.imshow('im_subtROI',im_subtROI), cv2.waitKey(0)  
                            del im_subtROI                   
                            im_ROI= np.copy(im[miny[r]:maxy[r]+1,minx[r]:maxx[r]+1]) #im[0:rows,0:cols]                             
                            # change 4-piece BG triangles of ROI to zero                
                            im_ROI[~mask_subtROI] = 0           
                        ROI_CntOut = ROI_CntTotal - ROI_CntIn[r]
                        if ROI[r][-1] == 'c': 
                            roi_mask = grid_mask0
                            del grid_mask0        
                    else:   # need to set 1x1 grid, then unclick ROI 
                        im_ROI= im[:imH,:imW] #im[0:rows,0:cols] 
                        roi_mask = im_mask
                    #print('im_ROI=',im_ROI.shape,im_ROI.dtype,', roi_mask=',roi_mask.shape,roi_mask.dtype)
                    #cv2.imshow('grid_mask',grid_mask), cv2.waitKey(0)   
                    #cv2.imshow('roi_mask',roi_mask), cv2.waitKey(0)   
                    #cv2.imshow('im_ROI',im_ROI), cv2.waitKey(0)   
                    
                    #Create a seged subROI image
                    im_ROI_masked = cv2.bitwise_and(im_ROI,im_ROI,mask=roi_mask) #roi image including masked pixels in 0 (black)            
                    #cv2.imshow('im_ROI_masked',im_ROI_masked), cv2.waitKey(0)   
                    if len(im_ROI_masked.shape)==2: im_ROI_masked = im_ROI_masked.reshape(im_ROI_masked.shape[0],im_ROI_masked.shape[1],1) # 1b img shape:(H,W)->(H,W,1) for im[:,:,i], ->(H,W) for ImageTk.PhotoImage(), Save_GeoTiff()
                    #print('im_ROI_masked.shape=',im_ROI_masked.shape,im_RI_masked.dtype)
                    del im_ROI
    
                    #create ROI_Cnt & ROI_ave array to later calculate Ave, band math
                    ROI_Cnt.append(np.count_nonzero(roi_mask)-ROI_CntOut)
                    #print('ROI_Cnt[r]=',ROI_Cnt[r],'np.count_nonzero(roi_mask)=',np.count_nonzero(roi_mask),'ROI_CntOut=',ROI_CntOut)
                    ROI_CntP.append(100*ROI_Cnt[r]/ROI_CntIn[r])
                    #print('r=',r,'ROI_CntP[r]=',ROI_CntP[r],'ROI_Cnt[r]=',ROI_Cnt[r],'ROI_CntIn[r]=',ROI_CntIn[r])
                                        
                    #ROI_Cnt.append(np.count_nonzero(im_ROI_masked[:,:,0]))
                    del roi_mask
                                        
                    ROI_ave_ch = []
                    ROI_std_ch = []
                    ROI_HSV_ave_ch = []
                    for i in range(nband):                         
                        ret = im_ROI_masked[:,:,i][im_ROI_masked[:,:,i]!=0]                        
                        if len(ret)==0: ret = [0]
                        ROI_ave_ch.append(np.nanmean(ret))    #px data of roi image excluding masked pixels in 0 (black)
                        ROI_std_ch.append(np.nanstd(ret))    #px data of roi image excluding masked pixels in 0 (black)
                        if flag_batch and i<3 and nband>=3: # 3b HSV image: 
                            im_ROI_HSV_masked = cv2.cvtColor(im_ROI_masked,cv2.COLOR_RGB2HSV_FULL) #baseimage of imScn before segmentation
                            ret = im_ROI_HSV_masked[:,:,i][im_ROI_HSV_masked[:,:,i]!=0]
                            if len(ret)==0: ret = [0]                            
                            ROI_HSV_ave_ch.append(np.nanmean(ret))    #px data of roi image excluding masked pixels in 0 (black)
                            del im_ROI_HSV_masked                        
                        # calculate NDVI if 4b image
                        if i==3:
                            ndvi,ndvi_min,ndvi_max = Calculate_NDVI(im_ROI_masked[:,:,3],im_ROI_masked[:,:,0]) 
                            ROI_ndvi.append(ndvi)                            
                            #print('ndvi=%.3f (%.3f~%.3f)'%(ROI_ndvi[r],ndvi_min,ndvi_max),end=',')
                    ROI_ave.append(ROI_ave_ch)
                    ROI_std.append(ROI_std_ch)
                    ROI_HSV_ave.append(ROI_HSV_ave_ch)
                    #print('self.ROI_Cnt=',ROI_Cnt[r],'ROI_ave=',ROI_ave[r],'ROI_std=',ROI_std[r])
                    del im_ROI_masked
            else: # if undefined ROI, GRID as a whole image
                self.GRID_H, self.GRID_W = im.shape[0:2]
                print('im_masked.shape=',im_masked.shape,'im.shape=',im.shape)
                if len(im_masked.shape)==2: im_masked = im_masked.reshape(im.shape[0],im.shape[1],1) # 1b img shape:(H,W)->(H,W,1) for im[:,:,i], ->(H,W) for ImageTk.PhotoImage(), Save_GeoTiff()
                for i in range(nband):    
                    ret = im_masked[:,:,i][im_masked[:,:,i]!=0]
                    if len(ret)==0: ret = [0]                        
                    GRID_ave.append(np.nanmean(ret))
                    GRID_std.append(np.nanstd(ret))
                    if flag_batch and i<3 and nband>=3: # 3b HSV image: 
                        im_HSV_masked = cv2.cvtColor(im_masked,cv2.COLOR_RGB2HSV_FULL) #baseimage of imScn before segmentation
                        ret = im_HSV_masked[:,:,i][im_HSV_masked[:,:,i]!=0]
                        if len(ret)==0: ret = [0]                          
                        GRID_HSV_ave.append(np.nanmean(ret))
                        del im_HSV_masked                    
                    #print('GRID_ave=%.2f'%GRID_ave[i],'GRID_std=%.2f'%GRID_std[i],end=', ')
                    #if i==nband-1: print('')                                    
                    # calculate NDVI if 4b image
                    if i==3:
                        ndvi,ndvi_min,ndvi_max = Calculate_NDVI(im_masked[:,:,3],im_masked[:,:,0]) 
                        ROI_ndvi.append(ndvi)    # ROI=entire image
                        #print('ndvi=%.3f (%.3f~%.3f)'%(ROI_ndvi[0],ndvi_min,ndvi_max))                                            
                GRID_Cnt = np.count_nonzero(im_masked[:,:,0])
                #print(np.nanmean(im_masked[:,:,0][im_masked[:,:,0]!=0]))    # = im[np.nonzero(im)]
                #cv2.imshow('im_masked',im_masked), cv2.waitKey(0)                
                ROI_Cnt.append(GRID_Cnt)
                ROI_ave.append(GRID_ave)
                ROI_std.append(GRID_std)
                if flag_batch: ROI_HSV_ave.append(GRID_HSV_ave)
                del im_masked            
            GRID_CntP = 100*GRID_Cnt/self.GRID_CntIn
            print('GRID_CntP=%.2f%% (%d/%d)'%(GRID_CntP,GRID_Cnt,self.GRID_CntIn))
            #print ('len(ROI_ave)=%d'%(len(ROI_ave)), end=', Ref:')
            #for i in range(self.im_nband): print('%.2f%% '%(GRID_ave[i]*100/255), end='')
            #print('')
            
        #del grid_mask
        del im_mask, im
        self.GRID_CntP = GRID_CntP
        self.GRID_Cnt = GRID_Cnt
        self.GRID_ave = GRID_ave
        self.GRID_std = GRID_std
        self.GRID_HSV_ave = GRID_HSV_ave
        self.ROI_Cnt = ROI_Cnt
        self.ROI_CntP = ROI_CntP
        self.ROI_CntIn = ROI_CntIn
        self.ROI_ave = ROI_ave
        self.ROI_std = ROI_std
        self.ROI_ndvi = ROI_ndvi
        self.ROI_HSV_ave = ROI_HSV_ave
        (self.X1,self.Y1),(self.X2,self.Y2) = (X1,Y1),(X2,Y2)
        return 
        
    #Read the mouse position inside the image canvas
    def MouseMotion_Image(self, event):     
        self.UpdateMouse_Image(event,'MouseMotion',self.zf)
        return 

    #Read mouse click in the image canvas
    def MouseClickStart(self, event):
        self.frame.canvas.scan_mark(event.x, event.y)
        
    #Read mouse click inside the image canvas & display a rectangular ROI
    def MouseClick_Image(self, event):     
        #print('self.btn_ROI_var=',self.btn_ROI_var,'self.ROI_var.get()=',self.ROI_var.get())
        if(self.btn_ROI_var == 1):
            self.Reset_ROI(event)
            self.ROI[0][-1]='r'
            self.GRID[0] = int(self.entR.get())
            self.GRID[1] = int(self.entC.get())            
            self.ROI = Draw_ROI(self.GRID, self.ROI, self.zf, self.frame.canvas, self.ROI_rot) 
        return

    #Read Ctrl + mouse click&drage inside the image canvas for panning
    def CtrlMouseClick_Image(self, event):     
        # Drag (move) canvas to the new position
        self.frame.canvas.scan_dragto(event.x, event.y, gain=1)
        return

    #Read Shift + mouse click inside the image canvas & display an oval ROI
    def ShiftMouseClick_Image(self, event):     
        if(self.btn_ROI_var == 1):
            self.Reset_ROI(event)
            self.ROI[0][-1]='c'
            self.GRID[0] = int(self.entR.get())
            self.GRID[1] = int(self.entC.get())
            self.ROI = Draw_ROI(self.GRID, self.ROI, self.zf, self.frame.canvas, self.ROI_rot) 
        return

    #Read Alt + mouse click&drage inside the image canvas
    def AltMouseClick_Image(self, event):     
        self.Image_Curtain(event,'MouseClick',self.zf)
        return
    
    #Draw ROI
    def Reset_ROI(self, event):     
        self.UpdateMouse_Image(event,'MouseClick',self.zf)        
        #self.frame.canvas.delete('all')
        #self.frame.canvas.create_image(0,0,anchor="nw",image=img_photo)        
        self.ROI[0] = [int(self.x1/self.zf),int(self.y1/self.zf),int(self.x2/self.zf),int(self.y2/self.zf),self.ROI[0][-1]]              
        #print('ROI = [%d,%d]-[%d,%d]'%(self.ROI[0][0],self.ROI[0][1],self.ROI[0][2],self.ROI[0][3]))
        # recreate ROI with GRID created
        self.ROI = Update_ROIs(self.GRID,[self.ROI[0]],self.ROI_rot)            
        return

    #Read the Mounse buttone-2 (middle) key inside the image canvas to draw an enclosed circle for ROI rotation
    def MouseBtn2Click_Image(self, event):  
        self.MouseBtn2 = not self.MouseBtn2
        
        if(self.btn_ROI_var == 1):
            theta_increment = 0.1           
            self.MouseWheel_ROIrot(event,theta_increment)
            
            ## update ROIpx if changed
            #if len(self.ROI) != len(self.ROIpx) or self.pre_ROI0 != self.ROI:
                #self.ROIpx = [[] for i in range(len(self.ROI))] # pixels (i,j) in tROI
                #for i in range (len(self.ROI)):
                    #self.ROIpx[i] = Extract_tROIpx(self.ROI[i])   # 3D array: [n-th ROI of ROIs], [n-th px in ROI], [i or j value]                 
                #self.pre_ROI0 = self.ROI[0]
        return
    
    #Read the mouse wheel inside the image canvas
    def MouseUnclick_Image(self, event):  
        if(self.btn_ROI_var == 1):
            ##save the changed ROI to Config file                            
            #self.Save_Config('filename')        
            #print('ROI=',self.ROI[0], end=' ')            
            
            # update ROIpx if changed
            #if len(self.ROI) != len(self.ROIpx) or self.pre_ROI0 != self.ROI:
                #self.ROIpx = [[] for i in range(len(self.ROI))] # pixels (i,j) in tROI
                #for i in range (len(self.ROI)):
                    #self.ROIpx[i] = Extract_tROIpx(self.ROI[i])   # 3D array: [n-th ROI of ROIs], [n-th px in ROI], [i or j value]                    
                #self.pre_ROI0 = self.ROI[0]
                
            # update segmentation in ROI
            #if self.SEG_var.get()==1: self.btn_SEG()            
            try:                
                print('ent=',self.ent_,'vars_=',self.vars_)            
            except:
                #print('Reference target was not selected yet.')
                pass
        return
    
    #Read mouse wheel inside the grid entry for GridR & update the value of GridR
    def MouseWheel_entR(self, event):             
        #update the entry when a manual change
        self.GRID[0] = int(self.entR.get())            
        self.GRID[0] += 1*int(event.delta/120)
        #print(self.GRID[0])        
        #limit the grid range 1~999
        if(self.GRID[0]<1): self.GRID[0] = 1                
        elif(self.GRID[0]>999): self.GRID[0] = 999  
        self.entR.delete(0,tk.END)
        self.entR.insert(0,self.GRID[0])
        if(self.ROI_var.get() == 1): 
            self.frame.canvas.delete('all')
            self.frame.canvas.create_image(0,0,anchor="nw",image=img_photo)            
            self.btn_ROI()                 
        return

    #Read mouse wheel inside the grid entry for GridC & update the value of GridC
    def MouseWheel_entC(self, event):             
        #update the entry when a manual change
        self.GRID[1] = int(self.entC.get())            
        self.GRID[1] += 1*int(event.delta/120)
        #print(self.GRID[1])        
        #limit the grid range 1~999
        if(self.GRID[1]<1): self.GRID[1] = 1
        elif(self.GRID[1]>999): self.GRID[1] = 999
        self.entC.delete(0,tk.END)
        self.entC.insert(0,self.GRID[1])
        if(self.ROI_var.get() == 1): 
            self.frame.canvas.delete('all')
            self.frame.canvas.create_image(0,0,anchor="nw",image=img_photo)            
            self.btn_ROI()                 
        return
    
    #Zoom the image 
    def Img_Zoom(self):        
        #Display the zoomed image
        im = self.imScn_base
        
        # detect 3b (-1) or 1b image (0,1,2)
        #print('combobox=',self.ch,'Band=',self.B)        
        if (self.B>=0 and (self.ch=='RGB' or self.ch=='HSV' or self.ch=='LAB')): im=im[:,:,self.B]   #012:RGB        if self.SEG_var.get()==1: self.Img_Seg()
        if self.MASK_var.get()==1: self.btn_MASK()
        if self.rot!=0: self.btn_ROTATE()
        if self.MASK_var.get()!=1 and self.rot==0:  #self.SEG_var.get()!=1 and 
            self.Display_Image(im)       
        return    

    #Rotate the image 
    def Img_Rotate(self):        
        # set the default status of buttons
        self.MASK_var.set(0)
        self.SEG_var.set(0)
        
        #Display the rotated image
        im_rot = imu.rotate_bound(self.im_stretched,self.rot)
        
        if self.rot == 0: 
            self.imScn_base = self.im_stretched
        else: 
            self.imScn_base = im_rot
        print('self.im.shape=',self.im.shape,'im_rot.shape=',im_rot.shape)
        self.Display_Image(im_rot)        
        return    

    #Read mouse wheel inside the zoom entry for zoom factor & update the image zoomed
    def MouseWheel_entZm(self, event):  
        #update the entry when a manual change
        self.MouseWheelCnt_Zm = round(self.MouseWheelCnt_Zm) + int(1*(event.delta/120))
        #adjust MouseWheelCnt 0->1/2, -1->1/3
        if(self.MouseWheelCnt_Zm>=1): self.zf = self.MouseWheelCnt_Zm    #zoom factor of multiplier
        else: self.zf = -1/(self.MouseWheelCnt_Zm-2)    #-2 to skip the mousewheel 0,-1
        #limit the zoom range 
        zm_lim = 20 #within +\- 10x
        if(self.zf<1/zm_lim): 
            self.zf = 1/zm_lim   
            self.MouseWheelCnt_Zm = -8
        elif(self.zf>zm_lim): 
            self.zf = zm_lim 
            self.MouseWheelCnt_Zm = zm_lim
        print('delta=',event.delta,'self.MouseWheelCnt_Zm=',self.MouseWheelCnt_Zm,'self.zf=',self.zf)
        #update the zoom entry
        self.entZm.delete(0,tk.END)
        self.entZm.insert(0,"%.3f"%self.zf)
        self.Img_Zoom() #Display the zoomed image
        return

    #Read Ctrl + mouse wheel inside the image canvas to zoom the image
    def CtrlMouseWheel_Image(self, event): 
        self.ctrl_btn = 1    
        self.MouseWheel_entZm(event)
        return

    #Read Shift + mouse wheel inside the image canvas to rotate the image
    def ShiftMouseWheel_Image(self, event): 
        self.shft_btn = 1    
        self.MouseWheel_entRot(event)
        return

    #Read Alt + mouse wheel inside the image canvas to rotate the ROI
    def AltMouseWheel_Image(self, event): 
        self.alt_btn = 1    
        if(self.btn_ROI_var == 1):
            theta_increment = 1
            self.MouseWheel_ROIrot(event,theta_increment)
        return

    #Read Ctrl+Alt + mouse wheel inside the image canvas to rotate the ROI in fine tunning
    def CtrlAltMouseWheel_Image(self, event): 
        self.alt_btn = 1    
        if(self.btn_ROI_var == 1):
            theta_increment = 0.1
            self.MouseWheel_ROIrot(event,theta_increment)
            
            ## update ROIpx if changed
            #if len(self.ROI) != len(self.ROIpx) or self.pre_ROI0 != self.ROI:
                #self.ROIpx = [[] for i in range(len(self.ROI))] # pixels (i,j) in tROI
                #for i in range (len(self.ROI)):
                    #self.ROIpx[i] = Extract_tROIpx(self.ROI[i])   # 3D array: [n-th ROI of ROIs], [n-th px in ROI], [i or j value]                 
                #self.pre_ROI0 = self.ROI[0]
            # draw ROI pixels in 2nd subROI for verification purpose, skip if no need
            #print('self.alt_btn=',self.alt_btn,'self.ctrl_btn=',self.ctrl_btn)
            #Draw_ROIpx(self.ROIpx,self.frame.canvas)
        return
    
    def MouseWheel_ROIrot(self, event,theta_increment):
        rotGRID,a,b,r,min_theta, max_theta = Calculate_GRID_corners(self.ROI,self.ROI_rot)        
        
        #update the rotational angle when a manual change
        self.ROI_rot += theta_increment*int(event.delta/120)

        if min_theta < self.ROI_rot < max_theta:
            self.ROI = Draw_ROI(self.GRID, self.ROI, self.zf, self.frame.canvas, self.ROI_rot)
        else:
            self.ROI_rot -= theta_increment*int(event.delta/120)
        # filter digi residual error 
        if -0.001<self.ROI_rot<0.001: self.ROI_rot = 0
        #print('ROI_rot=',self.ROI_rot,'theta=%.3f ~ %.3f'%(min_theta,max_theta))
        #print('self.ROI=',self.ROI)

        # display ROI_rot while scrolling
        self.t0.delete('1.0',tk.END)
        self.t0.insert(tk.END,'ROI rotation:%.2f deg'%(self.ROI_rot))

        # draw a circle enclosing a grid
        if self.MouseBtn2 and self.ROI_rot!=0:
            X1,Y1,X2,Y2,zf = a-r,b-r,a+r,b+r,self.zf
            self.frame.canvas.create_oval(X1*zf,Y1*zf,X2*zf,Y2*zf, outline='red')    #oval ROI
        return

    #Read mouse wheel inside the image canvas to scroll the image
    def MouseWheel_Img(self, event):  
        #print('self.ctrl_btn=',self.ctrl_btn,'self.shft_btn=',self.shft_btn,'winfo_height()=',self.frame.canvas.winfo_height())
        if(self.ctrl_btn != 1 and self.shft_btn != 1):  #cf)&: bitwise operation, dealing w numbers
            if self.flag_im:
                self.frame.canvas.yview_scroll(-1*int(event.delta/120), "units")             
                if(self.imH*self.zf > self.frame.canvas.winfo_height()):
                    self.frame.canvas.yview_scroll(-1*int(event.delta/120), "units")             
        return

    #Read mouse wheel inside the rotate entry for angle & update the image rotated
    def MouseWheel_entRot(self, event):       
        #update the entry when a manual change
        self.MouseWheelCnt_Rot += 5*int(event.delta/120)
        self.rot = self.MouseWheelCnt_Rot    #Rotate factor of angle
        #update the Rotate entry
        self.entRot.delete(0,tk.END)
        self.entRot.insert(0,"%.3f"%self.rot)
        self.btn_ROTATE()
        return
    
    def Image_Curtain(self,event,mode,zf):
        if mode=='MouseClick':  
            self.cursor = 'right_side'        
            self.frame.canvas.config(cursor=self.cursor)             

            if self.im_nband==1 and len(self.imScn.shape)==2: 
                im_curtain = np.invert(np.zeros((self.imH,self.imW,1),dtype=np.uint8))    #height(row) x width(col), white 
                im_curtain[:,event.x:] = self.im_stretched[:,event.x:]   # assign 1b base image from cursor
                im_curtain[:,:event.x] = cv2.cvtColor(self.imScn[:,:event.x], cv2.COLOR_GRAY2RGB) # assign 3b base image from cursor
            else:
                im_curtain = np.invert(np.zeros((self.imH,self.imW,3),dtype=np.uint8))    #height(row) x width(col), white 
                if self.im_nband==1: 
                    im_curtain[:,event.x:,:] = cv2.cvtColor(self.im_stretched[:,event.x:], cv2.COLOR_GRAY2RGB)   # assign 3b base image from cursor
                else:
                    im_curtain[:,event.x:,:3] = self.im_stretched[:,event.x:,:3]   # assign 3b base image from cursor
                if len(self.imScn.shape)==2:                    
                    im_curtain[:,:event.x] = cv2.cvtColor(self.imScn[:,:event.x], cv2.COLOR_GRAY2RGB)   # assign 3b top image up to cursor
                else:
                    im_curtain[:,:event.x,:] = self.imScn[:,:event.x,:3] # assign 1b top image up to cursor
            #print('cursor (x,y) =',event.x,event.y,'imScn_base=',self.imScn_base.shape,self.imScn_base.dtype,'imScn=',self.imScn.shape,self.imScn.dtype,'im_curtain=',im_curtain.shape,im_curtain.dtype)        
            
            # display the curtain image
            global img_photo     #must be here to clear img_photo memory buildup
            # convert imScn to 3b img
            if len(im_curtain.shape)>2 and im_curtain.shape[2]>3:
                img_photo = ImageTk.PhotoImage(Image.fromarray(cv2.cvtColor(im_curtain, cv2.COLOR_BGRA2BGR)))   # buffer (x3), processing time (1s for 166MB)
            else:
                if len(im_curtain.shape)>2 and im_curtain.shape[2]==1:
                    im_curtain = im_curtain.reshape(im_curtain.shape[0],im_curtain.shape[1])   # (H,W,1)->(H,W) for ImageTk.PhotoImage()
                img_photo = ImageTk.PhotoImage(Image.fromarray(im_curtain))   # (file=self.src_file) buffer (x3), processing time (1s for 166MB)
            #Display image
            self.frame.canvas.delete('all')    # clear the canvas, e.g., removing old ROI drawing
            self.frame.canvas.create_image(0,0,anchor="nw",image=img_photo) # buffer (x2), processing time (1s for 166MB)
            #Update the scrollbar length    
            self.frame.canvas.config(scrollregion=(0,0,int(self.imW*self.zf),int(self.imH*self.zf)))    
        return

    def UpdateMouse_Image(self,event,mode,zf):
        # update imScnW & imScnH of im->imScn if the image zoomed
        if self.flag_im: 
            imScnW = int(self.imW*self.zf) #Width\Col of the input image, =img.size[0]=im.shape[1] 
            imScnH = int(self.imH*self.zf) #Height\Row of the input image, =img.size[1]=im.shape[0]        
        #print('im(W,H)=',imScnW,imScnH)
        
        #Shift the cursor coordinate (1,1)=>(0,0) to match to image buffer address starting [0,0]
        #event.x = event.x - 1
        #event.y = event.y - 1
        
        #Adjust the cursor position according to scrollbar changes
        px_width = round(imScnW*self.frame.sbarH.get()[1])
        px_height = round(imScnH*self.frame.sbarV.get()[1])
        #print('self.frame.sbarH.get()[1]=',self.frame.sbarH.get()[1],'self.frame.sbarV.get()[1]=',self.frame.sbarV.get()[1])
        #print('winfo_width()=%d'%self.frame.canvas.winfo_width(),'px_width=',px_width)
        #print('winfo_height()=%d'%self.frame.canvas.winfo_height(),'px_height=',px_height)
        if px_width > self.frame.canvas.winfo_width():
            event.x = event.x + int(px_width -self.frame.canvas.winfo_width())
        if px_height > self.frame.canvas.winfo_height():
            event.y = event.y + int(px_height-self.frame.canvas.winfo_height())

        #Adjust the scrollbars according to ROI extending
        if self.btn_ROI_var==1 and mode=='MouseClick':            
            if px_width >= self.frame.canvas.winfo_width():
                if event.x > px_width-1:            
                    self.frame.canvas.xview_scroll(1, "units")                
                if event.x < px_width-self.frame.canvas.winfo_width()+1:            
                    self.frame.canvas.xview_scroll(-1, "units")        
            if px_height >= self.frame.canvas.winfo_height():
                if event.y > px_height-1:
                    self.frame.canvas.yview_scroll(1, "units")                            
                if event.y < px_height-self.frame.canvas.winfo_height()+1:            
                    self.frame.canvas.yview_scroll(-1, "units")        

        #print ('Mouse position: (%s %s)'%(event.x,event.y))
        
        ts = self.ts
        t0 = self.t0
        t1 = self.t1
        t2 = self.t2
        # delete Text widgets
        ts.delete('1.0',tk.END)
        t0.delete('1.0',tk.END)
        t1.delete('1.0',tk.END)
        t2.delete('1.0',tk.END)

        ts.place(rely=0.00, height=18)     #1st line for statistics
        t0.place(rely=0.04, height=18)     #1st line for ROI
        t1.place(rely=0.08, height=18)     #2nd line for RGB
        t2.place(rely=0.12, height=18)     #2nd line for GPS 
        
        #Display image stats
        ts.insert(tk.END,'%db:'%self.im_nband)
        t1.insert(tk.END,'Cursor:')
        #Display mean of each img channel
        for i in range(len(self.Ave)): 
            if self.Ave[i]<1:   #32-bit geotiff NDVI
                ts.insert(tk.END,'%.2f'%(self.Ave[i]))
            elif self.Ave[i]>255:   #64-bit raw geotiff image
                ts.insert(tk.END,'%d'%(round(self.Ave[i])))
            else:
                ts.insert(tk.END,'%d'%(round(self.Ave[i])))
            if i<len(self.Ave)-1: ts.insert(tk.END,',')
        #Display min~max of each img channel
        for i in range(len(self.Ave)): 
            if i==0: ts.insert(tk.END,'(')
            if self.Ave[i]<1:   #32-bit geotiff NDVI
                ts.insert(tk.END,'%.2f~%.2f'%(self.Min[i],self.Max[i]))
            elif self.Ave[i]>255:   #64-bit raw geotiff image
                ts.insert(tk.END,'%d~%d'%(round(self.Min[i]),round(self.Max[i])))
            else:
                ts.insert(tk.END,'%d~%d'%(round(self.Min[i]),round(self.Max[i])))
            if i<len(self.Ave)-1: ts.insert(tk.END,',')
            else: ts.insert(tk.END,')')            
        
        #Display ROI statistics
        if self.ROI_var.get()==1 and self.flag_segStats and self.GRID_ave!=[]:
            self.Insert_ROIstats()

        if self.georef == 1:
            #Read GPS coordinate from geotiff (a raster layer)                
            #print('self.src_file=',self.src_file)
            self.UTM_E = self.minx + event.x*self.E_res #UTM of original EPSG
            self.UTM_N = self.maxy + event.y*self.N_res #UTM of original EPSG
            
            #Convert UTM to WGS 84 Lat\Lon
            if self.flag_im: Lat, Lon = Convert_UTM2LL(self.UTM_E,self.UTM_N,None,self.ds.GetProjection())  #LL of WGS 84
            else:            Lat, Lon = Convert_UTM2LL(self.UTM_E,self.UTM_N,12)  #LL of WGS 84 
            #print('LUTM_E,UTM_N=',self.UTM_E,self.UTM_N)
            #print('Lon, Lat=',Lon, Lat)
            #Display GPS UTM coordinate of image px of the cursor
            t2.insert(tk.END,'GPS:%.5f,%.5f[%.2f,%.2f]'%(Lon,Lat,self.UTM_E,self.UTM_N))    #NOTE: lon\lat seems swapped in gdal 3.0?
                
        #change the cursor to d-arrow if on the ROI corners     
        if mode=='MouseMotion':
            (self.X1,self.Y1),(self.X2,self.Y2) = (self.ROI[0][:2], list(np.add(self.ROI[len(self.ROI)-1][2:4],-1)))    # need -1, e.g. 640 should be W=0~639
            if event.x < imScnW and event.y < imScnH:   #exclude becuz 640x480 becomes 0~639 x 0~479
                #print ('(W,H)=',imScnW,imScnH,'event.(x,y)=',event.x,event.y,'self.ROI=',self.X1,self.Y1)
                # detect ROI corners to resize
                if self.btn_ROI_var==1:
                    if self.X1*zf-2<=event.x<=self.X1*zf+2 and self.Y1*zf-2<=event.y<=self.Y1*zf+2:                
                        self.cursor = 'top_left_corner'
                    elif self.X2*zf-2<=event.x<=self.X2*zf+2 and self.Y2*zf-2<=event.y<=self.Y2*zf+2:                
                        self.cursor = 'bottom_right_corner'
                    elif self.X2*zf-2<=event.x<=self.X2*zf+2 and self.Y1*zf-2<=event.y<=self.Y1*zf+2:                
                        self.cursor = 'top_right_corner'
                    elif self.X1*zf-2<=event.x<=self.X1*zf+2 and self.Y2*zf-2<=event.y<=self.Y2*zf+2:                
                        self.cursor = 'bottom_left_corner'
                    # detect ROI edges to resize
                    elif self.X1*zf+20<event.x<self.X2*zf-20 and self.Y1*zf-2<event.y<self.Y1*zf+2: 
                        self.cursor = 'sb_v_double_arrow'
                        self.v_arrow = 'top'
                    elif self.X1*zf+20<event.x<self.X2*zf-20 and self.Y2*zf-2<event.y<self.Y2*zf+2:
                        self.cursor = 'sb_v_double_arrow'
                        self.v_arrow = 'bottom'
                    elif self.X1*zf-2<event.x<self.X1*zf+2 and self.Y1*zf+20<event.y<self.Y2*zf-20:
                        self.cursor = 'sb_h_double_arrow'
                        self.h_arrow = 'left'
                    elif self.X2*zf-2<event.x<self.X2*zf+2 and self.Y1*zf+20<event.y<self.Y2*zf-20:
                        self.cursor = 'sb_h_double_arrow'
                        self.h_arrow = 'right'
                    # detect ROI body at around corners to transit the whole ROI
                    elif ((self.Y1*zf-2<event.y<self.Y1*zf+2 and (self.X1*zf+5<event.x<self.X1*zf+20 or self.X2*zf-20<event.x<self.X2*zf-5)) or
                          (self.Y2*zf-2<event.y<self.Y2*zf+2 and (self.X1*zf+5<event.x<self.X1*zf+20 or self.X2*zf-20<event.x<self.X2*zf-5)) or                                              
                          (self.X1*zf-2<event.x<self.X1*zf+2 and (self.Y1*zf+5<event.y<self.Y1*zf+20 or self.Y2*zf-20<event.y<self.Y2*zf-5)) or
                          (self.X2*zf-2<event.x<self.X2*zf+2 and (self.Y1*zf+5<event.y<self.Y1*zf+20 or self.Y2*zf-20<event.y<self.Y2*zf-5))):
                        self.cursor = 'fleur'
                        #self.X1,self.Y1,self.X2,self.Y2,r = self.ROI[0]
                        #print('self.ROI[0]=',self.ROI[0])
                    else:                
                        self.cursor = 'left_ptr'                          
                else:                
                    self.cursor = 'left_ptr'      
                
                t1.insert(tk.END,'[%d,%d]'%(event.x,event.y))                
                if self.btn_ROI_var==1: # find plot ID (R,C) when ROI selected
                    #print('self.GRID=',self.GRID,'len(self.ROI)=',len(self.ROI))
                    #print('self.ROI=',self.ROI)                    
                    Row = Col = 0
                    for i in range(len(self.ROI)):
                        x1, y1 = self.ROI[i][0], self.ROI[i][1]
                        x2, y2 = self.ROI[i][2], self.ROI[i][1]
                        x3, y3 = self.ROI[i][2], self.ROI[i][3]
                        x4, y4 = self.ROI[i][0], self.ROI[i][3]                    
                        polygon = [(x1,y1),(x2,y2),(x3,y3),(x4,y4)]
                        points = [(event.x,event.y)]
                        inside = pltPath(polygon).contains_points(points)
                        #print('inside=',inside)                
                        if inside: 
                            Row = int(i/self.GRID[1])+1
                            Col = i%self.GRID[1]+1
                            #print('R%dC%d'%(Row,Col))
                            break                    
                    t1.insert(tk.END,'@R%dC%d'%(Row,Col))
                if 0<=event.x < imScnW and 0<=event.y < imScnH:                                        
                    if self.ch == 'NDVI':
                        #print('imNDVI=',self.imNDVI.shape,self.imNDVI.dtype)
                        t1.insert(tk.END,' %s (%s)'%(self.imScn[event.y,event.x],self.imNDVI[event.y,event.x]))
                    else:
                        t1.insert(tk.END,' %s (%s)'%(self.imScn[event.y,event.x],self.im[event.y,event.x][0]))
                
                self.pre_x = event.x    # save for ROI(pre_x,pre_y,x2,y2)
                self.pre_y = event.y    # save for ROI(pre_x,pre_y,x2,y2)
            else:
                self.cursor = 'left_ptr'                        
            self.frame.canvas.config(cursor=self.cursor)     
            
            #Switch if opposite corners of ROI
            if self.X1*zf > self.X2*zf: self.X1,self.X2 = self.X2,self.X1
            if self.Y1*zf > self.Y2*zf: self.Y1,self.Y2 = self.Y2,self.Y1
            
            # display ROI_rot while mouse motion
            self.t0.delete('1.0',tk.END)
            self.t0.insert(tk.END,'ROI rotation:%.2f deg'%(self.ROI_rot))
            
        #print('mode=',mode,'cursor=',self.cursor)
        else:
            #Limit the cursor coord within the image size           
            if event.x < 0: event.x = 0
            if event.y < 0: event.y = 0
            if event.x >= imScnW: event.x = imScnW-1  # if ROI on image edge, substract 1, e.g. 640x480 should be W=0~639
            if event.y >= imScnH: event.y = imScnH-1  # if ROI on image edge, substract 1, e.g. 640x480 should be H=0~479
            if self.cursor == 'top_left_corner':
                self.x1 = event.x
                self.y1 = event.y
                self.x2 = self.X2*zf
                self.y2 = self.Y2*zf                   
            elif self.cursor == 'bottom_right_corner':
                self.x1 = self.X1*zf
                self.y1 = self.Y1*zf               
                self.x2 = event.x
                self.y2 = event.y            
            elif self.cursor == 'top_right_corner':
                self.x1 = self.X1*zf
                self.y1 = event.y
                self.x2 = event.x
                self.y2 = self.Y2*zf
            elif self.cursor == 'bottom_left_corner':
                self.x1 = event.x
                self.y1 = self.Y1*zf
                self.x2 = self.X2*zf
                self.y2 = event.y
            elif self.cursor == 'sb_v_double_arrow':
                self.x1 = self.X1*zf
                self.y1 = self.Y1*zf
                self.x2 = self.X2*zf                
                self.y2 = self.Y2*zf
                if self.v_arrow == 'top':      self.y1 = event.y
                elif self.v_arrow == 'bottom': self.y2 = event.y
            elif self.cursor == 'sb_h_double_arrow':
                self.x1 = self.X1*zf
                self.y1 = self.Y1*zf
                self.x2 = self.X2*zf                
                self.y2 = self.Y2*zf
                if self.h_arrow == 'left':    self.x1 = event.x
                elif self.h_arrow == 'right': self.x2 = event.x
            elif self.cursor == 'fleur':
                self.x1 = self.X1*zf + event.x - self.pre_x
                self.y1 = self.Y1*zf + event.y - self.pre_y
                self.x2 = self.X2*zf + event.x - self.pre_x+1   #add +1 to fix 1-px shorter unknown 
                self.y2 = self.Y2*zf + event.y - self.pre_y+1   #add +1 to fix 1-px shorter unknown
            else:            
                self.x1 = self.pre_x     
                self.y1 = self.pre_y 
                self.x2 = event.x
                self.y2 = event.y
            t0.delete('1.0',tk.END)
            #if self.zf!=1:  # convert float->int when zoomed
            self.x1 = int(self.x1)
            self.y1 = int(self.y1)
            self.x2 = int(self.x2)
            self.y2 = int(self.y2)
            
            #print('self.x2,self.y2=',self.x2,self.y2)
            t0.insert(tk.END,'ROI:[%d,%d]-[%d,%d] %s'%(self.x1,self.y1,self.x2,self.y2,self.imScn[self.y2,self.x2]))
            #print('(x,y)=[%d,%d]-[%d,%d]'%(self.x1,self.y1,self.x2,self.y2))
            #print('elf.cursor=',self.cursor,'pre_x,y=',self.pre_x,self.pre_y,'x,y=',event.x,event.y,'x-pre_x=',event.x-self.pre_x,'y-pre_y=',event.y-self.pre_y,'self.ROI=',self.ROI[0])
        
        return 

    #Read the mouse position inside the slider
    def MouseMotion_SliderL(self, event):           
        #Limit the sliders not to cross over
        for i in range(self.im_nband):
            if self.RdBtn[i]['state']=='normal' and self.sldL[i].get() > self.sldU[i].get():                
                self.sldU[i].set(self.sldL[i].get())
            #print('sldL',self.sldL[i].get())
            self.bdry[i][0] = self.sldL[i].get()
        #Update the thresholding lines
        if self.flag_im:      
            self.Histogram()
            #self.SEG_var.set(1) 
            if(self.SEG_var.get() == 1):
                self.btn_SEG()  # update ROI_seg insert as slider moves
        return    
    
    #Read the mouse position inside the slider
    def MouseMotion_SliderU(self, event):             
        #Limit the sliders not to cross over
        for i in range(self.im_nband):
            if self.sldL[i].get() > self.sldU[i].get():
                self.sldL[i].set(self.sldU[i].get())
            #print('sldU',self.sldU[i].get())
            self.bdry[i][1] = self.sldU[i].get()
        #Update the thresholding lines
        if self.flag_im:      
            self.Histogram()
            #self.SEG_var.set(1) 
            if(self.SEG_var.get() == 1):
                self.btn_SEG()  # update ROI_seg insert as slider moves
        return    
    
    #Read the mouse position in the histogram
    def MouseMotion_Histo(self, event):        
        #print('Mouse position: (%s %s)'%(event.x,event.y))   
        t = self.t
        t.delete('1.0',tk.END)
        t.place(relx=0.91, rely=0.20, relwidth=.1, height=20)
        X = event.x - 58
        if X<0:     X = 0
        elif X>255: X = 255
        t.insert(tk.END,X)

        #change the cursor to d-arrow if on the thresholding lines
        #print('self.ch=',self.ch)
        if X==self.sldL[0].get() or X==self.sldU[0].get() or X==self.sldL[1].get() or X==self.sldU[1].get() or X==self.sldL[2].get() or X==self.sldU[2].get() or X==self.sldL[3].get() or X==self.sldU[3].get():
            self.canvas.get_tk_widget().config(cursor='sb_h_double_arrow')
            #find which line is detected
            if self.ch in [0,1,2]:  #if 1-ch image
                if   X==self.sldL[self.ch].get(): self.cursor_on_line = 1+self.ch*2
                elif X==self.sldU[self.ch].get(): self.cursor_on_line = 2+self.ch*2
                
            else:   #if color image
                if   X==self.sldL[0].get(): self.cursor_on_line = 1
                elif X==self.sldU[0].get(): self.cursor_on_line = 2
                elif X==self.sldL[1].get(): self.cursor_on_line = 3
                elif X==self.sldU[1].get(): self.cursor_on_line = 4
                elif X==self.sldL[2].get(): self.cursor_on_line = 5
                elif X==self.sldU[2].get(): self.cursor_on_line = 6
                elif X==self.sldL[3].get(): self.cursor_on_line = 7
                elif X==self.sldU[3].get(): self.cursor_on_line = 8                
        else:
            self.canvas.get_tk_widget().config(cursor='arrow')
        
        return
    
    #Read the mouse position on the thresholding line in the histogram
    def MouseMotion_Histo_Line(self, event):        
        #print('Mouse position: (%s %s)'%(event.x,event.y))
        t = self.t
        t.delete('1.0',tk.END)
        t.place(relx=0.91, rely=0.20, relwidth=.1, height=20)        
        X = event.x - 58
        if X<0:     X = 0
        elif X>255: X = 255   
        t.insert(tk.END,X)
        #print('X=%d'%X,'self.cursor_on_line=%d'%self.cursor_on_line,'self.sldU[%d].get()=%d'%(-(-self.cursor_on_line//2)-1,self.sldU[-(-self.cursor_on_line//2)-1].get()))

        #change the thresholding slider corresponding to the cursor move
        if   self.cursor_on_line == 1: 
            bdry = [[X,self.sldU[0].get()],[self.sldL[1].get(),self.sldU[1].get()],[self.sldL[2].get(),self.sldU[2].get()],[self.sldL[3].get(),self.sldU[3].get()]]
        elif self.cursor_on_line == 2: 
            bdry = [[self.sldL[0].get(),X],[self.sldL[1].get(),self.sldU[1].get()],[self.sldL[2].get(),self.sldU[2].get()],[self.sldL[3].get(),self.sldU[3].get()]]            
        elif self.cursor_on_line == 3: 
            bdry = [[self.sldL[0].get(),self.sldU[0].get()],[X,self.sldU[1].get()],[self.sldL[2].get(),self.sldU[2].get()],[self.sldL[3].get(),self.sldU[3].get()]]                            
        elif self.cursor_on_line == 4: 
            bdry = [[self.sldL[0].get(),self.sldU[0].get()],[self.sldL[1].get(),X],[self.sldL[2].get(),self.sldU[2].get()],[self.sldL[3].get(),self.sldU[3].get()]]            
        elif self.cursor_on_line == 5: 
            bdry = [[self.sldL[0].get(),self.sldU[0].get()],[self.sldL[1].get(),self.sldU[1].get()],[X,self.sldU[2].get()],[self.sldL[3].get(),self.sldU[3].get()]]            
        elif self.cursor_on_line == 6: 
            bdry = [[self.sldL[0].get(),self.sldU[0].get()],[self.sldL[1].get(),self.sldU[1].get()],[self.sldL[2].get(),X],[self.sldL[3].get(),self.sldU[3].get()]]                        
        elif self.cursor_on_line == 7: 
            bdry = [[self.sldL[0].get(),self.sldU[0].get()],[self.sldL[1].get(),self.sldU[1].get()],[self.sldL[2].get(),self.sldU[2].get()],[X,self.sldU[3].get()]]                        
        elif self.cursor_on_line == 8: 
            bdry = [[self.sldL[0].get(),self.sldU[0].get()],[self.sldL[1].get(),self.sldU[1].get()],[self.sldL[2].get(),self.sldU[2].get()],[self.sldL[3].get(),X]]
            
        #Update the thresholding lines
        self.reset_sliders(bdry)
        if self.flag_im:     
            self.Histogram()
            self.SEG_var.set(1) 
            self.btn_SEG()  # update ROI_seg insert as slider moves
        return    
 
    def reset_sliders(self,bdry):
        #print('bdryp',bdry)
        self.bdry = bdry
        for i in range(4): 
            #print('bdry=%d~%d'%(int(bdry[i][0]),int(bdry[i][1])))
            self.sldL[i].set(bdry[i][0])
            self.sldU[i].set(bdry[i][1])        

    # Calculate image statistics
    def Calculate_dsStat(self,ds,i):
        # reset stats before new calculations
        if i==0:
            self.Min = []
            self.Max = []
            self.Ave = []
            self.Stdev = []        
        
        band = ds.GetRasterBand(i+1)  #Get the first band
        stats = band.GetStatistics(0,1) # [min, max, mean, stdev] 
        self.Min.append(stats[0])
        self.Max.append(stats[1])
        self.Ave.append(stats[2])
        self.Stdev.append(stats[3])   
        #print('%d: Min=%.3f, Max=%.3f, d=%.3f, Ave=%.3f, Std=%.3f, %d~%d %d+/-%d'%(i,self.Min[i],self.Max[i],self.Max[i]-self.Min[i],self.Ave[i],self.Stdev[i],self.Min[i]*scale,self.Max[i]*scale,self.Ave[i]*scale,self.Stdev[i]*scale))
    
    def Calculate_imStat(self,im,i):
        self.Min.append(np.nanmin(im[i]))
        self.Max.append(np.nanmax(im[i]))
        self.Ave.append(np.nanmean(im[i]))
        self.Stdev.append(np.nanstd(im[i]))   
        #print('%d: Min=%d, Max=%d, d=%d, Ave=%.3f, Std=%.3f, %d~%d %d+/-%d'%(i,self.Min[i],self.Max[i],self.Max[i]-self.Min[i],self.Ave[i],self.Stdev[i],self.Min[i]*scale,self.Max[i]*scale,self.Ave[i]*scale,self.Stdev[i]*scale))

    def Load_Image(self, file=None):           
        if file == None: self.src_file = LoadFile(self,'img')     
        else: self.src_file = file
        if len(self.src_file)==0:
            return        

        # set the default status of buttons
        self.MASK_var.set(0)
        self.SEG_var.set(0)
        self.ROI_var.set(0)
        # set the default of Rotate button
        self.entRot.delete(0,tk.END)
        self.entRot.insert(0, "0.0")
        self.rot = 0
        
        #os.sys('gdalwarp infile.tif outfile.tif -t_srs "+proj=longlat +ellps=WGS84"')

        # load a geotiff image
        self.flag_im_base = 0        
        self.flag_im = 1    # image loaded
        gdal.SetConfigOption('GDAL_PAM_ENABLED','NO')  # avoid "ERROR 6: band 1: SetUnitType() not supported on this raster band."
        im,ds = LoadGeoTiff(self.src_file) # buffer (x2.5)        
        self.ds = ds
        self.im = im    # R,G,B,N =im[:,:,0],im[:,:,1],im[:,:,2],im[:,:,3] 
        #cv2.imshow('im0',im[:,:,0]), cv2.imshow('im1',im[:,:,1]),cv2.imshow('im2',im[:,:,2]),cv2.waitKey()
        #cv2.imshow('im',cv2.cvtColor(im,cv2.COLOR_RGB2BGR)), cv2.waitKey(0) # cv2.imshow() needs RGB->BGR or im[:,:,::-1]

        self.UNIT = osr.SpatialReference(wkt=ds.GetProjection()).GetAttrValue('UNIT',0)
        if self.UNIT == None:
            print('not georeferenced')
            self.flag_georef = False
        else: 
            self.flag_georef = True
        nband = ds.RasterCount
        if nband>4: nband = 4  # assign nband=4 if 8b PS2 image or +5b image
        self.im_nband = nband
        #if nband>4: messagebox.showwarning(title='Warning',message='The image is +4 bands!')
        self.imH = im.shape[0]    #Height\Row of the image, =im.shape[0]=ds.RasterWSize
        self.imW = im.shape[1]    #Width\Col of the image, =im.shape[1]=ds.RasterXSize 
        #print('self.imHxW=',self.imH,self.imW,'nband=',nband,'dtype=',im.dtype,'UNIT=',self.UNIT)

        self.master.title("IMAP"+'  - '+self.src_file+' [%s,%s]'%(im.shape,im.dtype)) #change the title of master widget        
        #print('src_file=',self.src_file)

        #Display the loaded image in the scrolled window          
        if   nband == 1: self.btn_cmap = 'R';   self.ch = 'Gray'   #1-ch image
        elif nband == 2: self.btn_cmap = 'RG';  self.ch = 'C&S'    #2-ch Cloudy&Shadow mask image                                
        elif nband == 3: self.btn_cmap = 'RGB'; self.ch = 'RGB'    #3-ch color image
        elif nband == 4: 
            self.btn_cmap = 'RGBY'
            self.ch = 'RGBN'   #4-ch multi-band image
            if len(im[:,:,3])==0: nband = 3 #detect RGBA img        
                #self.ViewMenu.entryconfig(4,state='normal') 
        #else: messagebox.showerror(title='Error',message='Image is +4 bands!')            
        self.Update_BtnSlider(nband,'normal')   

        # define a scale factor for file type to 0~255
        #print('np.nanmax(im)=',np.nanmax(im))
        if np.nanmax(im)<=1: scale = 255                 #mono or float32 1b image
        elif im.dtype == 'uint16' or im.dtype == 'int16': scale = 255/pow(2,16) #10000.     #64-bit raw geotiff image (0-65535)
        else:   scale = 1   # uint8
        self.im_scale = scale
        #print('scale=',scale)

        # stretch image if stretch option checked or not uint8 
        if self.stretch[0]>=0 and im.dtype != 'uint8':   # Stretch options: 0 for Percentile, 1 for min/max, unit16->unit8 w percentile stretch
            self.im_stretched = self.Update_StretchVar(im,self.stretch)
        else: 
            self.im_stretched = im
        
        self.imScn_base = self.im_stretched    #baseimage of imScn before segmentation
        # calculate max_y for histogram display
        self.max_y = 0
        for i in range(self.im_nband): 
            self.max_y = max(self.max_y,np.max(cv2.calcHist([self.imScn_base],[i],None,[256],[1,256]))) # 1: exclude saturated 0s
        
        # reset Img_Select(), btn_Band()
        self.combxViewImg.current(0)    # {'RGB':0,'HSV':1,'LAB':2,'NDVI':3,'Heatmap':4}
        self.Band_var.set(None) # deselect all Band Radiobuttons
        
        # get the ROI resized to fit the image loaded if oversized ROI
        if self.ROI[len(self.ROI)-1][2]>=self.imW or self.ROI[len(self.ROI)-1][3]>=self.imH:
            self.ROI = Update_ROIs(self.GRID,self.ROI,self.ROI_rot) 
        
        (self.X1,self.Y1),(self.X2,self.Y2) = (self.ROI[0][:2], self.ROI[len(self.ROI)-1][2:4])
        
        #print('im=',np.nanmax(im),'im_stretched=',np.nanmax(self.im_stretched))
        #if self.zf!=1: self.Img_Zoom()
        #else: 
        self.Display_Image(self.im_stretched)  # im_stretched to fit display in uint8 format
        
        #Calculate statistics of the image channels
        #if os.stat(self.src_file).st_size/1024/1024 < 25:  # do if file size < 25MB
        for i in range(nband):  #nband goes 1(B), 2(G), 3(R), 4(N)
            self.Calculate_dsStat(ds,i)
        self.GRID_ave = self.Ave
        #print('self.GRID_ave=',self.GRID_ave)
        
        #Display image if not 4b 
        if len(ds.GetProjection())==0:   #check if image is not geotiff            
            self.georef = 0 # flag for non-georeferenced image
            return       
        else: self.georef = 1   # flag for georeferenced image
            
        #Calculate geospatial location to display GPS coord on screen
        gt = ds.GetGeoTransform()   #ulx, xres, xskew, uly, yskew, yres
        self.PxRol = gt[1]  # pixel resolution
        if gt:
            #print('width=',self.imW,'height=',self.imH)
            #print('Pixel Resolution = ({},{})'.format(gt[1],gt[5]))  #pixel resol [ft] in pos & neg
            #print('Origin = ({},{})'.format(gt[0],gt[3]))    #left-top (x,y) in UTM
            print('gt[]=',gt[0],gt[1],gt[2],gt[3],gt[4],gt[5])
            self.minx = gt[0]
            self.miny = gt[3] + self.imW*gt[4] + (self.imH-1)*gt[5] 
            self.maxx = gt[0] + (self.imW-1)*gt[1] + self.imH*gt[2]
            self.maxy = gt[3]  
            print('minx=',self.minx,'miny=',self.miny,'maxx=',self.maxx,'maxy=',self.maxy)            
            self.E_res = gt[1]
            self.N_res = gt[5]
                
        ##Display image px means on histogram canvas        
        #t = tk.Text(self.canvas.get_tk_widget(),relief=tk.FLAT)
        #t.place(height=20)  
        return  

   
    def Filter_Erode(self):   
        self.cnt_erode += 1
        print('self.cnt_erode=',self.cnt_erode)
        self.FilterMenu.entryconfig(0,label='Erode (%d done)'%self.cnt_erode) 
        im = self.imScn        
        #im_proc = cv2.morphologyEx(im,cv2.MORPH_ERODE,np.ones((2,2)))
        im_proc = cv2.erode(im,np.ones((2,2),np.uint8),1)
        self.Display_Image(im_proc) 
       
    def Filter_Dilate(self):   
        self.cnt_dilate += 1
        self.FilterMenu.entryconfig(1,label='Dilate (%d done)'%self.cnt_dilate) 
        im = self.imScn   
        im_proc = cv2.morphologyEx(im,cv2.MORPH_DILATE,np.ones((2,2)))
        self.Display_Image(im_proc)

    def Filter_Opening(self):   #Opening (erode+dilate) to remove noise blobs
        im = self.imScn 
        im_proc = cv2.morphologyEx(im,cv2.MORPH_OPEN,np.ones((2,2)))
        self.Display_Image(im_proc)
       
    def Filter_Closing(self):   #Closing (dilate+erode) to fill holes
        im = self.imScn 
        im_proc = cv2.morphologyEx(im,cv2.MORPH_CLOSE,np.ones((2,2)))
        self.Display_Image(im_proc)
        
    def Filter_Opening5(self):   #Opening (erode+dilate) to remove noise blobs
        im = self.imScn 
        im_proc = cv2.morphologyEx(im,cv2.MORPH_OPEN,np.ones((5,5)))
        self.Display_Image(im_proc)
       
    def Filter_Closing5(self):   #Closing (dilate+erode) to fill holes
        im = self.imScn 
        im_proc = cv2.morphologyEx(im,cv2.MORPH_CLOSE,np.ones((5,5)))
        self.Display_Image(im_proc)
           
    def Filter_FFT(self):           
        im_fft = FFT(self.imScn)        
        self.Display_Image(im_fft)

    def Filter_Gradient(self):   #edge detection as difference b/w erosion and dilation of the image
        im = self.imScn
        im_proc = cv2.morphologyEx(im,cv2.MORPH_GRADIENT,np.ones((2,2)))
        self.Display_Image(im_proc)
       
    def Filter_TopHat(self):   #difference b/w input & opening of the image 
        im = self.imScn
        im_proc = cv2.morphologyEx(im,cv2.MORPH_TOPHAT,np.ones((2,2)))
        self.Display_Image(im_proc)
        
    def Filter_Mean(self):
        im = self.imScn
        #print 'len(im.shape)=',len(im.shape),'im.shape=',im.shape
        if len(im.shape)>2 and im.shape[2]==3:    #if color image
            im = cv2.cvtColor(im,cv2.COLOR_BGR2GRAY)         
        mean = cv2.adaptiveThreshold(im,255, cv2.ADAPTIVE_THRESH_MEAN_C, cv2.THRESH_BINARY, 115,1)
        im_threshold = cv2.bitwise_and(im,im,mask=mean)
        self.Display_Image(im_threshold)
        
    def Filter_Gaus(self):
        im = self.imScn
        if len(im.shape)>2 and im.shape[2]==3:    #if color image
            im = cv2.cvtColor(im,cv2.COLOR_BGR2GRAY)         
        gaus = cv2.adaptiveThreshold(im,255, cv2.ADAPTIVE_THRESH_GAUSSIAN_C, cv2.THRESH_BINARY, 115,1)
        im_threshold = cv2.bitwise_and(im,im,mask=gaus)
        self.Display_Image(im_threshold)

    def Filter_Otsu(self):  #work only w 1-ch image
        from skimage.viewer import ImageViewer
        im = self.imScn
        if len(im.shape)>2 and im.shape[2]==3:    #if color image
            im = cv2.cvtColor(im,cv2.COLOR_BGR2GRAY)                 
            #messagebox.showerror(title='Error',message='Otsu filter works only with grayscale image. Please load 1-ch image.')
        
        # Otsu's thresholding after Gaussian filtering
        blur = cv2.GaussianBlur(im,(5,5),0)
        ret,otsu = cv2.threshold(blur,0,255,cv2.THRESH_BINARY+cv2.THRESH_OTSU)
        im_threshold = cv2.bitwise_and(im,im,mask=~otsu)
        #cv2.imshow('im_threshold',im_threshold), cv2.waitKey(0)
        
        #otsu = sk.filters.threshold_otsu(im)
        #im_threshold = im <= otsu
        #plt.imshow(im_threshold, cmap='gray', interpolation='nearest')
        #plt.show()
        #im[im<=otsu] = 0
        #im[im>180] = 0
        #im[im>otsu] = 255
        ##cv2.imshow('im',im), cv2.waitKey(0)
        #self.Display_Image(im)
        #im_threshold[im_threshold<(255*0.3)] = 0    # remove saturated px if NDVI<0.3
        im_threshold[im_threshold>(255*0.8)] = 0    # remove saturated px if NDVI>0.8
        im_threshold[im_threshold>0] = 255
        self.Display_Image(im_threshold)

    def Filter_GausFiltered(self):
        im = self.imScn
        if len(im.shape)>2 and im.shape[2]==3:    #if color image
            im = cv2.cvtColor(im,cv2.COLOR_BGR2GRAY)                         
        gaus = cv2.adaptiveThreshold(im,255, cv2.ADAPTIVE_THRESH_GAUSSIAN_C, cv2.THRESH_BINARY, 115,1)        
        gausFiltered = cv2.GaussianBlur(gaus,(5,5),0)    #Gausian filtering
        im_threshold = cv2.bitwise_and(im,im,mask=gausFiltered)        
        self.Display_Image(im_threshold)

    def Proc_Calib(self):     
        if self.ROI_var == 0:
            messagebox.showwarning(title='Warning',message='Select ROI first!')
            return

        # find the target reflectance (c_REFr) from dialog
        obj = Dialog_CalibTarget(self.root,self.iCal,self.iCal_Ent)   
        self.iCal = obj.iCal
        self.iCal_Ent = obj.iCal_Ent
        self.fields= obj.fields
        self.iCal_Seg = obj.iCal_Seg
        print('iCal=',self.iCal,'iCal_Ent=',self.iCal_Ent,'iCal_Seg=',self.iCal_Seg)
        if obj.cancel: #close if canceled
            return
        if self.iCal==2:
            if len(self.iCal_Ent.split(','))==1: c_REFr = float(self.iCal_Ent)
            else:
                Ent = self.iCal_Ent.split(',')
                c_REFr = [0]*len(Ent)
                for i in range(len(Ent)): c_REFr[i] = float(Ent[i])
        else:
            c_REFr = float(self.fields[self.iCal].replace('%',''))                
        if self.iCal<2 or len(self.iCal_Ent.split(','))==1: 
            if c_REFr == 0: c_REFr = [3.5855979,3.6552659,3.7512322,3.935356]   #4% target (BGRN)
            else:          c_REFr = [c_REFr,c_REFr,c_REFr,c_REFr]   # % reflectance RGBN
            
        # read the target reflectance (c_REFr) from file
        fdir,fname,fbasename,ftype = ExtractFilename(self.src_file)
        cal_dir = FindParentDir(fdir,1)
        src_filename = '%sCalib/%s_cal.txt'%(cal_dir,fbasename) #'20170511'                
        #print('src_filename=',src_filename)
        #if os.path.exists(src_filename):
            #CF = read_CalF4HS(src_filename)
        #c_REFr = [10,20,30,40]
        #c_REFr = (48,48,48,48)

        self.ROI_var.set(1) # get the ROI selected to read the calib target
            
        # get the ROI resized to fit the image loaded if oversized ROI
        print('self.ROI=',self.ROI)
        print(self.ROI[len(self.ROI)-1][2],self.ROI[len(self.ROI)-1][3])
        if self.ROI[len(self.ROI)-1][2]>self.imW or self.ROI[len(self.ROI)-1][3]>self.imH:
            self.btn_ROI()  

        #ROI = (self.ROI[0][0], self.ROI[0][1], self.ROI[len(self.ROI)-1][2], self.ROI[len(self.ROI)-1][3])
        self.c_REFr = c_REFr

        bdry = self.bdry
        if not obj.iCal_Seg: self.reset_sliders([[0,255],[0,255],[0,255],[0,255]])  # Reset thresholding values if unchecked
        self.Calib(self.imScn_base, c_REFr, GSD=1)
        if not obj.iCal_Seg: self.reset_sliders(bdry)   # Retrieve the thresholding values         
        
        # retrieve the ROI button unchecked
        self.ROI_var.set(0), self.btn_ROI()
  
    # Calibrate the image based on ROI on known reflectance target per channel
    def Calib(self, im, c_REFr, GSD, flag_batch=None):  
        #print('self.GRID=',self.GRID,'self.ROI=',self.ROI)
        scaleup = 1 #scale up the brightness in RGB visualization
        #print('self.imScn_base=',self.imScn_base.shape,self.imScn_base.dtype)
        # calculate ROI (reflectance target) metrics by Seg()
        self.Img_Seg(flag_segStats=True,flag_batch=True)    # skip display if batch
        print('GRID_ave=',self.GRID_ave,'GRID_Cnt=',self.GRID_Cnt)
        if self.im_nband == 4: print('ROI_ndvi=',self.ROI_ndvi)
        
        CF = [1,1,1,1]   # calib factor
        REFr = [0,0,0,0]
        flag_nan = False
        for i in range(self.im_nband):
            if np.isnan(self.GRID_ave[i]): 
                flag_nan = True
                break
            else:
                REFr[i] = self.GRID_ave[i]*self.im_scale/255.0*100    # ref%
                if REFr[i]==0: REFr[i]=c_REFr[i]  # filter the denominator = 0 or inf -> 0.1
                CF[i] = c_REFr[i]/REFr[i]  # R,G,B,N    
            #print('REFr[%d]=%.3f%%'%(i,REFr[i]),'CF[%d]=%.3f'%(i,CF[i]))
        print('c_REFr=',c_REFr,'REFr=',REFr,'CF=',CF)
        
        #max_CF = 0
        #for k in range(self.im_nband): max_CF = max(max_CF,CF[k])
        ##print('nband=',self.im_nband,'max_CF=',max_CF)
        
        #fdir,fname,fbasename,ftype = ExtractFilename(self.src_file)
        #dst_dir = fdir+'Calib/'
        #if not os.path.exists(dst_dir): os.makedirs(dst_dir)                
        #outFile = dst_dir+fbasename+'_cal.txt'

        ##save Calibration factors into a file
        #with open(outFile,'w') as f:
            #f.write('Calibration factors on %s%% target for %s\n'%(c_REFr,self.src_file))
            #for k in range(self.im_nband): 
                #f.write('ch%d: %f\n'%(k,CF[k]))
            #f.write('max: %f\n'%(max_CF))
        
        ## save LAI & ha into an output file
        #Fgnd = self.GRID_Cnt
        #Fgnd_ha = Fgnd*GSD*GSD/(100*100)    # ha = 100m x 100m, PS2 GSD=3.125m
        #PxCnt = self.imH*self.imW
        ##print('PxCnt=',PxCnt,np.count_nonzero(im))
        #if PxCnt>0: LAI = Fgnd*100.0/PxCnt
        #else:       LAI = 0
            
        #Calibrate the image
        im_Cal = np.zeros((self.imH,self.imW,self.im_nband),dtype=float)   # must be float to scale the raw data properly, generate another buffer
        #print('im=',im.shape,im.dtype,np.nanmean(im),np.nanmax(im),'im_Cal=',im_Cal.dtype,end='->')
        #stats = [[0 for j in range(4)] for i in range(self.im_nband)]
        for k in range(self.im_nband):  # R,G,B,N
            #stats[k] = self.ds.GetRasterBand(k+1).ComputeStatistics(True)   #min, max, mean, stdev
            im_Cal[:,:,k] = im[:,:,k]*CF[k]
            #print('%.3f:%d~%d'%(CF[k],np.nanmin(im_Cal),np.nanmax(im_Cal)))
        #print(im_Cal.dtype,np.nanmean(im_Cal),np.nanmax(im_Cal))
        
        
        ##save the calibrated raster image                
        ##print('im=%s,%.1f,%d'%(im.dtype,np.nanmean(im),np.nanmax(im)),'->im_Cal=%s,%.1f,%d'%(im_Cal.dtype,np.nanmean(im_Cal),np.nanmax(im_Cal)),end='->')
        #if self.im.dtype!='uint8' and flag_nan==False:  
            #Save_GeoTiff(im_Cal.astype(self.im.dtype),self.src_file, dst_dir+fbasename+'_cal.tif') # save im_Cal as in the original image type        
        
        # stretch & clip the calibrated image to 0~255 for display
        #im_Cal = im_Cal   # convert px values to 8-bit(0~255) scale
        im_Cal[im_Cal > 255] = 255  # clip the saturated px -> white
        im_Cal[im_Cal < 0] = 0  # clip the saturated px -> black
        im_Cal = im_Cal.astype(np.uint8)    # ->uint8 to display
        
        #if self.im.dtype=='uint8' and flag_nan==False:
            #dst_file = dst_dir+fbasename+'_cal.tif'
            #Save_GeoTiff(im_Cal,self.src_file,dst_file) # save im_Cal as in the original image type        
        
        # display the calibrated image, im_Cal
        #print('%s,%.1f,%d'%(im_Cal.dtype,np.nanmean(im_Cal),np.nanmax(im_Cal)))
        #cv2.imshow('im_Cal',im_Cal), cv2.waitKey(0)
        #print('0SEG_var.get()=',self.SEG_var.get())
        if flag_nan==False and flag_batch==None: # skip display if batch
            self.Display_Image(im_Cal) 
        #print('1SEG_var.get()=',self.SEG_var.get())
        return im_Cal #,dst_dir
        
    def Proc_LoadMask(self):    
        self.MASK_var.set(1)
        self.btn_MASK()

    def Raster_BandMath(self):     
        # find the targer reflectance (c_REFr) from dialog        
        if self.flag_im: self.Bands_Math[0] = self.src_file
        fdir,fname,fbasename,ftype = ExtractFilename(self.src_file)
        obj = Dialog_BandMath(self.root,self.eqBM,self.eqBM_Ent,self.Bands_Math,fdir)   
        self.eqBM = obj.eqBM
        self.eqBM_Ent = obj.eqBM_Ent
        self.Bands_Math = obj.Bands
        self.fields= obj.fields
        print('eqBM=',self.eqBM,'eqBM_Ent=',self.eqBM_Ent,'Bands_Math=',self.Bands_Math)
        if obj.cancel: #close if canceled
            return
        if self.eqBM==2: c_BM = self.eqBM_Ent
        else:            c_BM = self.fields[self.eqBM]
            
        # calculate band math
        print('c_BM=',c_BM)
        if self.eqBM == 0: 
            im_b1,ds = LoadGeoTiff(self.Bands_Math[0]) 
            del ds
            im_b2,ds = LoadGeoTiff(self.Bands_Math[1])         
            del ds            

            # filter if 2 images are in same size
            print('im_b1.shape=',im_b1.shape,'im_b2.shape=',im_b2.shape)
            if im_b1.shape == im_b2.shape: 
                #imNDVI = (im_b1.astype(np.float32)-im_b2.astype(np.float32))/(im_b1.astype(np.float32)+im_b2.astype(np.float32))
                imNDVI = Generate_NDVI(im_b1,im_b2) # (b1-b2)/(b1+b2)
                print('im_b1.shape=',im_b1.shape,im_b1.dtype,'imNDVI.shape=',imNDVI.shape,imNDVI.dtype)            
                #if np.nanmin(imNDVI)<0:
                    #imNDVI = (imNDVI - np.nanmin(imNDVI)) / (np.nanmax(imNDVI) - np.nanmin(imNDVI)) # change the range from -1~1 to 0~1
                #imNDVI[imNDVI<0] = np.nan   # filter negative NDVI pixels to nan (not calculated in mean, but changing color in display)
                fdir,fname,fbasename,ftype = ExtractFilename(self.src_file)
                Save_GeoTiff(imNDVI,self.src_file,fdir+'ndvi_float32.tif')
                self.imScn_base = (255*imNDVI).astype(np.uint8) # (b1-b2)/(b1+b2)
            
        elif self.eqBM == 1: self.imScn_base = (255*Generate_NDVI(self.im[:,:,2],self.im[:,:,0])).astype(np.uint8) # (B-R)/(B+R)
        elif self.eqBM==2: self.imScn_base = (255*Generate_NDVI(self.im[:,:,1],self.im[:,:,0])).astype(np.uint8) # (G-R)/(G+R)
        elif self.eqBM==3: self.imScn_base = (255*Generate_NDVI(self.im[:,:,2],self.im[:,:,1])).astype(np.uint8) # (B-G)/(B+G), TO-DO: parse the string to equation
        
        self.im_nband = 1
        self.ch='NDVI'
        self.Band_var.set(None) # deselect all Band Radiobuttons
        if self.Stretch_var.get():
            print('self.stretch=',self.stretch)
            self.imScn_base = self.Update_StretchVar(imNDVI,self.stretch)
        self.Display_Image(self.imScn_base)            

    def Raster_Merge(self):     
        # find the targer reflectance (c_REFr) from dialog        
        if self.flag_im: self.Bands_Merge[0] = self.src_file
        fdir,fname,fbasename,ftype = ExtractFilename(self.src_file)
        obj = Dialog_Merge(self.root,self.Bands_Merge,fdir)   
        self.Bands_Merge = obj.Bands
        print('Bands_Merge=',self.Bands_Merge)
        if obj.cancel: #close if canceled
            return
        
        # detect no. of bands
        bands = 0
        for i,band in enumerate(self.Bands_Merge):
            print(i,band,band.split('.')[-1])
            if band.split('.')[-1] == 'tif' or band.split('.')[-1] == 'jpg':
                bands = bands + 1
        print ('bands=',bands)
        self.Bands_Merge = self.Bands_Merge[:bands]
        
        # Merge bands to a composite image
        ds = []
        IM = []
        #if bands>=1: 
        for i,band in enumerate(self.Bands_Merge):
            im0,ds0 = LoadGeoTiff(band)
            ds.append(ds0)
            IM.append(im0)    
        if IM[0].shape == IM[1].shape and IM[0].shape == IM[2].shape:
            im = np.array([ds[b].ReadAsArray() for b in range(bands)])
            im = np.transpose(im, [1,2,0]) # reorder dimensions, so that # of band is last                    
            print('im=',im.shape,im.dtype,np.nanmean(im),np.nanmin(im),np.nanmax(im))    
            if bands == 4:
                im = im[:,:,[2,1,0,3]]   # Image changed: BGRN -> RGBN 
            
            fdir,fname,fbasename,ftype = ExtractFilename(self.src_file)
            filename = fdir+'im_merged.tif'
            Save_GeoTiff(im,self.src_file,filename)
            
            # reload the merged image
            self.Load_Image(filename) # reset the self.im to the saved img
            
            ## display the merged image
            #self.master.title("IMAP"+'  - '+filename+' [%s,%s]'%(im.shape,im.dtype)) #change the title of master widget        
            #self.im = im 
            #self.im_nband = bands
            #if self.stretch[0]>=0 or im.dtype != 'uint8':   # Stretch options: 0 for Percentile, 1 for min/max, unit16->unit8 w percentile stretch
                #self.im_stretched = self.Update_StretchVar(im,self.stretch)                        
            #else: self.im_stretched = im
            #self.imScn_base = self.im_stretched    #baseimage of imScn before segmentation
            ##print('self.imScn_base=',self.imScn_base.shape,self.imScn_base.dtype,np.nanmean(self.imScn_base),np.nanmin(self.imScn_base),np.nanmax(self.imScn_base))    
            #self.Display_Image(self.imScn_base)  

        else: messagebox.showwarning(title='Warning',message='Load images in the same size!')

    def btn_MASK(self):    
        self.SEG_var.set(0)
        if(self.MASK_var.get() == 1):        
            #Load a mask image if not opened
            if len(self.mask_file) == 0 or kb.is_pressed("ctrl"): 
                self.mask_file = LoadFile(self,'img')
            #Check if a mask file is loaded
            if len(self.mask_file)==0: 
                self.MASK_var.set(0)
                messagebox.showwarning(title='Warning',message='Load a mask image first!')
                return        
            
            if   self.im_nband == 1: im = self.imScn_base   # 1b image
            elif self.im_nband == 2: im = 255-self.im*255 #b2(shadow):0(fgnd)->255,1(shadow)->0 
            else: 
                if self.B>=0: im = self.imScn_base[:,:,self.B]   #012:RGB
                else:         im = self.imScn_base   # 1b image
            
            mask = Get_Mask(self.mask_file)
            #Check if a mask file size = display image size
            if im.shape[0]!=mask.shape[0] or im.shape[1]!=mask.shape[1]: 
                self.MASK_var.set(0)
                messagebox.showwarning(title='Warning',message='%s!=%s: Load a same-sized mask image!'%(im.shape,mask.shape))
                return                    
            im_masked = Get_Masked(im,mask)
            self.Display_Image(im_masked)
        else:
            self.Display_Image(self.im_stretched)   
    
    #def btn_STAT(self):    
        #for i in range(self.im_nband):  #nband goes 1(B), 2(G), 3(R), 4(N)
            #self.Calculate_dsStat(self.ds,i)    #Display the img stat
                
    #Reset the thresholding bars and lines
    def btn_ResetThreshold(self):   
        bdry = [[0,255],[0,255],[0,255],[0,255]]
        self.reset_sliders(bdry)    
        if self.flag_ResetROI: 
            ROI_outer = [0,0,self.imW,self.imH,self.ROI[0][-1],self.ROI[0][-1]] 
            self.ROI = Update_ROIs(self.GRID, [ROI_outer], self.ROI_rot)
            #self.ROI = Draw_ROI(self.GRID, self.ROI, self.zf, self.frame.canvas, self.rot)
            (self.X1,self.Y1),(self.X2,self.Y2) = (self.ROI[0][:2], self.ROI[len(self.ROI)-1][2:4])

    # option to include ROI in reset button
    def btn_ResetROI(self):
        self.flag_ResetROI = self.ResetROI_var.get()        
            
    # option to strecth the image to percentile
    def btn_Stretch(self):
        if self.Stretch_var.get(): 
            self.btn_StretchMin()
        else:                      
            self.stretch=[-1,0,100,0,255]
            im_stretched = self.imScn_base 
        
        # calculate max_y for histogram display
        self.max_y = 0
        if self.ch == 'NDVI': im_nband = 1
        else: im_nband = self.im_nband
        for i in range(im_nband): 
            self.max_y = max(self.max_y,np.max(cv2.calcHist([self.imScn_base],[i],None,[256],[1,256]))) # 1: exclude saturated 0s
        
        self.Display_Image(im_stretched)  # im_stretched to fit display in uint8 format
            
    # option to strecth the image to Min/Max
    def btn_StretchMin(self):
        option = self.StretchMin_var.get()  # Stretch options: 0 for Percentile, 1 for min/max
        self.stretch=[option, float(self.entStretchL.get()), float(self.entStretchU.get()), float(self.entStretchMin.get()), float(self.entStretchMax.get())]
        im_stretched = Stretch_Image(self.im,self.stretch) # 16-bit img, buffer (x1)                
        
        # calculate max_y for histogram display
        self.max_y = 0
        if self.ch == 'NDVI': im_nband = 1
        else: im_nband = self.im_nband
        for i in range(im_nband): 
            self.max_y = max(self.max_y,np.max(cv2.calcHist([self.imScn_base],[i],None,[256],[1,256]))) # 1: exclude saturated 0s
        
        self.Display_Image(im_stretched)  # im_stretched to fit display in uint8 format

    # option to calibration in batch process
    def btn_BatchCal(self):
        self.flag_BatchCal = self.BatchCal_var.get()
        print('self.BatchCal_var.get()=',self.BatchCal_var.get())

    # option to clip to ROI in batch process
    def btn_BatchClip2ROI(self):
        self.flag_BatchClip2ROI = self.BatchClip2ROI_var.get()
        print('self.BatchClip2ROI_var.get()=',self.BatchClip2ROI_var.get())

    # option to save Heatmap image in batch process
    def btn_BatchSaveImgHeatmap(self):
        self.flag_BatchSaveImgHeatmap = self.BatchSaveImgHeatmap_var.get()
        print('self.BatchSaveImgHeatmap_var.get()=',self.BatchSaveImgHeatmap_var.get())

    # option to save HeatmapSeged image in batch process
    def btn_BatchSaveImgHeatmapSeged(self):
        self.flag_BatchSaveImgHeatmapSeged = self.BatchSaveImgHeatmapSeged_var.get()
        print('self.BatchSaveImgHeatmapSeged_var.get()=',self.BatchSaveImgHeatmapSeged_var.get())

    # option to save NDVI image in batch process
    def btn_BatchSaveImgNDVI(self):
        self.flag_BatchSaveImgNDVI = self.BatchSaveImgNDVI_var.get()
        print('self.BatchSaveImgNDVI_var.get()=',self.BatchSaveImgNDVI_var.get())

    # option to save mask image in batch process
    def btn_BatchSaveImgMask(self):
        self.flag_BatchSaveImgMask = self.BatchSaveImgMask_var.get()
        print('self.BatchSaveImgMask_var.get()=',self.BatchSaveImgMask_var.get())

    # option to strecth the heatmap to min/max
    def btn_HeatmapStretch(self):
        if self.HeatmapStretch_var.get(): 
            self.htstch=['yes', float(self.entHeatmapStretchL.get()), float(self.entHeatmapStretchU.get())]
        else: self.htstch=['no', float(self.entHeatmapStretchL.get()), float(self.entHeatmapStretchU.get())]

    def btn_HIST(self):    
        if self.flag_im: #if not segmentation
            self.Histogram() #do not update Histogram if doing segmentation            
                
    def btn_ZOOM(self):   
        self.zf = float(self.entZm.get())        
        # update MouseWheelCnt_Zm for next scroll count if manually typed
        self.MouseWheelCnt_Zm = 2-1/self.zf
        self.Img_Zoom() #Display the zoomed image
    
    def btn_ROTATE(self):   
        self.rot = float(self.entRot.get())
        self.MouseWheelCnt_Rot = round(self.rot)
        self.Img_Rotate()   #Display the rotated image
        
    #Display ROI on the image
    def btn_ROI(self):
        self.btn_ROI_var = self.ROI_var.get()
        #print ('btn_ROI=',self.ROI_var.get(),'GRID=',self.GRID,'ROI= %d,%d,%d,%d,%c'%(self.ROI[0][0],self.ROI[0][1],self.ROI[len(self.ROI)-1][2],self.ROI[len(self.ROI)-1][3],self.ROI[0][-1]))
        
        if(self.ROI_var.get() == 1):   
            self.GRID = [int(self.entR.get()),int(self.entC.get())]
            self.ROI = Update_ROI(self.GRID,self.ROI)   # creating new 2-corner ROIs
            
            if self.ROI[0][-1] == 'p':  # polygon
                for i in range(len(self.ROI)):
                    self.GRID = [1,1]
                    self.ROI = Draw_ROI(self.GRID, [self.ROI[i]], self.zf, self.frame.canvas, self.ROI_rot) 
            else:   
                self.ROI = Draw_ROI(self.GRID, self.ROI, self.zf, self.frame.canvas, self.ROI_rot) 
            #print('self.ROI=',self.ROI)
        else:
            self.frame.canvas.delete('all')
            self.frame.canvas.create_image(0,0,anchor="nw",image=img_photo)

    def Save_ROI(self,filename=None):
        if(filename==None):
            filename = tkinter.filedialog.asksaveasfilename(initialdir=self,title='Save as a text file',filetypes = (("txt files","*.txt"),("all files","*.*")))
            print('filename.find()=',filename.find('.txt'))                
            if filename.find('.txt') < 0: filename += '.txt' #add extension if not typed                
            print('filename=',filename)
        else:
            filename = os.getcwd()+'\Config\ROI.txt' 
        # exit if canceled to open a file
        if len(filename)==0:    
            return        
        
        print('filename=',filename, 'len(self.ROI))=',len(self.ROI))
        print('len(self.plots)=',len(self.plots))
        #print('ROI=',self.ROI)
        if self.georef == 1:
            self.plots = Convert_ROI2LL(self.ROI,self.ds,self.flag_im)         
        else:
            messagebox.showerror(title='Error',message='Load a geotiff image first!')
            self.plots = 0
            
        with open(filename,'w') as f:  
            f.write('%d, %d, %d, %s, %.2f, GRID, number of ROIs, ROI shape, ROI rot,NW_E,NW_N, NE_E,NE_N, SE_E,SE_N, SW_E,SW_N\n'%(self.GRID[0],self.GRID[1],len(self.ROI),self.ROI[0][8],self.ROI_rot))
            for i in range(len(self.ROI)):  # ROI: NW,SE,NE,SW
                f.write('%d, %d, %d, %d, %d, %d, %d, %d, %d'%(i,self.ROI[i][0],self.ROI[i][1],self.ROI[i][2],self.ROI[i][3],self.ROI[i][4],self.ROI[i][5],self.ROI[i][6],self.ROI[i][7]))
                if self.georef == 1:        # LL: NW,NE,SE,SW            
                    f.write(', %f, %f, %f, %f, %f, %f, %f, %f'%(self.plots[i][0][0],self.plots[i][0][1],self.plots[i][1][0],self.plots[i][1][1],self.plots[i][2][0],self.plots[i][2][1],self.plots[i][3][0],self.plots[i][3][1]))
                f.write('\n')
        return
        
    #Process images to extract ROI metrics 
    def Save_Metrics(self,filename,flag_batch=None,NDVI_CF=None): 
        #self.Read_Config()        
        fdir,fname,fbasename,ftype = ExtractFilename(self.src_file)
        
        with open(filename,'a+') as f: 
            if flag_batch==None:
                f.write(time.strftime('%m/%d/%y_%H:%M,'))
                f.write('%s%s\n'%(fdir,fname))                
                f.write('Threshoolding\n')          
                #print('self.im_nband=',self.im_nband,'min(self.im_nband,3)=',min(self.im_nband,3))
                for i in range(min(self.im_nband,3)):
                    if min(self.im_nband,3)==1: 
                        f.write('R:, %d, %d\n'%(self.sldL[i].get(),self.sldU[i].get()))
                    else:
                        f.write('%s:, %d, %d\n'%(self.ch[i],self.sldL[i].get(),self.sldU[i].get()))
                    self.bdry[i][0] = self.sldL[i].get()
                    self.bdry[i][1] = self.sldU[i].get()   
                f.write('GRID(RxC):,%d,%d\n'%(self.GRID[0],self.GRID[1]))            
                
                if self.ROI_var.get()==1:
                    f.write('ROI:,(%d,x %d), %d, %d, %d, %d, %c, rot=%.2f'%(self.GRID_W,self.GRID_H,self.ROI[0][0],self.ROI[0][1],self.ROI[len(self.ROI)-1][2], self.ROI[len(self.ROI)-1][3],self.ROI[0][-1],self.ROI_rot))
                else:
                    f.write('ROI:,(%d,x %d), %d, %d, %d, %d, %c'%(self.GRID_W,self.GRID_H,0,self.imW,0,self.imH,'r'))                
                
                try: 
                    f.write(',%.1f,%.1f\n'%(self.UTM[0][0],self.UTM[0][1]))
                except:     #NameError
                    f.write('\n')     
    
                f.write('ImgSize (WxHxb):,%d,%d,%d,%.1fMB,%.2f'%(self.imW,self.imH,self.im_nband,os.stat(self.src_file).st_size/1024/1024,self.GRID_CntP))
                for i in range(self.im_nband):                     
                    if self.im.dtype == 'float32' and np.nanmax(self.GRID_ave[i])<=1: # -1~1 px -> 0-100%
                        f.write(',%.2f'%(self.GRID_ave[i]*100))  
                    elif self.im.dtype == 'uint16': # 0~65024 px -> 0-100%
                        f.write(',%.2f'%(self.GRID_ave[i]*100/(255*255)))  
                    else: f.write(',%.2f'%(self.GRID_ave[i]*100/255))
                    
                if self.im_nband>=3: 
                    for i in range(3): f.write(',%.2f'%(self.GRID_HSV_ave[i]))
    
                # add column headings
                f.write('\nIter,Row,Col,cnt_Fgnd,cnt_ROI,%Canopy')        
                ch = ['R','G','B','N']
                for i in range(self.im_nband):                     
                    f.write(',%c [%%]'%ch[i])
                if self.im_nband>=3: f.write(',H,S,V')
                f.write(',NDVI\n')        
            
            # calculate a multiplier to % reflectance
            #print ('%.2f,'%(self.ROI_ave[r][i]*100/255),end=' ')
            if self.im.dtype == 'uint8': m_percent = 100/255    # 0-255 px -> 0-100%
            elif self.im.dtype == 'uint16': m_percent = 100/pow(2,16)    # 0-255 px -> 0-100%
            elif self.im.dtype == 'float32':
                if np.nanmax(self.ROI_ave[0][0])<=1: m_percent = 100    # -1~1 px -> 0-100%
                else: m_percent = 100/255    # 0~255 px -> 0-100%
            else:                        m_percent = 100
            print('m_percent=',m_percent)
            
            if self.ROI_var.get()==1: len_ROI = len(self.ROI)
            else:                     len_ROI = 1             
            for r in range(len_ROI):                      
                Row = int(r/self.GRID[1]) + 1
                Col = r%self.GRID[1] + 1
                
                # calculate & write LAI
                if self.ROI_var.get()==1: 
                    if self.ROI_rot == 0:
                        cnt_ROI = (self.ROI[r][2]-self.ROI[r][0])*(self.ROI[r][3]-self.ROI[r][1])
                        cntP_ROI = 100*self.ROI_Cnt[r]/cnt_ROI
                    else: 
                        cnt_ROI = self.ROI_CntIn[r]
                        cntP_ROI = self.ROI_CntP[r]
                else: 
                    cnt_ROI = self.GRID_W*self.GRID_H
                    cntP_ROI = 100*self.ROI_Cnt[r]/cnt_ROI
                    
                #print('%d,%d,%d,%d,%d,%.2f%%, Ave='%(r+1,Row,Col,self.ROI_Cnt[r],cnt_ROI,cntP_ROI),end=' ')
                if r==0: f.write('%d,%d,%d,%d,%d,%.2f'%(r+1,Row,Col,self.ROI_Cnt[r],cnt_ROI,cntP_ROI))
                else:    f.write('\n%d,%d,%d,%d,%d,%.2f'%(r+1,Row,Col,self.ROI_Cnt[r],cnt_ROI,cntP_ROI))
                
                # calculate & write Ref
                for i in range(self.im_nband):
                    f.write(',%.2f'%(self.ROI_ave[r][i]*m_percent))
                #print('')   #create a linefeed, cf)'\n' creates additional linefeed                            
                # calculate & write HSV if 3b img
                if self.im_nband >= 3:        
                    for i in range(3):
                        f.write(',%.2f'%(self.ROI_HSV_ave[r][i]))
                
                # calculate & write NDVI
                if self.im_nband>3:   
                    #print('ndvi=',self.ROI_ndvi[r])
                    f.write(',%.3f'%(self.ROI_ndvi[r]))
                elif self.im_nband == 1: # write normalized value if 1b image                    
                    f.write(',%.3f'%(self.ROI_ave[r][0]*m_percent/100))                    
                else:  # for 3b MAPIR RGN image
                    #print('self.ds.GetMetadata()=',self.ds.GetMetadata())
                    if self.maskSrc == 'RGN':
                        f.write(',%.3f'%((self.ROI_ave[r][2]-self.ROI_ave[r][0])/(self.ROI_ave[r][2]+self.ROI_ave[r][0])))
                    elif self.maskSrc == 'HSV' or self.maskSrc == 'Sat' or self.maskSrc == 'RGB': 
                        if NDVI_CF==None: NDVI_CF = 100 # Hue_full->NDVI: 80 for RDA_corn, 100 for RDA_wheat_oblique, 70 for RDA_wheat_nadir
                        f.write(',%.3f'%(self.ROI_HSV_ave[r][0]/NDVI_CF))  
                        
            else: f.write('\n')
    
    #Save ROI to config file
    def btn_SAVE(self):
        print('%d,%d,%d,%d,%c'%(self.ROI[0][0],self.ROI[0][1],self.ROI[0][2],self.ROI[0][3],self.ROI[0][-1]))
        
        #save the changed ROI to Config file                            
        self.Save_Config('filename')  
        self.Save_ROI('filename')  
        print('done')

    # options to include statistics in segmentation
    def btn_SegStats(self):
        self.flag_segStats = self.SegStats_var.get()
        #print('self.flag_segStats=',self.flag_segStats)            
        
    #Display the segmented image
    def btn_SEG(self,filterOpen=None,flag_batch=None,mask=None):
        self.MASK_var.set(0)
        #print('SEG_var.get()=',self.SEG_var.get())
        if(self.SEG_var.get() == 1):            
            #Update the image w seg
            if self.flag_im:     
                self.Img_Seg(self.flag_segStats,filterOpen,flag_batch,mask=mask)                
            else:
                messagebox.showwarning(title='Warning',message='Open an image first!')
                self.SEG_var.set(0)
        else:
            # calculate stats of the original image in ROI or no ROI
            self.GRID_ave = []
            self.GRID_CntP = 100
            if self.ROI_var.get()==1:
                im_GRID = self.im[self.Y1:self.Y2,self.X1:self.X2]  # im[H,W]
                #cv2.imshow('im_GRID',im_GRID), cv2.waitKey(0)
                for i in range(self.im_nband): 
                    #print(np.mean(im_GRID[:,:,i]))
                    self.GRID_ave.append(np.mean(im_GRID[:,:,i]))
                del im_GRID
            else:                
                self.GRID_ave = self.Ave
            
            #print('self.ch=',self.ch,'self.B=',self.B)
            if self.B == -1: # 1b selection: -1(none), 0(R), 1(G), 2(B)
                self.Display_Image(self.imScn_base)    # processing time: 6s for 166MB img
            else: self.btn_Band()
            
        #Display ROI statistics
        self.t0.delete('1.0',tk.END)        
        if self.flag_segStats and self.ROI_var.get()==1 and flag_batch==None: 
            self.Insert_ROIstats()

    def Insert_ROIstats(self):
        #if self.SEG_var.get()==1:
        self.t0.insert(tk.END,'ROI:')
        if self.im_nband==1:
            self.t0.insert(tk.INSERT,'[%.3f] %.1f%%'%(round(self.GRID_ave[0]),self.GRID_CntP))
        elif self.im_nband==3 and  len(self.imScn.shape)>2:
            self.t0.insert(tk.INSERT,'[%d,%d,%d] %.1f%%'%(round(self.GRID_ave[0]),round(self.GRID_ave[1]),round(self.GRID_ave[2]),self.GRID_CntP))
        elif self.im_nband==4:
            if len(self.GRID_ave)==1:
                self.t0.insert(tk.INSERT,'[%d] %.1f%%'%(round(self.GRID_ave[0]),self.GRID_CntP))
            elif len(self.GRID_ave)==4:
                self.t0.insert(tk.INSERT,'[%d,%d,%d,%d] %.1f%%'%(round(self.GRID_ave[0]),round(self.GRID_ave[1]),round(self.GRID_ave[2]),round(self.GRID_ave[3]),self.GRID_CntP))
        #else:
            #self.t0.insert(tk.INSERT,'[%d,%d,%d] %.1f%%'%(round(self.GRID_ave[0]),round(self.GRID_ave[1]),round(self.GRID_ave[2]),round(self.GRID_ave[3]),self.GRID_CntP))
        if(self.SEG_var.get() == 1):
            self.t0.insert(tk.INSERT,'(%d/%d)'%(self.GRID_Cnt,self.GRID_W*self.GRID_H))
        self.t0.place(rely=0.04, height=18)                        
    
    def Batch_FmCalendar(self):
        start_time = time.time()
        #Read config file   
        self.Read_Config()
        
        #self.stretch = [5.,95.]    #percentile of stretch range, [0,100]=[min,max]
        #print('self.stretch=',self.stretch)
        src_dir = tk.filedialog.askdirectory(parent=self.root,title='Open a folder to load images')
        fmC.GenerateVImapCalendar(src_dir) 
        #fmC.GenerateVImapCalendar(stretch=[0,2,98])    #Percentiles must be in the range [0, 100]
        #messagebox.showinfo(title='fmCalendar',message='Done')
        hh,mm,ss = ConvertSec2hhmmss(float(time.time()-start_time))
        print("Done! (run time = %02d:%02d:%02.1f seconds)"%(hh,mm,ss))
        return
 
    # Load and convert picture files to a video files
    def Batch_Pic2Video(self): 
        start_time = time.time()
        src_dir = tkinter.filedialog.askdirectory(parent=self,title='Open a folder to load images')
        if len(src_dir)==0:    
            return 
        
        zf = .25
        fps = 15    
        d = P2V.Dialog_Input4Vid(self.root)
        print('d.zf=',d.zf,'d.fps=',d.fps)
        if d.zf != 0 and d.fps !=0:
            zf = d.zf
            fps = d.fps          
            P2V.Pic2Video(src_dir,zf,fps)
        hh,mm,ss = ConvertSec2hhmmss(float(time.time()-start_time))
        print("Done! (run time = %02d:%02d:%02.1f seconds)"%(hh,mm,ss))
        
    # Process image to extract segmented metrics in whole image or ROI
    def Proc_Seg(self,ofilename=None,flag_batch=None,cnt_img=None): 
        start_time = time.time()
        
        # update boundary of grid if rotated ROI
        if self.ROI_rot==0: X1,Y1, X2,Y2 = self.X1,self.Y1, self.X2,self.Y2
        else:          
            # update ROIpx if changed or not defined
            #print('len(self.ROI)=',len(self.ROI),'len(self.ROIpx)=',len(self.ROIpx))
            #print('pre_ROI0=',self.pre_ROI0,'ROI[0]=',self.ROI[0])
            if len(self.ROI) != len(self.ROIpx) or self.pre_ROI0 != self.ROI[0]:
                self.ROIpx = [[] for i in range(len(self.ROI))] # pixels (i,j) in tROI
                self.GRID_CntIn = 0
                for i in range (len(self.ROI)):
                    self.ROIpx[i] = Extract_tROIpx(self.ROI[i])   # 3D array: [n-th ROI of ROIs], [n-th px in ROI], [i or j value] 
                    #print('i=',i,'len(self.ROIpx[i]=',len(self.ROIpx[i]))
                    self.GRID_CntIn = self.GRID_CntIn + len(self.ROIpx[i])    # total # of pixels in GRID                
                self.pre_ROI0 = self.ROI[0]
                #print('self.GRID_CntIn=',self.GRID_CntIn)
            # update 4-side boundary of GRID from ROIpx
            mask_tROI, OuterBoundary, InnerBoundary = Calculate_GRID_boundary(self.ROIpx,self.imW,self.imH)
            self.tX1, self.tY1, self.tX2, self.tY2 = OuterBoundary
            #print('self.tX1, self.tY1, self.tX2, self.tY2=',self.tX1, self.tY1, self.tX2, self.tY2)
            X1,Y1, X2,Y2 = self.tX1,self.tY1, self.tX2,self.tY2
            
            # update mask to exclude 4-piece triagles around tROI
            
        if(self.SEG_var.get() == 0): self.SEG_var.set(1)
        self.flag_segStats = 1
        self.SegStats_var.set(1)
        
        #update the filename if ROI for one of GRIDs as a CropID
        if self.ROI_var: CropID = self.cropID
        else: CropID = ''   
        if CropID == '': CropID = 'seg'
        print('CropID=',CropID)
        
        # create an output filename
        fdir,fname,fbasename,ftype = ExtractFilename(self.src_file)
        fdir = fdir+'Seg/'
        if ofilename == None:
            if not os.path.exists(fdir): os.makedirs(fdir) 
            ofilename = '%s%s%s.csv'%(fdir,fbasename,CropID)
            ofilename = fdir+fname.replace(fname.split('_')[-1],'%s.csv'%CropID)

        # detect if image is rotated, then process the rotated image
        if self.rot!=0: 
            im = imu.rotate_bound(self.im,self.rot)
        else: 
            im = np.copy(self.im)   # NOTE: not im = self.im due to change on im => change on self.im
        #print('self.im=',self.im.shape,self.im.dtype,'im=',im.shape,im.dtype)
        #im = self.imScn_base
        
        if self.flag_BatchClip2ROI: window_ROI = ((Y1,Y2),(X1,X2)) 
        else:                       window_ROI = ((0,im.shape[0]),(0,im.shape[1]))  #update ROI as a whole in case of rotated image
        #cv2.imshow('self.im',self.im), cv2.waitKey(0)        
        # apply radiometric calibration from ROI of a reference target
        if self.flag_BatchCal:                        
            bdry = self.bdry
            iCal_ROI = self.ROI            
            iCal_GRID = self.GRID
            if not self.iCal_Seg: 
                self.reset_sliders([[0,255],[0,255],[0,255],[0,255]])  # Reset ROI seg if unchecked
                self.ROI = [[563,742,572,750,self.ROI[0][-1]]]                
                self.GRID = [1,1]
            im = self.Calib(self.im, self.c_REFr, GSD=1, flag_batch=True)
            if not self.iCal_Seg: 
                self.reset_sliders(bdry)   # Retrieve the thresholding values 
                self.ROI = iCal_ROI                
                self.GRID = iCal_GRID
            #print('self.iCal_Seg=',self.iCal_Seg,'self.c_REFr=',self.c_REFr)            
            
            # save a calibrated image
            dst_file = fdir+fname.replace(fname.split('_')[-1],'%s_cal.tif'%CropID)
            Save_GeoTiff(im,self.src_file,dst_file,window_ROI)   
            
            # retrieve the btn_SEG
            if(self.SEG_var.get() == 0): self.SEG_var.set(1)                   
            self.im = im    # Assign the calibrated image as an input 8-bit image (self.im)              
        
        ROI = [X1, Y1, X2, Y2, self.ROI[0][-1]]
        
        # create a grey VI image and a mask image
        if self.maskSrc == 'RGN':
            # create a grey VI image
            im_grey = (255*Generate_NDVI(im[:,:,2],im[:,:,1])).astype(np.uint8)  # (B-G)/(B+G)
            
            # thresholding by Gaussian Otsu 
            blur = cv2.GaussianBlur(im_grey,(5,5),0)
            ret,otsu = cv2.threshold(blur,0,255,cv2.THRESH_BINARY+cv2.THRESH_OTSU)
            
            # create a mask image
            im_mask = cv2.bitwise_and(im_grey,im_grey,mask=otsu)
            im_mask[im_mask>(255*0.8)] = 0    # remove saturated px if NDVI>0.8
            im_mask[im_mask>0] = 255
            del blur,otsu                    
        else: # self.maskSrc == 'RGB'            
            #print('im=',im.shape,im.dtype,'np.amax(im)=',np.amax(im),np.max(im),np.nanmax(im))
            #if im.dtype == 'uint16': im = self.im_stretched #(im/256).astype('uint8')
            #elif im.dtype == 'float32': im = self.im_stretched 

            # thresholding by HSV
            loB = np.array([self.bdry[0][0],self.bdry[1][0],self.bdry[2][0],self.bdry[3][0]])
            upB = np.array([self.bdry[0][1],self.bdry[1][1],self.bdry[2][1],self.bdry[3][1]])                        
            
            # create a grey VI image & mask image
            if self.im_nband==1:  
                im_grey = im
                loB = np.array([loB[0]])
                upB = np.array([upB[0]])
                #print('imScn_base=',self.imScn_base.shape,self.imScn_base.dtype,np.nanmean(self.imScn_base),np.nanmin(self.imScn_base),np.nanmax(self.imScn_base))
                im_base = self.imScn_base
            else:
                if self.maskSrc=='RGB':                  
                    im_base = cv2.cvtColor(self.im_stretched[:,:,:3],cv2.COLOR_RGB2HSV_FULL)  # RGB->HSV 
                    im_grey = im_base[:,:,0]  # HSV->Hue                     
                    #print('imScn_base=',self.imScn_base.shape,self.imScn_base.dtype,np.nanmean(self.imScn_base),np.nanmin(self.imScn_base),np.nanmax(self.imScn_base))
                    #print('im=',im.shape,im.dtype,np.nanmean(im),np.nanmin(im),np.nanmax(im))
                    im_base = self.imScn_base
                elif self.maskSrc=='HSV' or self.maskSrc=='Sat':
                    im_base = cv2.cvtColor(self.im_stretched[:,:,:3],cv2.COLOR_RGB2HSV_FULL)  # RGB->HSV                                  
                    im_grey = im_base[:,:,0]  # HSV->Hue
                    if self.maskSrc=='Sat': im_grey = im_base[:,:,1]  # HSV->Sat                          
                elif self.maskSrc=='LAB':
                    im_base = cv2.cvtColor(self.im_stretched[:,:,:3],cv2.COLOR_RGB2LAB)  # RGB->LAB
                    im_grey = im_base[:,:,0]  # LAB->L                               
                elif self.maskSrc=='NDVI':
                    im_grey = Generate_NDVI(im[:,:,3],im[:,:,0]) # R,G,B,N =im[:,:,0],im[:,:,1],im[:,:,2],im[:,:,3]
                    if self.stretch[0]>=0 or im.dtype != 'uint8':   # Stretch options: 0 for Percentile, 1 for min/max, unit16->unit8 w percentile stretch
                        im_grey = self.Update_StretchVar(im_grey,self.stretch)                        
                    #print('im_grey=',im_grey.shape,im_grey.dtype,np.nanmean(im_grey),np.nanmin(im_grey),np.nanmax(im_grey))                                                            
                    im_base = im_grey
                if self.im_nband==4:
                    im_grey = Generate_NDVI(im[:,:,3],im[:,:,0])  # float32
            #print('im_base=',im_base.shape,im_base.dtype,np.nanmean(im_base),np.nanmin(im_base),np.nanmax(im_base))
            if len(im_base.shape)>2: ch = im_base.shape[2]
            else: ch = 1
            im_mask= cv2.inRange(im_base,loB[:ch],upB[:ch])
            del im_base
                
            ## create a grey VI image
            #im_HSV = None
            #if self.im_nband==1:                 
                #im_grey = im
                #loB = np.array([loB[0]])
                #upB = np.array([upB[0]])
            #else:             
                #im_HSV = cv2.cvtColor(self.im_stretched[:,:,:3],cv2.COLOR_RGB2HSV_FULL)  # RGB->HSV                                  
                #if self.im_nband == 3:
                    #if self.maskSrc=='Sat': im_grey = im_HSV[:,:,1]  # HSV->Sat 
                    #else:                   im_grey = im_HSV[:,:,0]  # HSV->Hue            
                #elif self.im_nband==4:
                    ##im_grey = (255*Generate_NDVI(im[:,:,3],im[:,:,0])).astype(np.uint8)  # int8
                    #im_grey = Generate_NDVI(im[:,:,3],im[:,:,0])  # float32
            
            ## create a mask image
            ##print('loB=',loB,'upB=',upB)            
            #if self.im_nband>=3 and (self.maskSrc=='HSV' or self.maskSrc=='Sat'): 
                #im_mask= cv2.inRange(im_HSV,loB[:3],upB[:3])
            #elif self.im_nband>=3 and self.maskSrc=='NDVI':
                ## create a grey VI image
                #if len(im.shape)>2 and im.shape[2]>3: 
                    #im_ndvi = Generate_NDVI(im[:,:,3],im[:,:,0]) # R,G,B,N =im[:,:,0],im[:,:,1],im[:,:,2],im[:,:,3]
                    #if self.stretch[0]>=0 or im.dtype != 'uint8':   # Stretch options: 0 for Percentile, 1 for min/max, unit16->unit8 w percentile stretch
                        #im_ndvi = self.Update_StretchVar(im_ndvi,self.stretch)                        
                    #print('im_ndvi=',im_ndvi.shape,im_ndvi.dtype,np.nanmean(im_ndvi),np.nanmin(im_ndvi),np.nanmax(im_ndvi))                                                            
                #im_mask= cv2.inRange(im_ndvi,loB[:1],upB[:1])   # masking based on 8-bit imScn_base image                
            #else:                    
                #print('imScn_base=',self.imScn_base.shape,self.imScn_base.dtype,np.nanmean(self.imScn_base),np.nanmin(self.imScn_base),np.nanmax(self.imScn_base))
                #im_mask= cv2.inRange(self.imScn_base,loB,upB)   # masking based on 8-bit imScn_base image
            #del im_HSV, im_LAB        
            
            #print('im_mask.dtype=',im_mask.dtype,'im_mask.shape=',im_mask.shape,np.nanmean(im_mask),np.nanmin(im_mask),np.nanmax(im_mask))
            
            # filter if ROI is bigger than img size for RDA images
            if self.ROI[len(self.ROI)-1][2]>self.imW:             
                self.ROI[len(self.ROI)-1][2] = self.imW-1
            if self.ROI[len(self.ROI)-1][3]>self.imH: 
                self.ROI[len(self.ROI)-1][3] = self.imH-1                  
        
        # apply filtering
        if self.filterSrc=='Opening': im_mask = cv2.morphologyEx(im_mask,cv2.MORPH_OPEN,np.ones((2,2))) # Opening(erode+dilate) to fill frgnd holes
        elif self.filterSrc=='Closing': im_mask = cv2.morphologyEx(im_mask,cv2.MORPH_CLOSE,np.ones((2,2))) # Closing(dilate+erode) to remove bkgnd particles        

        #cv2.imshow('im_mask',im_mask), cv2.waitKey(0)
        # rotote im_mask if image is rotated
        #print('im_grey.shape=',im_grey.shape[:2],'im_mask.shape',im_mask.shape[:2])
        if im_grey.shape[:2] == im_mask.shape[:2]: im_mask = imu.rotate_bound(im_mask,self.rot)                
        self.btn_SEG(filterOpen=2,flag_batch=1,mask=im_mask)
        
        # save a raw image with a grid
        if self.flag_BatchClip2ROI: 
            dst_file = fdir+fname.replace(fname.split('_')[-1],'%s_clip.tif'%CropID)
            #print('self.im.shape=',self.im.shape,'self.im.dtype=',self.im.dtype,'window_ROI=',window_ROI)
            #if(self.rot==0):
                #if len(im.shape)>2 and im.shape[2]==3: grid_color = (255,255,0)
                #else:                                  grid_color = (0,0,0)
                #im = Add_Grid(im,ROI,self.GRID,color=grid_color,LineSize=2,rot=self.ROI_rot) # add a grid on the heatmap
                #print('im=',im.shape,im.dtype)
                
            if self.ROI_rot!=0: im[~mask_tROI] = 255
            if window_ROI != None and self.flag_georef: 
                Save_GeoTiff(im,self.src_file,dst_file,window_ROI)            
        
        # save a heatmap with a grid
        heatmap = self.Img_Select(im=im_grey,ch='Heatmap')   # modify to feed im
        #print('ROI=',ROI)
        #print('self.ROI=',self.ROI)
        #print('self.ROI=',self.ROI[0][:2],self.ROI[-1][2:4])
        if self.ROI_rot!=0 and self.flag_BatchClip2ROI: 
            ROI = [self.ROI[0][0],self.ROI[0][1],self.ROI[-1][2],self.ROI[-1][3],self.ROI[0][-1]]
        heatmap = Add_Grid(heatmap,ROI,self.GRID,color=(255,255,0),LineSize=2,rot=self.ROI_rot) # add a grid on the heatmap
        if self.flag_BatchSaveImgHeatmap:             
            dst_file = fdir+fname.replace(fname.split('_')[-1],'%s_heatmap.tif'%CropID)
            if self.ROI_rot!=0 and self.flag_BatchClip2ROI: heatmap[~mask_tROI] = 255
            if self.flag_georef: Save_GeoTiff(heatmap,self.src_file,dst_file,window_ROI)

        # save a grey    
        if self.flag_BatchSaveImgNDVI and self.im_nband>=3: 
            dst_file = fdir+fname.replace(fname.split('_')[-1],'%s_ndvi.tif'%CropID)        
            if self.im_nband>=4:
                im_grey = Generate_NDVI(im[:,:,3],im[:,:,0]) # R,G,B,N =im[:,:,0],im[:,:,1],im[:,:,2],im[:,:,3]
                if self.ROI_rot!=0: im_grey[~mask_tROI] = np.NaN            
            if self.flag_georef: Save_GeoTiff(im_grey,self.src_file,dst_file,window_ROI)
        del im_grey,im

        # save a masked heatmap with a grid                
        if self.flag_BatchSaveImgHeatmapSeged:             
            dst_file = fdir+fname.replace(fname.split('_')[-1],'%s_heatmap_seged.tif'%CropID)
            heatmap_masked = Get_Masked(heatmap,im_mask) # create a masked heatmap                    
            heatmap_masked = Add_Grid(heatmap_masked,ROI,self.GRID,color=(255,255,0),LineSize=2,rot=self.ROI_rot) # add a grid on the masked heatmap                    
            if self.ROI_rot!=0: heatmap_masked[~mask_tROI] = 255
            if self.flag_georef: Save_GeoTiff(heatmap_masked,self.src_file,dst_file,window_ROI)
            del heatmap_masked
            #cv2.imwrite(dst_file,cv2.cvtColor(heatmap_masked, cv2.COLOR_RGB2BGR)) # cv2:BGR
        del heatmap
        
        # save a mask        
        if self.flag_BatchSaveImgMask: 
            dst_file = fdir+fname.replace(fname.split('_')[-1],'%s_greyVI_mask.tif'%CropID)
            if self.ROI_rot!=0: im_mask[~mask_tROI] = 255
            if self.flag_georef: Save_GeoTiff(im_mask,self.src_file,dst_file,window_ROI)
        del im_mask


        # assign a Cal Factor for pseudo NDVI 
        NDVI_CF = 100    # Hue_full->NDVI: 80 for RDA_corn, 100 for RDA_wheat_oblique, 70 for RDA_wheat_nadir
        if fbasename.lower().find('nadir')>=0: NDVI_CF = 70
        elif fbasename.lower().find('oblique')>=0: NDVI_CF = 100
        
        if cnt_img==0 or cnt_img==None: self.Save_Metrics(ofilename,NDVI_CF=NDVI_CF)
        else:                           self.Save_Metrics(ofilename,flag_batch=True,NDVI_CF=NDVI_CF)
        
        hh,mm,ss = ConvertSec2hhmmss(float(time.time()-start_time))
        print("done (seg) (run time = %02d:%02d:%02.1f seconds)"%(hh,mm,ss))        
        
        return

    #Read Avg raw files to extract data and write to a file
    def Batch_Avg(self):         
        Proc4Avg(self)        

    # Remove invalid dark images acquied at night (17:00-07:00)
    def Batch_Valid(self): 
        start_time = time.time()
        src_dir = tkinter.filedialog.askdirectory(parent=self,title='Open a folder to load 4b images')         
        if src_dir == '': return
        
        # read image filenames and remove files at 17:00-07:00 
        cnt_img = cnt_del = 0
        for filename in os.listdir(src_dir): #read files in the folder
            fname = filename.lower()
            if fname.endswith('.jpg') or fname.endswith('.tif'):  #read jpg or tif image  
                self.master.title("IMAP"+'  - '+fname)        #change the title of master widget                                        
                
                hh = int(fname[10:12])
                mm = int(fname[12:14])
                print ('fname=',fname,'at hh:mm=%02d:%02d'%(hh,mm), end='')
                if hh>=17 or hh<7:  # remove images                    
                    print(' => Deleted!')
                    os.remove(src_dir+'/'+fname)
                    cnt_del = cnt_del + 1
                else: print('')
                cnt_img = cnt_img + 1
        hh,mm,ss = ConvertSec2hhmmss(float(time.time()-start_time))
        print("Done! (%d/%d images deleted. Run time = %02d:%02d:%02.1f seconds)" % (cnt_del,cnt_img,hh,mm,ss))
        
    #Process images (jpg or RGN tif) to calculate ROI metrics and NDVI 
    def Batch_Seg(self): 
        start_time = time.time()
        src_dir = tkinter.filedialog.askdirectory(parent=self,title='Open a folder to load 4b images')         
        if src_dir == '': return
        
        #Read config file   
        if self.flag_im==0: self.Read_Config()
        print('len(self.ROI)=',len(self.ROI))
                
        # create an output filename for cvs
        fdir = src_dir+'/Seg/'
        ofilename = '%s/%s_seg.csv'%(fdir,src_dir.split('/')[-1])
        #print('fdir=',fdir)
        if not os.path.exists(fdir): os.makedirs(fdir)        
        
        # detect a specific image type: RDA=iPhone X, RGN=MAPIR Survey3 Camera
        for i in range(len(list(self.ds.GetMetadata().keys()))): #'TIFFTAG_DATETIME', 'TIFFTAG_IMAGEDESCRIPTION', 'TIFFTAG_RESOLUTIONUNIT', 'TIFFTAG_SOFTWARE', 'TIFFTAG_XRESOLUTION', 'TIFFTAG_YRESOLUTION'
            if list(self.ds.GetMetadata().keys())[i]=='TIFFTAG_IMAGEDESCRIPTION' and list(self.ds.GetMetadata().values())[i]=='MAPIR Survey3 Camera':   # for MAPIR tif
                self.maskSrc = 'RGN'
                self.filterSrc = 'Opening'                
            elif list(self.ds.GetMetadata().keys())[i]=='EXIF_Model' and list(self.ds.GetMetadata().values())[i]=='Survey3W_RGN': # for MAPIR jpg
                self.maskSrc = 'RGN'
                self.filterSrc = 'Opening' 
            elif list(self.ds.GetMetadata().keys())[i]=='EXIF_Model' and list(self.ds.GetMetadata().values())[i]=='iPhone X': # for RDA corn jpg images
                self.maskSrc = 'HSV'   
                self.entHeatmapStretchL.delete(0,tk.END)
                self.entHeatmapStretchL.insert(0, 5)
                self.entHeatmapStretchU.delete(0,tk.END)
                self.entHeatmapStretchU.insert(0, 95)
                #self.flag_BatchClip2ROI = 0
                #self.BatchClip2ROI_var.set(0)

        # write header info to output csv file
        cnt_img = 0
        # enable the btn_SEG
        self.SEG_var.set(1)
        self.flag_segStats = 1
        self.SegStats_var.set(1)
        for filename in os.listdir(src_dir): #read files in the folder
            fname = filename.lower()
            if fname.endswith('.jpg') or fname.endswith('.tif'):  #read jpg or tif image                    
                print('fname=',fname)                                        
                self.master.title("IMAP"+'  - '+fname)        #change the title of master widget                        
                
                # load input images
                self.im = cv2.imread(src_dir+'/'+fname)[:,:,[2,1,0]]   # BGR (cv2 image) -> RGB (np image) for Img_Seg()\masking
                self.imH,self.imW = self.im.shape[0:2]
                #self.im = np.asarray(Image.open(src_dir+'/'+fname))     # for Img_Seg()\masking
                self.imScn_base = self.im   # for Img_Seg()\metrics
                self.src_file = src_dir+'/'+fname
                
                # apply masking using NDVI for MAPIR tif images
                self.Proc_Seg(ofilename=ofilename,flag_batch=True,cnt_img=cnt_img)
                del self.im,self.imScn_base

                cnt_img = cnt_img + 1
        hh,mm,ss = ConvertSec2hhmmss(float(time.time()-start_time))
        print("Done! (%d images, run time = %02d:%02d:%02.1f seconds)"%(cnt_img,hh,mm,ss))

    #Process images to calculate NDVI and ROI metrics 
    def Batch_Calib(self): 
        start_time = time.time()
        src_dir = tkinter.filedialog.askdirectory(parent=self,title='Open a folder to load 4b images')         
        if src_dir == '': return
        #src_dir = 'C:/Users/james.kim2/Documents/Research/Python/IMAP/Data/PS2/rgbn'
        #print('src_dir=',src_dir)
        
        #Read config file   
        if self.flag_im==0: self.Read_Config()
        print('len(self.ROI)=',len(self.ROI))
                
        # create an output filename for cvs
        outDir = src_dir+'/Calib/'
        #print('outDir=',outDir)
        if not os.path.exists(outDir):
            os.makedirs(outDir)        
        ofilename = '%s/NDVI_cal.csv'%(outDir)  
        
        # write header info to output csv file
        with open(ofilename,'w+') as f:            
            f.write('Threshoolding\n')
            ch = ['R','G','B']
            for i in range(3):
                f.write('%s:, %d, %d\n'%(ch[i],self.sldL[i].get(),self.sldU[i].get()))
                self.bdry[i][0] = self.sldL[i].get()
                self.bdry[i][1] = self.sldU[i].get()                  
            f.write('ROI:, %d, %d, %d, %d, %c'%(self.ROI[0][0],self.ROI[0][1],self.ROI[0][2],self.ROI[0][3],self.ROI[0][-1]))
            try: 
                f.write(',%.1f,%.1f\n'%(self.UTM[0][0],self.UTM[0][1]))
            except:     #NameError
                f.write('\n')                
            f.write('Filename,ImgSize,Fgnd_ROI,Fgnd[ha]_ROI,PxCnt_ROI,%Canopy_ROI,R,G,B,NIR,NDVI,NDVI_stdev,NDVIc,Hue,Sat,Val\n')        
        
        cnt_img,imCal_dir,imNDVI_dir,imcNDVI_dir = self.CALproc(src_dir,ofilename)   
        if imCal_dir!=None and self.im_nband==1: fmC.GenerateVImapCalendar(imCal_dir)   # 1b input image as NDVI
        if imNDVI_dir!=None: fmC.GenerateVImapCalendar(imNDVI_dir)  # 4b input image -> NDVI
        if imcNDVI_dir!=None: fmC.GenerateVImapCalendar(imcNDVI_dir)  # 4b input image -> NDVI 
        hh,mm,ss = ConvertSec2hhmmss(float(time.time()-start_time))
        print("Done! (%d images run time = %02d:%02d:%02.1f seconds)"%(cnt_img,hh,mm,ss))
        
    #Process images to calculate NDVI and ROI metrics 
    def CALproc(self,src_dir=None,ofilename=None): 
        if src_dir==None: 
            messagebox.showwarning(title='Warning',message='EnterOpen a folder to load images')
        cnt_img = 0
        imCal_dir = None
        imNDVI_dir = None
        imcNDVI_dir = None
        for filename in os.listdir(src_dir): #read files in the folder
            if filename.endswith('.tif'):  #read only NDVI image                    
                print('filename=',filename)                        
                fdir,fname,fbasename,ftype = ExtractFilename(src_dir+'/'+filename)
                #self.master.title("IMAP"+'  - '+fname)        #change the title of master widget                        
                
                # load input images
                self.Load_Image(src_dir+'/'+fname) #,display='Batch_ROIproc')                
                #print 'len(self.im)=',len(self.im)
                
                # get the ROI resized to fit the image loaded if oversized ROI
                self.ROI_var.set(1) # get the ROI selected to read the calib target
                #print(self.ROI[len(self.ROI)-1][2],self.ROI[len(self.ROI)-1][3])
                if self.ROI[len(self.ROI)-1][2]>self.imW or self.ROI[len(self.ROI)-1][3]>self.imH:
                    self.btn_ROI()  
                # calibrate the image and save 
                #ROI = (self.ROI[0][0], self.ROI[0][1], self.ROI[len(self.ROI)-1][2], self.ROI[len(self.ROI)-1][3])
                c_REFr = (10,20,30,40)
                #c_REFr = (48,48,48,48)
                im_Cal,imCal_dir = self.Calib(self.im, c_REFr, GSD=1)
                print('im_Cal=',im_Cal.shape,im_Cal.dtype,np.nanmean(im_Cal))
                
                # save LAI & ha into an output file
                print('self.ROI=',self.ROI)
                print('self.GRID_Cnt=',self.GRID_Cnt)
                Fgnd = self.GRID_Cnt
                Fgnd_ha = Fgnd*3.125*3.125/(100*100)    # ha = 100m x 100m	  
                PxCnt = self.GRID_H*self.GRID_W   # imH*self.imW
                #print('PxCnt=',PxCnt,np.count_nonzero(im))
                if PxCnt>0: LAI = Fgnd*100.0/PxCnt
                else:       LAI = 0                
                
                # save NDVI images if 4b img
                NDVI = NDVI_std = cNDVI = 0
                if self.im_nband == 4:
                    fdir,fname,fbasename,ftype = ExtractFilename(self.src_file)
                    
                    # save the NDVI images
                    im_NDVI = self.Img_Select(ch='NDVI')          
                    imNDVI_dir = fdir+'NDVI/'
                    if not os.path.exists(imNDVI_dir): os.makedirs(imNDVI_dir)        
                    dst_file = imNDVI_dir+fbasename+'_ndvi.'+ftype                                              
                    print('im_NDVI=',im_NDVI.shape,im_NDVI.dtype,np.nanmean(im_NDVI))
                    Save_GeoTiff(im_NDVI,self.src_file,dst_file)                 
                    NDVI = np.nanmean(im_NDVI)
                    NDVI_std = np.nanstd(im_NDVI)                       

                    # save the calibrated NDVI images
                    im_cNDVI = self.Img_Select(im=im_Cal,ch='NDVI')   # modify to feed im          
                    imcNDVI_dir = fdir+'NDVI_cal/'
                    if not os.path.exists(imcNDVI_dir): os.makedirs(imcNDVI_dir)        
                    dst_file = imcNDVI_dir+fbasename+'_ndvi_cal.'+ftype                
                    print('im_cNDVI=',im_cNDVI.shape,im_cNDVI.dtype,np.nanmean(im_cNDVI))
                    Save_GeoTiff(im_cNDVI,self.src_file,dst_file)                 
                    cNDVI = np.nanmean(im_cNDVI) 
                
                # calculate image metrics
                REFr = [0,0,0,0]
                for i in range(self.im_nband):
                    REFr[i] = self.GRID_ave[i]/255.0*100    # ref%                
                Hue = Sat = Val = 0                

                # save image metrics to csv
                with open(ofilename,'a') as f:
                    f.write('%s,'%(fname))
                    f.write('%d,%d,%.3f,%d,%.2f,'%(self.im.size,Fgnd,Fgnd_ha,PxCnt,LAI))
                    f.write('%.3f,%.3f,%.3f,%.3f,'%(REFr[0],REFr[1],REFr[2],REFr[3]))
                    f.write('%.3f,%.3f,%.3f,%.3f,%.3f,%.3f\n'%(NDVI,NDVI_std,cNDVI,Hue,Sat,Val))                
                #time.sleep(.1) #sec
                cnt_img = cnt_img+1    
        return cnt_img,imCal_dir,imNDVI_dir,imcNDVI_dir
        
    # Create a blank base image of GPS points
    def Create_Basemap(self,E_min,E_max,N_min,N_max):
        #print('flag_im_base=',self.flag_im_base)
        if self.flag_im_base==0 and self.flag_im==1:
            if self.UNIT == 'foot':
                sc = 1/(self.PxRol * 0.3048)   # NAD83:[ft] x 0.3048 = [m]
            else:
                sc = 1/(self.PxRol)   # WGS84: [m]
            #print('UNIT=',self.UNIT,'Pixel Resolution =',self.PxRol,'sc=',sc)        
            #im_base = self.im   # ERROR, causing a long proc time
            im_base = cv2.cvtColor(self.im,cv2.COLOR_BGRA2BGR)   # BGRA->BGR
    
            # convert baseimage EPSG to WGS84
            #print(self.minx,self.miny)
            Lat, Lon = Convert_UTM2LL(self.minx,self.miny,None,self.ds.GetProjection())  #LL of WGS 84
            #print(Lon, Lat)
            Lon_deg,Lat_deg,E_min,N_min = Convert_LL2UTM([Lat,Lon])    #UTM of WGS 84
            #print(self.minx,self.miny)
            Lat, Lon = Convert_UTM2LL(self.maxx,self.maxy,None,self.ds.GetProjection())  #LL of WGS 84
            #print(Lon, Lat)
            Lon_deg,Lat_deg,E_max,N_max = Convert_LL2UTM([Lat,Lon])    #UTM of WGS 84                
            #print(E_min,N_max)        
        else:           
            self.georef = 0 # TO-DO: make it georef
            self.PxRol = 0
            max_L = max(E_max-E_min,N_max-N_min)
            sc = 2000/max_L #scale factor = 10
            print('E_max-E_min=%.2f, N_max-N_min=%.2f, max_L=%.2f, sc=%.2f'%(E_max-E_min,N_max-N_min,max_L,sc))
            self.imW = round((E_max - E_min +2)*sc)+30
            self.imH = round((N_max - N_min +2)*sc)+30
            im_base = 255*np.ones((self.imH,self.imW,3),dtype=np.uint8)          
            self.flag_im_base = 1
            self.im = im_base
        return im_base,E_min,E_max,N_min,N_max,sc 

    # Locate Lat,Lon to ROI index
    def Locate_LL2ROI(self,bidir,pre_iR,pre_iC,pID,Easting=None,Northing=None,plots=None):
        #print('pID,Easting,Northing=',pID,Easting,Northing)
        if pre_iR==1 and pre_iC==1 and pID==1:            
            E = [0]*4
            N = [0]*4
            min_x = 100000
            min_y = 100000
            for i,plot in enumerate(plots):       
                for j in range(1):  # 0:NW, 1:NE, 2:SE, 3:SW
                    Lon_deg,Lat_deg,E[j],N[j] = Convert_LL2UTM([plot[j][1],plot[j][0]])    #UTM of WGS 84 
                    #print(i,j,plot[j][0],plot[j][1],E[j],N[j])
                min_x = min(abs(Easting-E[0]),min_x)
                min_y = min(abs(Northing-N[0]),min_y)            
                if abs(Easting-E[0]) == min_x: 
                    iC = i%self.GRID[1] + 1
                if abs(Northing-N[0]) == min_y:
                    iR = int(i/self.GRID[1]) + 1
            print(i,'iR x iC =',iR,iC,'min x,y=',min_x,min_y)
        else: 
            if pID == 1: 
                iC = pre_iC + 1
                if bidir == True: iR = pre_iR            
                else:             iR = 1
            else:        
                iC = pre_iC
                if bidir == True: 
                    if iC%2 == 0: iR = pre_iR - 1
                    else:         iR = pre_iR + 1
                else:             iR = pre_iR + 1
        #print('iR,iC=',iR,iC)
        return iR,iC    
    
    # update imScn to display as georeferenced
    def Update_imScn(self,E_min,E_max,N_min,N_max):
        self.im_nband = 3
        self.Ave = []
        self.Stdev = []
        self.Min = []
        self.Max = []        
        #print('self.PxRol=',self.PxRol)
        if self.PxRol != 0: # skip if no geotiff image opened
            for i in range(3):
                self.Calculate_imStat(self.imScn,i)
        self.minx = E_min
        self.miny = N_min
        self.maxx = E_max
        self.maxy = N_max     
        #print('minx=',self.minx,'miny=',self.miny,'maxx=',self.maxx,'maxy=',self.maxy)  

    # update Stretch variables
    def Update_StretchVar(self,im,stretchVar):
        # update min/max of the image pixels
        self.entStretchMin.delete(0,tk.END)
        self.entStretchMin.insert(0, np.nanmin(im))
        self.entStretchMax.delete(0,tk.END)
        self.entStretchMax.insert(0, np.nanmax(im))            
        
        im_stretched = Stretch_Image(im,stretchVar) # 16-bit img, buffer (x1)
        self.Stretch_var.set(1)
        self.StretchMin_var.set(0)
        #cv2.imshow('im_stretched',im_stretched),cv2.waitKey()
        print ('im_stretched.dtype=',im_stretched.dtype, 'im_stretched.shape=',im_stretched.shape)
        return im_stretched

    # load csv datafile to read GPS and metrics to plot
    def Layer_CSV(self, file=None, display=None):  
        #filename = "C:/Users/james.kim2/Documents/Research/Data/Avenger/2019_09_19_F004_Run11_DOY262/Run 004/DOY262_9_19_2019_AvengerDataCollection_11_09_AM/Avg.csv"
        if file == None: filename = LoadFile(self,'CSV')     
        else: filename = file
        #print('filename=',filename)
        # exit if canceled to open a file
        if len(filename)==0:    
            return        

        fdir,fname,fbasename,ftype = ExtractFilename(filename)
        #print('fbasename=',fbasename)
        
        # open a Ent&CB dialog to select an input column
        with open(filename) as f:
            col_lst= f.readline().split(',')
        #print('col_lst=',col_lst)
        cmap = [255,0,0]
        d = Dialog_Input4Layer(self.root,var1=col_lst,var2=cmap,var3=fbasename) 
        if d.markerSize == 0:   # if canceled
            return

        start_time = time.time()
        # check if geotiff is opened
        if d.plotmap == True: 
            pre_iR = pre_iC = 1
            plots = self.Convert_ROI2LL() 
            if plots == 0:
                return 0

        self.master.title("IMAP"+'  - '+filename)        #change the title of master widget        
        
        #Detect # of lines
        num_lines = sum(1 for line in open(filename))
        #print(num_lines,'filename=',filename)        
        
        ms = d.markerSize  # marker size of data point circle
        cmap = d.cmap
        print('idx=',d.idxL,d.idxH,'pct=',d.pct1,d.pct2,'stretch=',d.stretch,'col=',d.col,'connect=',d.connect,'cmap=',cmap,'ms=',ms)
        if fbasename.lower().find('gt')>=0:
            print('plotmap=',d.plotmap,'R1C1=',d.R1C1,'bidir=',d.bidir) # 'RowCSV=',d.RowCSV

        # detect the heatmap color
        if cmap[0]>=255 and cmap[1]==0 and cmap[2]==0: color = 'red'
        else:   color = 'blue'

        # load GPS boundaries of all fields
        filename4FD = path+'Config/Border/ALARC_FD_GPS.csv'
        FD_GPS = Read_FD_GPS(filename4FD) #(Lon, Lat)
        FIDs = list(FD_GPS.keys()) # field ID
        FDs = list(FD_GPS.values())   # 4-corner coords of field boundary
        #print('FD_GPS=',FD_GPS)
        print('FIDs=',FIDs)
        #print('FDs=',FDs)
        
        # calculate UTM of field boundary
        FD_UTM = []
        FDs_UTM = []
        for i in range(len(FDs)):
            for j in range(4):
                LL = (FDs[i][j][1], FDs[i][j][0])                        
                Lon_deg,Lat_deg,Easting,Northing = Convert_LL2UTM(LL)
                FD_UTM.append((Easting,Northing))
            #print(i,'FDs[][0]=',FDs[i][0],'plot=',plot) 
            FDs_UTM.append(FD_UTM) 
            FD_UTM = []
        #print('FDs_UTM=',FDs_UTM)
        FDs = FDs_UTM
        
        #Calcuate a canvas size to display the GPS points
        lst = [0]*num_lines        
        E_min = N_min = VI_min = 10000000
        E_max = N_max = VI_max = -10000000
        VI_arr = []
        col_R = col_C = col_Lat = col_Lon = col_E = col_N = col_Ea = col_Na = col_Yaw = -1
        #col_N = -1
        with open(filename) as f:
            FID = []
            pre_FID = -1
            for i, line in enumerate(f):
                lst[i] = line.split(',')
                if i==0:
                    for j in range(len(lst[i])):
                        if lst[i][j].lower() == 'row': col_R = j
                        if lst[i][j].lower() == 'col': col_C = j 
                        if lst[i][j].lower() == 'plot': col_plotID = j 
                        if lst[i][j].lower() == 'easting': col_E = j
                        if lst[i][j].lower() == 'northing': col_N = j 
                        if lst[i][j].lower().find('easting_arm')>=0: col_Ea = j
                        if lst[i][j].lower().find('northing_arm')>=0: col_Na = j 
                        if lst[i][j].lower().find('lon')>=0 or lst[i][j].lower().find('longitude')>=0: col_Lon = j 
                        if lst[i][j].lower().find('lat')>=0 or lst[i][j].lower().find('latitude')>=0: col_Lat = j 
                        if lst[i][j].lower().find('yaw')>=0: col_Yaw = j
                else:
                    if col_Ea>=0 and col_Na>=0:
                        Easting = float(lst[i][col_Ea])
                        Northing = float(lst[i][col_Na])
                    elif col_Lon>=0 and col_Lat>=0:
                        Lon_deg,Lat_deg,Easting,Northing = Convert_LL2UTM([float(lst[i][col_Lat]),float(lst[i][col_Lon])])    #UTM of WGS 84
                    else:
                        messagebox.showerror(title='Error',message='Cannot find GPS data in the file!')
                        break
                    #print(Lat,Lon,Easting,Northing)                    
                    E_min = min(E_min,Easting)
                    E_max = max(E_max,Easting)
                    N_min = min(N_min,Northing)
                    N_max = max(N_max,Northing)
                    
                    if lst[i][d.col].lstrip(' -').replace('.','',1).isdigit() == True:  # check for number, cf) ' -0.1'.lstrip('-') won't work
                        VI = float(lst[i][d.col])
                        VI_arr.append(float(lst[i][d.col]))
                    else: VI = VI_min                    
                    VI_min = min(VI_min,VI)
                    VI_max = max(VI_max,VI)  
                    
                    # locate a field from GPS coord in CSV 
                    if d.plotmap == True:
                        for j in range(len(FDs)):                         
                            if pre_FID!=j:
                                for k in range(4):                                                        
                                    if abs(FDs[j][k][0]-Easting)<10 and abs(FDs[j][k][1]-Northing)<10:
                                        #print(i,j,k,'Easting,Northing=',Easting,Northing,"FD=",FDs[j][k][0],FDs[j][k][1])
                                        FID.append([i,j,k])
                                        pre_FID = j
                        #FID = Locate_Field()
            if d.plotmap == True:                
                for i in range(len(FID)):
                    FD_name = FIDs[FID[i][1]]
                    FD_start = FID[i][0]
                    if FD_name.find('F004')>=0:
                        # TO-DO: read the following FD info from ROI file
                        n_plot = 27*24  # number of plot for each of wet & dry fields
                        n_BufferCol = 8 # number of beffer columns between wet & dry fields
                    print('FID=',FID,FID[i][0],FIDs[FID[i][1]])

                FID.clear()
                ## read ROI file from FD_name
                #print('filename=',filename,'fdir=',fdir,'fname=',fname)                    
                #filename4ROI = r"C:\Users\james.kim2\Documents\Research\Python\IMAP\Config\ROI\ROI_grid_%s.txt"%FD_name
                ##filename4ROI = fdir+fname+FD_name+'.txt'
                #GRID,ROI,ROI_rot,plots = Read_ROI(filename4ROI)
                #n_ROI = GRID[0] * GRID[1]
                #print('n_ROI=',n_ROI,ROI[0][0],ROI[0][1],ROI[len(ROI)-1][2],ROI[len(ROI)-1][3])
                    
            else: 
                #FD_name = FIDs[FID[i][1]]
                FD_start = 0 
        
        # filter VI
        #if VI_min < 0: VI_min = 0
        #if VI_max < 1: VI_max = 1                
        
        #print(Easting,Northing,E_min,E_max,N_min,N_max)
        #Create a blank base image of GPS points
        if d.plotmap == True:
            if self.georef == False:
                self.imW = self.ROI[len(self.ROI)-1][2] - self.ROI[0][2]
                self.imH = self.ROI[len(self.ROI)-1][3] - self.ROI[0][3]
                self.im_nband = 3
                self.im_base = 255*np.ones((self.imH,self.imW,self.im_nband),dtype=np.uint8)
            else:
                self.im_base,E_min,E_max,N_min,N_max,sc = self.Create_Basemap(E_min-2,E_max,N_min,N_max+2)   # 2-m margin
        else:
            self.im_base,E_min,E_max,N_min,N_max,sc = self.Create_Basemap(E_min-2,E_max,N_min,N_max+2)   # 2-m margin
            #print('UTM min & max=',E_min,E_max,N_min,N_max,'sc=',sc)
        #cv2.imshow('im_base',self.im_base), cv2.waitKey(0)      

        #print('W x H =',self.imW,self.imH)     
        if d.pct1==0 and d.pct2==100: VI_pcL, VI_pcH = VI_min, VI_max
        elif d.stretch == 0:          VI_pcL, VI_pcH = d.idxL, d.idxH
        else:                         VI_pcL, VI_pcH = np.nanpercentile(VI_arr, (d.pct1, d.pct2))
        VI_arr.clear()
        print('VI min & max=',VI_min,VI_max,'percentile=',VI_pcL,VI_pcH)        

        #self.pre_zf = self.zf = 1  #skip to properly zoom out the plot map
        
        # read a platform config file
        if fbasename.lower().find('avg')>=0:   
        #if fbasename=='Avg':   
            filename_dat = fdir + fbasename + '_meta.dat'
            #print('filename_dat=',filename_dat)
            CC2BB,yaw_off,a,ctr_off,arm_n,RS,Heading_n,GPS,IMU,CC,T,Tb,H,P,GPS_n,IMU_n,CC_n,T_n,Tb_n,H_n,P_n = Read_Platform4Avg(filename_dat)

            filename4JSON = r"C:\Users\james.kim2\Documents\Research\Python\IMAP\Config\Offset\sensor_offsets_Avg.json"
            CC,IRt,Pe,Ho = LoadJSON(self,filename4JSON) # offsets in meter        
            print('CC=',len(CC),CC)
            print('IRt=',len(IRt),IRt)
            print('Pe=',len(Pe),Pe)
            print('Ho=',len(Ho),Ho)    
        
        #Plot GPS points   
        #print('FD_start=',FD_start)
        for i in range(FD_start,num_lines):
            if i== 0:                
                if fbasename.lower().find('avg')>=0:
                    # read column headings and find thier a, b offsets
                    print('col_heading=',d.col,lst[0][d.col]) 
                    aa = [0]*arm_n
                    bb = [0]*arm_n
                    if lst[0][d.col][:2] == 'VI':
                        #print(lst[0][d.col][3])
                        for n in range(arm_n):
                            name = '%d%s'%(n+1,lst[0][d.col][3])  # 1b or 1f
                            bb[n] = CC[name][0]
                            aa[n] = CC[name][1]                
                    elif lst[0][d.col][:3] == 'Hon':
                        for n in range(arm_n):
                            name = '%d'%(n+1)
                            bb[n] = Ho[name][0]
                            aa[n] = Ho[name][1]
                    elif lst[0][d.col][:3] == 'IRT':
                        #print(lst[0][d.col][-8],lst[0][d.col][-8].lower())
                        for n in range(arm_n):
                            if lst[0][d.col][-4:] == 'rget': name = '%d%s'%(n+1,lst[0][d.col][-8].lower())  # 1n or 1f
                            if lst[0][d.col][-4:] == 'Body': name = '%d%s'%(n+1,lst[0][d.col][-6].lower())  # 1n or 1f
                            bb[n] = IRt[name][0]
                            aa[n] = IRt[name][1]
                    elif lst[0][d.col][:3] == 'Pep':
                        for n in range(arm_n):
                            name = '%d'%(n+1)
                            bb[n] = Pe[name][0]
                            aa[n] = Pe[name][1]
                    else:   # 12 VIs & IMU (y,p,r)
                        for n in range(arm_n):
                            name = '%df'%(n+1)
                            bb[n] = CC[name][0]
                            aa[n] = CC[name][1]                                    
                    print('aa=',aa,'bb=',bb)                
                #elif fbasename.lower().find('gt')>=0:
                    #if lst[0][j].lower() == 'lat':
                        
                    #if lst[0][j].lower() == 'lon':
                print (num_lines,end=" lines:")                                         
                 
            if i%(int(num_lines/10)) == 0: print("-",end="")                
            if i>0:
                # determine color map for the marker        
                #VI @35th col
                #print(i,'lst[i][d.col]=',lst[i][d.col], lst[i][d.col].lstrip(' -')) 
                if lst[i][d.col].lstrip(' -').replace('.','',1).isdigit() == True:  # check for number, cf) ' -0.1'.lstrip('-') won't work
                    #print(float(lst[i][d.col]))
                    # noise filtering VI< VI_percentile Lower
                    if VI_pcL<=float(lst[i][d.col]) and float(lst[i][d.col])<=VI_pcH: 
                        VI_gray = round((float(lst[i][d.col])-VI_pcL)/(VI_pcH-VI_pcL)*255)
                    elif float(lst[i][d.col])<VI_pcL: VI_gray = 0   # default
                    elif float(lst[i][d.col])>VI_pcH: VI_gray = 255 # default
                else:
                    VI_gray = -1    # filter non-numerics to White
                R,G,B = VI2Color(VI_gray,color)
                #print('VI_gray=',VI_gray,R,G,B,ms)
                
                # draw plot heatmap based on plot ID 
                if d.plotmap == True:
                    # find values within ROI
                    if i-FD_start < len(self.ROI)+1:                        
                        #print(int(lst[i][col_plotID]),float(lst[i][col_Lat]),float(lst[i][col_Lon]))
                        Lon_deg,Lat_deg,Easting,Northing = Convert_LL2UTM([float(lst[i][col_Lat]),float(lst[i][col_Lon])])    #UTM of WGS 84                            
                        iR,iC = self.Locate_LL2ROI(d.bidir,pre_iR,pre_iC,int(lst[i][col_plotID]),Easting,Northing,plots)
                        #print(i,'pID=',int(lst[i][col_plotID]),'pre_iC=',pre_iC,'iR,iC=',iR,iC)
                        pre_iC = iC
                        pre_iR = iR
                        #print('col_R=',col_R,'col_C=',col_C)
                        #iR = int(lst[i][col_R]) - 1  # img coord. for Plot Row 
                        #iC = int(lst[i][col_C]) - 1  # img coord. for Plot Col
                        iR = iR - 1
                        iC = iC - 1
                        n =  self.GRID[1]*iR + iC
                        #skip the buffer between wet & dry fields
                        if i-FD_start >= n_plot: n = n + n_BufferCol   
                        #print('n=',n,'self.GRID[1]=',self.GRID[1],'iRxiC=',iR,iC)
                        if self.georef == True:
                            x1,y1,x2,y2,mode = self.ROI[n]
                            #self.frame.canvas.create_rectangle(x1*self.zf,y1*self.zf,x2*self.zf,y2*self.zf, outline='white', fill='#%02X%02X%02X'%(R,G,B))   #rectangular ROI                          
                            cv2.rectangle(self.im_base, (x1,y1), (x2,y2), (R,G,B), -1)   #-1: fill in the circle
                        else:
                            self.PxRol = 0
                            x1,y1,x2,y2,mode = self.ROI[n] 
                            x1 = x1-self.ROI[0][0]
                            y1 = y1-self.ROI[0][1]
                            x2 = x2-self.ROI[0][0]
                            y2 = y2-self.ROI[0][1]
                            cv2.rectangle(self.im_base, (x1,y1), (x2,y2), (R,G,B), -1)   #-1: fill in the circle
                # draw markers based on GPS coord.
                else:    
                    # correct GPS by applying offsets from json
                    if fbasename.lower().find('avg')>=0:
                        Easting0 = float(lst[i][col_E])
                        Northing0 = float(lst[i][col_N])
                        #print(Easting0,Northing0)
                        #print('col_Yaw=',col_Yaw,'yaw=',float(lst[i][24]),'yaw_off=',yaw_off)
                        yaw = radians(float(lst[i][col_Yaw])+yaw_off)   #deg2rad
                        #print ('bb=',bb[0],'aa=',aa[0],'yaw1=',yaw1)
                        
                        ## calculate a heading angle between two LL points
                        #if i<arm_n+1: 
                            #w=h=yaw = 0
                        #else:           
                            ## using UTM coord.
                            #w = Easting0 - float(lst[i-arm_n][col_E])
                            #h = Northing0 - float(lst[i-arm_n][col_N])
                            #angle = atan2(h,w)
                            #yaw = (pi/2 - angle)%(2*pi)                            
                            ### using LL coord.
                            ##pointA = (float(lst[i-arm_n][col_Lat]), float(lst[i-arm_n][col_Lon]))                            
                            ##pointB = (float(lst[i][col_Lat]), float(lst[i][col_Lon]))
                            ##yaw = Calculate_Heading(pointA,pointB)                            
                        ##if i>arm_n:
                            ##print ('yaw1=',radians(float(lst[i][col_Yaw])),yaw1,'yaw=',yaw,'yaw-yaw1=',yaw-yaw1,degrees(yaw-yaw1))
                        
                        # Correct GPS heading offsets                
                        Easting,Northing = Correct_GPSoffset(Easting0,Northing0,aa[i%arm_n],bb[i%arm_n],yaw)
                        #Easting = float(lst[i][col_Ea])
                        #Northing = float(lst[i][col_Na])                        
                        Easting = Easting + 0
                        Northing = Northing + 0
                        #print(Easting,Northing)
                    elif fbasename.lower().find('gt')>=0:
                        #print('col_Lat=',col_Lat,'col_Lon=',col_Lon)
                        if col_Lat<0 or col_Lon<0: 
                            messagebox.showerror(title='Error',message='Please load a file containing Lat, Lon!')
                            return
                        Lon_deg,Lat_deg,Easting,Northing = Convert_LL2UTM([float(lst[i][col_Lat]),float(lst[i][col_Lon])])    #UTM of WGS 84
                    
                    x = round((Easting - E_min)*sc)
                    y = round((N_max - Northing)*sc)
                    #print (i,x,y,Easting,Northing)
                    cv2.circle(self.im_base, (x,y), ms, (R,G,B), -1)  #-1: fill in the circle
                    #self.frame.canvas.create_oval((x-ms)*self.zf,(y-ms)*self.zf,(x+ms)*self.zf,(y+ms)*self.zf, fill='#%02x%02x%02x'%(R,G,B))    #draw circles
                    
                    # draw a line to connect data
                    if d.connect == 1:
                        if i>1: cv2.arrowedLine(self.im_base, (x_pre,y_pre), (x,y), (0,0,0), 1)  # line between circles
                        x_pre = x
                        y_pre = y
        
        lst.clear()
        #cv2.imshow('im_base',self.im_base), cv2.waitKey(0)            
        if d.plotmap == False or self.georef == False:
            self.Display_Image(self.im_base,seg='add_layer')  # seg='add_layer' to skip histogram update
            # update imScn to display as georeferenced
            self.Update_imScn(E_min,E_max,N_min,N_max)        
        else:
            self.Display_Image(self.im_base,seg='add_layer')  # seg='add_layer' to skip histogram update
            # update imScn to display as georeferenced
            self.Update_imScn(E_min,E_max,N_min,N_max)
        hh,mm,ss = ConvertSec2hhmmss(float(time.time()-start_time))
        print("Done! (run time = %02d:%02d:%02.1f seconds)"%(hh,mm,ss))
        
        return   

    # load a shapefile as a vetor layer
    def Layer_SHP(self, file=None, display=None):  
        #file = r"C:\Users\james.kim2\Documents\Research\Python\Border\FD_ALARC.shp"
        #file = r"C:\Users\james.kim2\Documents\Research\Python\Border\FD_F004\F004_Border_GPS.shp"
        if file == None: filename = LoadFile(self,'SHP')     
        else: filename = file
        # exit if canceled to open a file
        if len(filename)==0:    
            return        
        
        self.master.title("IMAP"+'  - '+filename)        #change the title of master widget        
        print('filename=',filename)
        
        shape, polys = Read_Shapefile(filename)        
        #print('polys=',polys)
        
        # Calcuate a canvas size to display the polygons
        E_min = N_min = VI_min = 10000000
        E_max = N_max = VI_max = -10000000    
        for i in range(len(polys)): 
            # convert LL2UTM
            if shape == 'POLYGON': 
                for j in range(4): 
                    if abs(polys[i][j][0])>180 and abs(polys[i][j][1])>180: Northing, Easting = polys[i][j] # UTM
                    else: Lon_deg,Lat_deg,Easting,Northing = Convert_LL2UTM(polys[i][j])                
                    E_min = min(E_min,Easting)
                    E_max = max(E_max,Easting)
                    N_min = min(N_min,Northing)
                    N_max = max(N_max,Northing)                

            elif shape == 'POINT': 
                if abs(polys[i][0])>180 and abs(polys[i][1])>180: Easting, Northing = polys[i] # UTM
                else: Lon_deg,Lat_deg,Easting,Northing = Convert_LL2UTM(polys[i])
                E_min = min(E_min,Easting)
                E_max = max(E_max,Easting)
                N_min = min(N_min,Northing)
                N_max = max(N_max,Northing)                                
            #print(i,j,Easting,Northing,E_min,E_max,N_min,N_max)                
        
        #Create a blank base image of GPS points
        self.im_base,E_min,E_max,N_min,N_max,sc = self.Create_Basemap(E_min-2,E_max,N_min,N_max+2)   # 2-m margin
        print('W x H =',self.imW,self.imH,'sc=',sc)        
        print('UTM min & max=',E_min,E_max,N_min,N_max)

        #self.pre_zf = self.zf = 1   #skip to properly zoom out the plot map        
        #cv2.imshow('im_base',self.im_base), cv2.waitKey(0)      
        
        print('len(self.ROI)',len(self.ROI),'len(polys)=',len(polys))    
        if shape == 'POLYGON': 
            if len(self.ROI)>len(polys): 
                del self.ROI[len(polys):]   # delete the unused extra elements in ROI list
            else:
                del self.ROI
                self.ROI = [[0]*5 for i in range(len(polys))]
                
        #Plot GPS points
        plot = [0]*4
        self.plot = []
        for i,poly in enumerate(polys):
            if shape == 'POLYGON': 
                # convert LL2UTM
                for j in range(4):
                    if abs(polys[i][j][0])>180 and abs(polys[i][j][1])>180: 
                        Northing, Easting = poly[j] # UTM
                    else: 
                        Lon_deg,Lat_deg,Easting,Northing = Convert_LL2UTM(poly[j])
                    plot[j] = (Easting, Northing)                        
                    
                    x = round((Easting - E_min)*sc)
                    y = round((N_max - Northing)*sc)  
                    #print (i,j,x,y)
                    ms = 5
                    #cv2.circle(self.im_base, (x,y), ms, (255,0,0), -1)  #-1: fill in the circle
                    if j==0: 
                        self.ROI[i][0] = x
                        self.ROI[i][1] = y
                    if j==2: 
                        self.ROI[i][2] = x
                        self.ROI[i][3] = y            
                self.ROI[i][-1] = 'p'    
                self.plots.append(plot)
                
                # update ROI from shapefile
                if self.ROI[i][0]>=0 and self.ROI[i][2]>=0:
                    #cv2.rectangle(self.im_base,(self.ROI[0][0],self.ROI[0][1]),(self.ROI[0][2],self.ROI[0][3]),(255,0,0),2)
                    self.btn_ROI_var = 1
                    self.ROI_var.set(1)
                    #self.btn_ROI()                         
            elif shape == 'POINT': 
                if abs(polys[i][0])>180 and abs(polys[i][1])>180: Easting, Northing = poly # UTM
                else: Lon_deg,Lat_deg,Easting,Northing = Convert_LL2UTM(poly)
                
                x = round((Easting - E_min)*sc)
                y = round((N_max - Northing)*sc)  
                #print (i,x,y)          
                ms = 5
                cv2.circle(self.im_base, (x,y), ms, (255,0,0), -1)  #-1: fill in the circle
            #print(i,Easting,Northing,E_min,E_max,N_min,N_max)                
            
        #print('self.ROI=',self.ROI)                        
        self.Display_Image(self.im_base,seg='add_layer')  # seg='add_layer' to skip histogram update but create imScn
                
        # update imScn to display as georeferenced
        self.Update_imScn(E_min,E_max,N_min,N_max)
        
        # save the shapefile to ROI file
        self.Save_ROI('filename') 
        
        print('done')        
        return           
                    
    # Convert LL in csv to a shapefile
    def Proc_Convert_CSV2SHP(self, file=None): 
        Convert_CSV2SHP(self, file=None)

    # Convert ROI to a shapefile
    def Proc_Convert_ROI2SHP(self, file=None): 
        #self.ROI = Draw_ROI(self.GRID, self.ROI, self.zf, self.frame.canvas, self.rot)
        
        plots = self.Convert_ROI2LL() 
        if plots == 0: return
        Write_shapefile(self, plots, file=None)
        print('done')

    def Read_Config(self,filename=None):
        if(filename==None): filename = os.getcwd()+'\Config\Config.txt'         
        print('config filename=',filename)
        self.bdry = []
        #self.ROI = [[0]*5 for i in range(2000)]   
        self.ROI = [[]]   
        self.GRID = [0]*2   # [Row, Col]
        self.plots = []
        self.stretch = []        
        self.dType = []           
        with open(filename) as f:
            lines = f.read().split('\n')        
        #print('lines=',lines)
        for i,line in enumerate(lines):
            if i>0 and i<5: self.bdry.append(re.findall(r'[-+]?\d*\.\d+|\d+',line)) 
            if i == 6: ROI_str_val = re.split(r', ',line)
            if i == 7: GRID_str_val = re.split(r', ',line)
            if i == 8: stretch_str_val = re.split(r', ',line)
            if i == 9: dType_str_val = re.split(r', ',line)
        #print('bdry=',self.bdry,'lines[2][1]=',lines[2][1],'bdry[2][0]=%d'%int(self.bdry[2][0]))
        
        # read ROI parameters
        ROI_str_val[0] = ROI_str_val[0].split(': ')[-1]
        for i in range(0,4): 
            self.bdry[i][0] = int(self.bdry[i][0])
            self.bdry[i][1] = int(self.bdry[i][1])
            self.ROI[0].append(int(ROI_str_val[i])) 
            #self.ROI[0][i] = int(ROI_str_val[i])
        #self.ROI[0][4] = ROI_str_val[4]
        self.ROI[0].append(ROI_str_val[4])
        #print('len(ROI_str_val)=',len(ROI_str_val))
        if len(ROI_str_val)>5:
            self.ROI_rot = float(ROI_str_val[5])
        else:
            self.ROI_rot = 0.0
        if len(ROI_str_val)>6:
            for i in range(6,8): self.ROI[0].append(float(ROI_str_val[i])) #UTM coord: Northing, Easting        
        #print('ROI_str_val=',ROI_str_val,'ROI[4]=',self.ROI[0][4])
        
        # read GRID parameters
        GRID_str_val[0] = GRID_str_val[0].split(': ')[-1]
        self.GRID[0] =int(GRID_str_val[0])
        self.GRID[1] =int(GRID_str_val[1])
        # update GRID in Control Panel
        self.entR.delete(0,tk.END)
        self.entR.insert(0, self.GRID[0])
        self.entC.delete(0,tk.END)
        self.entC.insert(0, self.GRID[1])
        #print('GRID=',self.GRID,'%d x %d'%(self.GRID[0],self.GRID[1]))
        
        ROI_n = len(self.ROI)
        #tROI = [[0]*9 for i in range(len(self.GRID[0]*self.GRID[1]))]    # 4 corners: NW,SE,NE,SW
        
        #Display ROI if btn_ROI is selected
        #if(self.ROI_var.get() == 1): 
            ##self.frame.canvas.delete('all')
            ##self.frame.canvas.create_image(0,0,anchor="nw",image=img_photo)            
            ##self.btn_ROI()            
            #pass
        #else:
        ROI_outer = [self.ROI[0][0], self.ROI[0][1], self.ROI[ROI_n-1][2], self.ROI[ROI_n-1][3], 'r']   # n: no display
        #print('ROI_outer=',ROI_outer)
        self.ROI = Update_ROIs(self.GRID, [ROI_outer], self.ROI_rot)    # Convert ROI to tROI if rotation
        #print('self.ROI=',self.ROI)

        (self.X1,self.Y1),(self.X2,self.Y2) = (self.ROI[0][:2], self.ROI[ROI_n-1][2:4])
        
        self.ROI_shape = self.ROI[0][-1]        
        print('ROI_shape=',self.ROI_shape,'ROI_rot=',self.ROI_rot)      
        
        # create ROIpx
        self.ROIpx = [[] for i in range(len(self.ROI))] # pixels (i,j) in tROI
        self.GRID_CntIn = 0
        for i in range (len(self.ROI)):
            self.ROIpx[i] = Extract_tROIpx(self.ROI[i])   # 3D array: [n-th ROI of ROIs], [n-th px in ROI], [i or j value] 
            self.GRID_CntIn = self.GRID_CntIn + len(self.ROIpx[i])    # total # of pixels in GRID
        self.pre_ROI0 = self.ROI[0]
        print('self.GRID_CntIn=',self.GRID_CntIn)
        
        # read Stretch parameters
        self.stretch.append(int(stretch_str_val[0].split(': ')[-1]))
        self.stretch.append(float(stretch_str_val[1]))
        self.stretch.append(float(stretch_str_val[2]))        
        self.stretch.append(float(stretch_str_val[3]))        
        self.stretch.append(float(stretch_str_val[4]))        
        #print('stretch=',self.stretch,'%s %.1f %.1f %.1f %.1f'%(self.stretch[0],self.stretch[1],self.stretch[2],self.stretch[3],self.stretch[4])) 
        option = int(self.stretch[0])    # Stretch options: 0 for Percentile, 1 for min/max
        if option>0: self.StretchMin_var.set(option)    
        self.entStretchL.delete(0,tk.END)
        self.entStretchL.insert(0, self.stretch[1])
        self.entStretchU.delete(0,tk.END)
        self.entStretchU.insert(0, self.stretch[2])        
        self.entStretchMin.delete(0,tk.END)
        self.entStretchMin.insert(0, self.stretch[3])
        self.entStretchMax.delete(0,tk.END)
        self.entStretchMax.insert(0, self.stretch[4])        
        
        # read dType parameters
        dType_str_val[0] = dType_str_val[0].split(': ')[-1]
        for i in range(0,2): self.dType.append(dType_str_val[i])
        for i in range(2,4): self.dType.append(float(dType_str_val[i]))
        self.dType.append(dType_str_val[4])                 
        #print('dType_str_val=',dType_str_val)
        #print('dType=',self.dType,'%s %s %.2f %.2f %s'%(self.dType[0],self.dType[1],self.dType[2],self.dType[3],self.dType[4]))         

        if(filename!=None): self.reset_sliders(self.bdry)
        return 
 
    #Save an image
    def Save_File(self,im=None,filename=None):        
        #try:
            #if im == None: pass  
        #except: imScn = im
        imScn = self.imScn 
        if filename == None:
            filename = tkinter.filedialog.asksaveasfilename(initialdir=self,title='Save an image file',filetypes = (("tif files","*.tif"),("tif files","*.jpg"),("all files","*.*")))
            #if filename:
            if len(filename)-filename.rfind('.') >4: filename += '.tif' #add extension if not typed
        # exit if canceled to open a file
        if len(filename)==0:    
            return   
        
        # update GSD if zoomed
        GSDx = self.ds.GetGeoTransform()[1] / self.zf
        GSDy = self.ds.GetGeoTransform()[5] / self.zf
        GSD = (GSDx, GSDy)
        print('GSD[0]=',self.ds.GetGeoTransform()[1],'/', self.zf, '=', GSD[0])
        print('GSD[1]=',self.ds.GetGeoTransform()[5],'/', self.zf, '=', GSD[1])
        print('GSD=', GSD)
        
        if self.rot!=0:
            (X1,Y1),(X2,Y2) = (0,0),(imScn.shape[1],imScn.shape[0])
        else:
            (X1,Y1),(X2,Y2) = (self.X1,self.Y1),(self.X2,self.Y2)
        
        #print('filename.find()=',filename.find('.tif'))
        print('len(imScn.shape)=',len(imScn.shape),'imScn.shape=',imScn.shape,'imScn.dtype=',imScn.dtype)
        if filename.find('mask') > 0: 
            self.mask_file = filename

            # apply binary only in ROI from imScn
            im_mask = np.full(imScn.shape[:2],255,np.uint8)                
            if len(imScn.shape)>2: #if color image  
                im_mask[Y1:Y2,X1:X2] =  cv2.inRange(imScn[Y1:Y2,X1:X2],np.array([1,1,1]),np.array([255,255,255]))
            else:                
                im_mask[Y1:Y2,X1:X2] =  cv2.inRange(imScn[Y1:Y2,X1:X2],np.array([1]),np.array([255])) 
            print('im_mask.shape=',im_mask.shape,'min,max=',np.min(im_mask),np.max(im_mask))
            
            #Create a geotif image
            window_ROI = ((0,imScn.shape[0]),(0,imScn.shape[1]))
            Save_GeoTiff(im_mask,self.src_file,filename,window_ROI)
            #Image.fromarray(im_mask).save(filename)   #may need to scale to 0-255?
        else:
            print('filename:%s\n'%(filename))              
            print('self.winfo_rootx(),self.winfo_geometry()=',self.winfo_rootx(),self.winfo_geometry())
            #x=self.winfo_rootx()
            #y=self.winfo_rooty()
            #w=x+self.winfo_width()
            #h=y+self.winfo_height()
            #ImageGrab.grab().crop((x,y,w,h)).save(filename)   

            #print('ROI[0]=',self.ROI[0])
            (X1,Y1),(X2,Y2) = (self.X1,self.Y1),(self.X2,self.Y2)
            border = (int(X1),int(Y1),int(X2)+1,int(Y2)+1)            
            
            # add ROI grid if filename contains '_roi'             
            fdir,fname,fbasename,ftype = ExtractFilename(filename)
            print("fdir,fname,fbasename,ftype=",fdir,fname,fbasename,ftype,'imScn.shape=',imScn.shape)
            if fbasename.lower().find('_roi') > 0: 
                ROI = [X1, Y1, X2, Y2, self.ROI[0][-1]]
                print('ROI size=',min(X2-X1,Y2-Y1), 'GridLineSize=',ceil(min(X2-X1,Y2-Y1)/1000))
                GridLineSize=ceil(min(X2-X1,Y2-Y1)/1000)
                im = Add_Grid(imScn, ROI, self.GRID, LineSize=GridLineSize)  # -> 3b image
                
                if self.btn_ROI_var == 1: 
                    window_ROI = ((Y1,Y2),(X1,X2))
                    Save_GeoTiff(im,self.src_file,filename,window_ROI)
                else: 
                    Image.fromarray(im).save(filename)
                    self.Load_Image(filename) # reset the self.im to the saved img
                #if self.btn_ROI_var == 1: Image.fromarray(im).crop(border).save(filename)
                #else:                     Image.fromarray(im).save(filename)
                del im
            else:
                # clip the image into geotif
                print('imScn.shape=',imScn.shape,'imScn.dtype=',imScn.dtype)
                if self.btn_ROI_var == 1:                     
                    window_ROI = ((Y1,Y2),(X1,X2))
                    if self.georef==1:   #save as geotiff if image is geotiff
                        #self.im_NDVI = Save_GeoTiff(imScn,self.src_file,filename,window_ROI)                    
                        self.im_NDVI = Save_GeoTiff(self.im,self.src_file,filename,window_ROI)
                    else: 
                        #Image.fromarray(imScn[Y1:Y2,X1:X2]).save(filename)
                        Image.fromarray(self.im[Y1:Y2,X1:X2]).save(filename)                        
                else:
                    #print('self.georef=',self.georef)
                    if self.georef==1:   #save as geotiff if image is geotiff
                        #Create a geotif image
                        if self.im_nband>=4 and filename.lower().find('ndvi')>=0:
                            Save_GeoTiff(Generate_NDVI(self.im[:,:,3],self.im[:,:,0]),self.src_file,filename)
                        else:                             
                            if self.im.shape[:2] != imScn.shape[:2]:
                                window_ROI = ((0,imScn.shape[0]),(0,imScn.shape[1]))
                                print("im.shape=",self.im.shape[:2], "imScn.shape=",imScn.shape[:2],"window_ROI=",window_ROI)                                
                                if self.im.dtype == np.uint8:
                                    Save_GeoTiff(imScn,self.src_file,filename,window_ROI,GSD)
                                else:
                                    if self.rot==0: 
                                        Save_GeoTiff(self.im,self.src_file,filename,window_ROI,GSD)                                        
                                    else:
                                        im = imu.rotate_bound(self.im,self.rot)                
                                        Save_GeoTiff(im,self.src_file,filename,window_ROI,GSD)                                                                            
                            else:
                                Save_GeoTiff(imScn,self.src_file,filename)
                                #Save_GeoTiff(self.im,self.src_file,filename)
                    else:
                        if imScn.dtype == 'float32':    #if display img is float32, e.g. FFT, convert to uint8
                            Image.fromarray(np.uint8(imScn)).save(filename)   #may need to scale to 0-255?
                        else: 
                            if Image.fromarray(imScn).mode=='RGBA': # if RGBA, ->RGB to save as a jpg                               
                                Image.fromarray(cv2.cvtColor(imScn,cv2.COLOR_RGBA2RGB)).save(filename)
                            else:
                                Image.fromarray(imScn).save(filename)   #may need to scale to 0-255?
                    #self.Load_Image(filename) # reset the self.im to the saved img    
        print('done')            

    def Save_Config(self,filename=None):       
        if(filename==None):
            filename = tkinter.filedialog.asksaveasfilename(initialdir=self,title='Save as a text file',filetypes = (("txt files","*.txt"),("all files","*.*")))
            print('filename.find()=',filename.find('.txt'))                
            if filename.find('.txt') < 0: filename += '.txt' #add extension if not typed                
            print('filename=',filename)
        else:
            filename = os.getcwd()+'\Config\Config.txt' 
        # exit if canceled to open a file
        if len(filename)==0:    
            return        
                
        ch = 'RGBN'
        print('bdry=',self.bdry,'bdry[2][0]=%d'%int(self.bdry[2][0]))
        with open(filename,'w') as f:
            f.write('Threshoolding\n')
            for i in range(4):
                f.write('%s: %d %d\n'%(ch[i],self.sldL[i].get(),self.sldU[i].get()))
                self.bdry[i][0] = self.sldL[i].get()
                self.bdry[i][1] = self.sldU[i].get()                
            f.write('Mask thresholding\n')
            #f.write('ROI: %d, %d, %d, %d, %c, %.2f'%(self.ROI[0][0],self.ROI[0][1],self.ROI[len(self.ROI)-1][2],self.ROI[len(self.ROI)-1][3],self.ROI[0][-1],self.ROI_rot))
            f.write('ROI: %d, %d, %d, %d, %c, %.2f\n'%(self.ROI[0][0],self.ROI[0][1],self.ROI[len(self.ROI)-1][2],self.ROI[len(self.ROI)-1][3],self.ROI[0][8],self.ROI_rot))
            #if len(self.ROI[0])>5: f.write(', %.1f, %.1f\n'%(self.ROI[0][5],self.ROI[0][6]))
            #else:  f.write('\n')
            f.write('GIRD: %d, %d\n'%(self.GRID[0],self.GRID[1]))
            f.write('Stretch: %d, %.1f, %.1f, %.1f, %.1f\n'%(self.stretch[0],self.stretch[1],self.stretch[2],self.stretch[3],self.stretch[4])) #percentile to stretech NDVI heatmap, 0-100
            f.write('dType: %s, %s, %.2f, %.2f, %s\n'%(self.dType[0],self.dType[1],self.dType[2],self.dType[3],self.dType[4])) #for fmCalendar cmap
            
        print('bdry=',self.bdry,'bdry[2][0]=%d'%int(self.bdry[2][0]))
        
        return
    
    def Load_Config(self):       
        filename = LoadFile(self,'txt')
        if len(filename)==0:    
            return 
        
        self.Read_Config(filename)
        
        # enable ROI button
        if not self.ROI_var.get():
            self.ROI_var.set(1), self.btn_ROI()
        else:
            self.ROI_var.set(0), self.btn_ROI()
            self.ROI_var.set(1), self.btn_ROI()            
        return

    
    def Load_ROI(self):
        filename = LoadFile(self,'txt')
        # exit if canceled to open a file
        if len(filename)==0:    
            return        

        #read a string of ROI filename
        if filename.find('_') > 0:
            if len(filename.split('_')) > 2:             
                self.cropID = filename.split('_')[-2]
            else: 
                self.cropID = filename.split('_')[-1].split('.')[-2]
        
        GRID,ROI,ROI_rot,plots = Read_ROI(filename)
        
        # reassign ROI	
        self.ROI_var.set(1)
        self.GRID = GRID
        del self.ROI        
        self.ROI = ROI
        self.ROI_rot = ROI_rot
        del self.plots        
        self.plots = plots
        print('self.GRID,self.ROI=',self.GRID, self.ROI)
        print('ROI_rot,plots=',ROI_rot,plots)

        # update GRID entry widgets in Control Panel
        self.entR.delete(0,tk.END)
        self.entR.insert(0, self.GRID[0])
        self.entC.delete(0,tk.END)
        self.entC.insert(0, self.GRID[1])        
        
        #Display ROI if btn_ROI is selected
        if(self.ROI_var.get() == 1): 
            self.frame.canvas.delete('all')
            self.frame.canvas.create_image(0,0,anchor="nw",image=img_photo)            
            self.btn_ROI()                    
        
        return 
    
    #Read an excel file to extract Arable data time-matching and write to TOA\columns
    def Read_AM(self):
        start_time = time.time()
        import xlrd     
        import openpyxl
        #from datetime import datetime
        #fdir = self
        fdir = 'C:/Users/james.kim2/Documents/Research/Python/IMAP/Data/AM'
        filename = LoadFile(fdir,'xlsx')
        src_dir = filename[:filename.rfind(filename.split('/')[-1])]
        dst_dir = filename[:filename.rfind(filename.split('/')[-2])]
        print('src_dir=',src_dir,'dst_dir=',dst_dir)
        
        with xlrd.open_workbook(filename) as book:                        
            sheet = book.sheet_by_name('hourly')  #0 corresponds for 1st worksheet, usually named 'Book1'     
            AM_date = [ AY for AY in sheet.col_values(26+25-1) ] # gets col AY values
            AM_time = [ AZ for AZ in sheet.col_values(26+26-1) ] # gets col AZ values
            AM_NDVI = [ BA for BA in sheet.col_values(26+26+1-1) ] # gets col BA values
            AM_NDVIuw = [ BB for BB in sheet.col_values(26+26+2-1) ] # gets col BB values
            AM_BLUuw = [ H for H in sheet.col_values(8-1) ] # gets col H values
            AM_GRNuw = [ J for J in sheet.col_values(10-1) ] # gets col J values
            AM_REDuw = [ N for N in sheet.col_values(14-1) ] # gets col N values
            AM_NIRuw = [ R for R in sheet.col_values(18-1) ] # gets col R values
            AM_BLU = [ AO for AO in sheet.col_values(26+15-1) ] # gets col AO values
            AM_GRN = [ AP for AP in sheet.col_values(26+16-1) ] # gets col AP values
            AM_RED = [ AR for AR in sheet.col_values(26+18-1) ] # gets col AR values
            AM_NIR = [ AT for AT in sheet.col_values(26+20-1) ] # gets col AT values            
            sheet = book.sheet_by_name('TOA')  #1 corresponds for 1st worksheet, usually named 'Book2'
            PS_date = [ R for R in sheet.col_values(26-9) ] # gets col R values
            PS_time = [ S for S in sheet.col_values(26-8) ] # gets col S values            
            #PS_NDVI = [ K for K in sheet.col_values(11-1) ] # gets col K values       
            #for i in range(len(AM_date)):
                #if i>0: dt = datetime(*xlrd.xldate_as_tuple(AM_date[i],0))
            print('AM_date=',len(AM_date),AM_date)
            print('AM_time=',AM_time)
            print('AM_NDVI=',AM_NDVI)
            print('PS_date=',len(PS_date),PS_date)
            print('PS_time=',PS_time)   
            #print('PS_NDVI=',PS_NDVI)
            workbook = openpyxl.load_workbook(filename)
            worksheet = workbook.get_sheet_by_name('TOA')            
            worksheet['AA1'] = 'AM_time'
            worksheet['AB1'] = 'AM_NDVI'
            worksheet['AC1'] = 'AM_NDVIuw'
            worksheet['AD1'] = 'AM_BLUuw'
            worksheet['AE1'] = 'AM_GRNuw'
            worksheet['AF1'] = 'AM_REDuw'
            worksheet['AG1'] = 'AM_NIRuw'
            worksheet['AH1'] = 'AM_BLU'
            worksheet['AI1'] = 'AM_GRN'
            worksheet['AJ1'] = 'AM_RED'
            worksheet['AK1'] = 'AM_NIR'            
            for i in range(1,len(PS_date)):
                min_t = 1.0
                jj = 0
                for j in range(1,len(AM_date)):                    
                    if AM_NDVI[j]!=15 and PS_date[i] == AM_date[j]: #15='#VALUE!'                        
                        min_t = min(min_t,abs(PS_time[i]-AM_time[j]))
                        #print 'i=',i,'j=',j,'min_t=',min_t
                        if min_t == abs(PS_time[i]-AM_time[j]):
                            jj = j
                if jj != 0:
                    print('PS_date=',PS_date[i],'min_t=',min_t,'PS_time[%d]=%f'%(i,PS_time[i]),'AM_time[%d]=%f'%(jj,AM_time[jj]))
                    worksheet['AA%d'%(i+1)] = AM_time[jj]   #(row,col,data) @col AA
                    worksheet['AB%d'%(i+1)] = AM_NDVI[jj]   #(row,col,data) @col AC
                    worksheet['AC%d'%(i+1)] = AM_NDVIuw[jj]   #(row,col,data) @col AD
                    worksheet['AD%d'%(i+1)] = AM_BLUuw[jj]   #(row,col,data) @col AE
                    worksheet['AE%d'%(i+1)] = AM_GRNuw[jj]   #(row,col,data) @col AF
                    worksheet['AF%d'%(i+1)] = AM_REDuw[jj]   #(row,col,data) @col AG
                    worksheet['AG%d'%(i+1)] = AM_NIRuw[jj]   #(row,col,data) @col AH
                    worksheet['AH%d'%(i+1)] = AM_BLU[jj]   #(row,col,data) @col AI
                    worksheet['AI%d'%(i+1)] = AM_GRN[jj]   #(row,col,data) @col AJ
                    worksheet['AJ%d'%(i+1)] = AM_RED[jj]   #(row,col,data) @col AK
                    worksheet['AK%d'%(i+1)] = AM_NIR[jj]   #(row,col,data) @col AL                    
                    #worksheet.cell(row=i+1,column=26+4-1).value = AM_time[jj] #(row,col,data) @col AA
            workbook.save(filename)

        hh,mm,ss = ConvertSec2hhmmss(float(time.time()-start_time))
        print("Done! (run time = %02d:%02d:%02.1f seconds)"%(hh,mm,ss))
        
    def update_Menu(self,str_Img):
        #print('str_Img=',str_Img)    
        #Reset the buttons
        self.SEG_var.set(0)
        self.MASK_var.set(0)
        #Reset the Menu check and commands to default
        self.FileMenu.entryconfig('Save As',state='normal') 
        for i in range(9): self.FilterMenu.entryconfig(i,state='normal') 
        for i in range(9): self.ProcMenu.entryconfig(i,state='normal') 
        for i in range(9): self.RasterMenu.entryconfig(i,state='normal') 
        #if self.im_nband<4:            
            #self.ProcMenu.entryconfig(1,state='disabled') #disable NDVI if 3b img 
        #Reset Erode and Dilate to default
        self.cnt_erode = self.cnt_dilate = 0
        self.FilterMenu.entryconfig(0,label='Erode') 
        self.FilterMenu.entryconfig(1,label='Dilate')         
        
        label_Img = ['RGBN','RGB','HSV','LAB','NDVI','Heatmap','Gray']
        if str_Img in label_Img:
            self.HSV_var.set(0)
            self.LAB_var.set(0)
            self.RGB_var.set(0)               
            self.NDVI_var.set(0)
            self.Heatmap_var.set(0)
            if str_Img == 'NDVI' or str_Img == 'Heatmap' or str_Img == 'Gray':                
                pass
            else:
                for i in [0,1,2]: self.RdBtn[i].config(text=str_Img[i]) # update Radiobuttons\R,G,B => L,A,B
        
        #self.ch = str_Img                
        if str_Img in ('RGBN','RGB'):
            #if self.im_nband==4: self.ch = 'RGBN'
            self.RGB_var.set(1)      
        elif str_Img == ('HSV'): self.HSV_var.set(1)      
        elif str_Img == ('LAB'): self.LAB_var.set(1)      
        elif str_Img in ('NDVI', 'Gray'): self.NDVI_var.set(1); #self.ch = 'R'   
        elif str_Img == 'Heatmap': self.Heatmap_var.set(1); #self.ch = 'R'  
        
    def btn_Band(self):   
        self.B = self.Band_var.get()
        #print('self.B=',self.B)
        self.btn_cmap = ['R','G','B','Y'][self.B]
        #print('self.src_file.find(udm)=',self.src_file.find('udm'))
        if self.ch == 'NDVI': self.Display_Image(self.imScn_base)   
        elif self.im_nband==1 or self.B==4: # 1b image            im = self.im_stretched
            if self.src_file.find('udm') != -1: #detect 8-bit udm file:0(fg),1(outBorder),2(cloud),4(B),8(G),16(R),32(RE),64(N)
                im[im==0] = 255 #0(fgnd)->255 
                im[im<255] = 0  #1(nan)->0, 2(cloud)->0, 64(N missed)->0 
            elif self.src_file.find('border') != -1:    #detect 8-bit border file:0(outBorder),1(fg)
                im[im==1] = 255  #1(fg)->255 
            self.Display_Image(im)
        elif self.im_nband==2: #2b Cloud&Shadow mask image
            self.Display_Image(255-self.im*255) #b2(shadow):0(fgnd)->255,1(shadow)->0 \        
        elif self.im_nband<=4:
            self.Display_Image(self.imScn_base[:,:,self.B])   #012:RGB
        #else: messagebox.showerror(title='Error',message='This is +4b image. Load 4b image!')

    # Update state of radiobuttons and sliders
    def Update_BtnSlider(self,nband,state):
        for i in range(nband):  
            self.RdBtn[i]['state'] = state # Update radiobutton
            self.sldL[i]['state'] = state # Update Lower slider
            self.sldU[i]['state'] = state # Update Upper slider        
        if nband<4: # disable bands if unavailable
            for i in range(nband,4):  
                self.RdBtn[i]['state'] = 'disabled' # Update radiobutton
                self.sldL[i]['state'] = 'disabled' # Update Lower slider
                self.sldU[i]['state'] = 'disabled' # Update Upper slider        
        
    # select an image conversion in combobox: RGB, HSV, LAB, NDVI, Heatmap
    def Img_Select(self, event=None, im=None, ch=None):   
        self.Band_var.set(None) # deselect all Band Radiobuttons
        self.B = -1 # 1b selection: -1(none), 0(R), 1(G), 2(B)
        self.max_y = 0
        
        try:
            if im == None: im = self.im
        except: pass
            
        if ch==None: ch = list(self.sel_Images)[self.combxViewImg.current()]
        
        self.btn_cmap = 'RGB'
        if ch == 'RGB':
            self.MASK_var.set(0)
            #self.ROI_var.set(0)
            self.SEG_var.set(0)
            if self.im_nband==1: self.btn_cmap = 'R'
            elif self.im_nband == 4: 
                self.btn_cmap = 'RGBY'
                ch = 'RGBN'
            if self.rot != 0: self.imScn_base = imu.rotate_bound(self.im_stretched,self.rot)  # rotate imScn_base if rot assigned
            else:             self.imScn_base = self.im_stretched #baseimage of imScn before segmentation
        elif ch == 'HSV': 
            if self.rot != 0: self.imScn_base = imu.rotate_bound(self.im_stretched,self.rot)  # rotate imScn_base if rot assigned
            else:             self.imScn_base = self.im_stretched #baseimage of imScn before segmentation
            self.imScn_base = cv2.cvtColor(self.imScn_base,cv2.COLOR_RGB2HSV_FULL) #baseimage of imScn before segmentation
        elif ch == 'LAB': 
            if self.rot != 0: self.imScn_base = imu.rotate_bound(self.im_stretched,self.rot)  # rotate imScn_base if rot assigned
            else:             self.imScn_base = self.im_stretched #baseimage of imScn before segmentation
            self.imScn_base = cv2.cvtColor(self.imScn_base,cv2.COLOR_RGB2LAB) #baseimage of imScn before segmentation
        elif ch == 'NDVI':
            if self.im_nband==4:                 
                self.btn_cmap = 'Y'
                # Calculate NDVI
                imNDVI = Generate_NDVI(im[:,:,3],im[:,:,0]) # R,G,B,N =im[:,:,0],im[:,:,1],im[:,:,2],im[:,:,3]
                self.imNDVI = imNDVI
                #print('im_NDVI=%s, %.3f (%.3f~%.3f)'%(self.imScn_base.dtype,np.nanmean(self.imScn_base),np.nanmin(self.imScn_base),np.nanmax(self.imScn_base)),end='->')
                # filter NDVI
                #imNDVI[imNDVI<0] = np.nan
                #imNDVI[imNDVI>1] = 1   # filter NDVI>1 -> 1
                #print('imNDVI=',imNDVI.shape,imNDVI.dtype,np.nanmean(imNDVI),np.nanmin(imNDVI),np.nanmax(imNDVI))
                
                #self.imScn_base = (255*imNDVI).astype(np.uint8)                
                if self.stretch[0]>=0 or im.dtype != 'uint8':   # Stretch options: 0 for Percentile, 1 for min/max, unit16->unit8 w percentile stretch
                    self.imScn_base = self.Update_StretchVar(imNDVI,self.stretch)
                    
                if self.rot != 0 and self.im.shape[:2] == self.imScn_base.shape[:2]: 
                    self.imScn_base = imu.rotate_bound(self.imScn_base,self.rot)  # rotate imScn_base if rot assigned                
                
                self.ch = ch
                self.max_y = max(self.max_y,np.max(cv2.calcHist([self.imScn_base],[0],None,[256],[1,256]))) # 1: exclude saturated 0s
                
                # Display NDVI                
                self.Display_Image(self.imScn_base)    

                # disable Radiobuttons
                self.Update_BtnSlider(1,'normal')
                
            else: messagebox.showerror(title='Error',message='Please load 4b image first!')            
            return self.imScn_base
        elif ch == 'Heatmap':
            #invert the image if veg<soil
            #if np.nanpercentile(im[self.Y1:self.Y2,self.X1:self.X2],(5,95))[1]>100: im = (255-im)   
            
            if len(im.shape)>2 and im.shape[2]==1:
                im = im.reshape(im.shape[0],im.shape[1])   # (H,W,1)->(H,W) for plt.imshow()
            if self.ROI_var.get()==1: im_GRID = im[self.Y1:self.Y2,self.X1:self.X2]
            else:                     im_GRID = im
            
            #print('im=',im.shape,im.dtype,'mean=',np.nanmean(im),'min=',np.nanmin(im),'max=',np.nanmax(im),np.nanmax(im_GRID),end=', ')
            if self.HeatmapStretch_var.get():
                self.htstch=['yes', float(self.entHeatmapStretchL.get()), float(self.entHeatmapStretchU.get())]
                minP,maxP = np.nanpercentile(im_GRID,(self.htstch[1],self.htstch[2]))            
                del im_GRID
                im = np.interp(im[:,:],(minP,maxP),(0,1))                
                #print('percentile_GRID=',minP,maxP)
                
            if len(im.shape)==2: #if 1-ch image 
                #self.imScn_base = (plt.get_cmap('RdYlGn')(im/maxP) * 255).astype(np.uint8)[:,:,:3] 
                heatmap = (plt.get_cmap(self.cmap)(im) * 255).astype(np.uint8)[:,:,:3] 
                #print('self.im=',self.im.shape,self.im.dtype,'im=',im.shape,im.dtype)
                if self.rot != 0 and self.im.shape[:2] == im.shape[:2]: 
                    heatmap = imu.rotate_bound(heatmap,self.rot)  # rotate imScn_base if rot assigned                
                
                del im
                if event==None: 
                    return heatmap
                else: self.imScn_base = heatmap
            else: messagebox.showerror(title='Error',message='Please load 1-ch image first!')
        
        self.Update_BtnSlider(len(self.btn_cmap),'normal')

        # calculate max_y for histogram display
        im_nband = len(self.btn_cmap)
        if self.im_nband == 1: im_nband = 1
        for i in range(im_nband): 
            self.max_y = max(self.max_y,np.max(cv2.calcHist([self.imScn_base],[i],None,[256],[1,256]))) # 1: exclude saturated 0s
        
        self.ch = ch
        self.Display_Image(self.imScn_base) 
        
    # select a cmap in combobox: RdYlGn':0,'hsv':1,'brg':2,'gist_rainbow':3,'nipy_spectral':4
    def Cmap_Select(self, event=None):   
        self.cmap = list(self.sel_Cmap)[self.combxViewCmap.current()]        
        print('self.cmap=',self.cmap)
        
    # select a Batch\Masking source in combobox: 'RGB':0,'HSV':1,'Sat':2,'NDVI':3,'RGN':4
    def Mask_Select(self, event=None):   
        self.maskSrc = list(self.sel_Mask)[self.combxViewMask.current()]
        print('self.maskSrc=',self.maskSrc)

    # select a Batch\Filtering source in combobox: 'None':0,'Opening':1,'Closing':2
    def Filter_Select(self, event=None):   
        self.filterSrc = list(self.sel_Filter)[self.combxViewFilter.current()]
        print('self.filterSrc=',self.filterSrc)

    # select a Batch\Filtering source in combobox: 'None':0,'Opening':1,'Closing':2
    def ROI_Select(self, event=None):   
        self.ROIsrc = list(self.sel_ROI)[self.combxViewROI.current()]
        print('self.ROIsrc=',self.ROIsrc)
        
    def Proc_Create_NDVI(self):
        self.combxViewImg.current(3)    # {'RGB':0,'HSV':1,'LAB':2,'NDVI':3,'Heatmap':4}
        #print('self.ch=',self.ch)
        im_NDVI = self.Img_Select()
        
        fdir,fname,fbasename,ftype = ExtractFilename(self.src_file)
        dst_dir = fdir+'Calib/'
        if not os.path.exists(dst_dir): os.makedirs(dst_dir)        
        dst_file = dst_dir+fbasename+'_ndvi.'+ftype
        
        # Save NDVI
        Save_GeoTiff(im_NDVI,self.src_file,dst_file)                
        print('Created NDVI')
  
    def init_buttons_sliders(self):
        self.btn = []
        self.sldL = []
        self.sldU = []
        
        BtnX = self.frame_w  #button anchor at the left top x-coordinate
        BtnX = 0
        BtnY = 0   #button anchor at the left top y-coordinate
        BtnW = 50
        BtnH = 30        
        BtnGapX = 60
        BtnGapY = 40
        
        ##Add Mask button        
        #ckbtn = tk.Button(self.frame2,text='Mask',command=self.Mask_btn)
        #ckbtn.place(x=420,y=10, width=50, height=30)        
        #Add Mask checkbutton   
        self.MASK_var = tk.IntVar()
        ckbtn = tk.Checkbutton(self.frame2,text='Mask',command=self.btn_MASK, bd=4, variable=self.MASK_var,indicatoron=False,onvalue=1,offvalue=0)
        ckbtn.place(x=BtnX,y=BtnY, width=BtnW, height=BtnH)        
        #Add SegStats checkbutton
        self.SegStats_var = tk.IntVar()
        ckbtn = tk.Checkbutton(self.frame2,text='ROI Stats',command=self.btn_SegStats, bd=4,variable=self.SegStats_var,onvalue=1,offvalue=0)
        ckbtn.place(x=BtnX+BtnGapX*1,y=BtnY+BtnGapY/2, width=BtnW+15, height=BtnH/2)         
        #Add Seg checkbutton
        self.SEG_var = tk.IntVar()
        ckbtn = tk.Checkbutton(self.frame2,text='SEG',command=self.btn_SEG, bd=4,variable=self.SEG_var,indicatoron=False,onvalue=1,offvalue=0)
        ckbtn.place(x=BtnX+20+BtnGapX*1,y=BtnY+BtnGapY, width=BtnW-10, height=BtnH)        
        #Add SAVE button for ROI
        btn = tk.Button(self.frame2,text='SAVE',command=self.btn_SAVE)
        btn.place(x=BtnX+5+BtnGapX*2,y=BtnY+BtnGapY, width=BtnW-10, height=BtnH)        
        #Add ROI checkbutton
        self.ROI_var = tk.IntVar()
        ckbtn = tk.Checkbutton(self.frame2,text='ROI',command=self.btn_ROI, bd=4,variable=self.ROI_var,indicatoron=False,onvalue=1,offvalue=0)
        ckbtn.place(x=BtnX-10+BtnGapX*3,y=BtnY+BtnGapY, width=BtnW-10, height=BtnH)        
        
        # add a ComboBox for multi-ROIs
        self.sel_ROI = {'0':0,'1':1,'2':2}
        self.combxViewROI = ttk.Combobox(self.frame2, state="readonly", value=list(self.sel_ROI.keys()), width=4)
        self.combxViewROI.place(x=BtnX+BtnGapX*3+35,y=BtnY+BtnGapY+10, width=BtnW-15, height=20)
        self.combxViewROI.current(0)    # {'0':0,'1':1,'2':2}
        self.combxViewROI.bind("<<ComboboxSelected>>", self.ROI_Select)
        
        #Display entry widgets for Grid
        text = tk.Text(self.frame2,relief=tk.FLAT,bg='SystemButtonFace')
        text.insert(tk.INSERT,"GRID:R x C")
        text.place(x=BtnX+BtnGapX*3-35,y=BtnY, width=100, height=20)
        self.entR = tk.Entry(self.frame2)
        self.entR.place(x=BtnX+BtnGapX*3,y=BtnY+20, width=25, height=20)               
        self.entR.insert(0, "1")
        self.entC = tk.Entry(self.frame2)
        self.entC.place(x=BtnX+BtnGapX*3+30,y=BtnY+20, width=25, height=20)
        self.entC.insert(0, "1")
        #Update the gird by mouse wheel
        #self.MouseWheelCnt_R = int(self.entR.get())  #Mouse Wheel Count
        #self.MouseWheelCnt_C = int(self.entC.get())  #Mouse Wheel Count
        self.entR.bind('<MouseWheel>',self.MouseWheel_entR)       
        self.entC.bind('<MouseWheel>',self.MouseWheel_entC)       
        
        #Read Shift + mouse wheel in the image canvas for Rotating
        self.frame.canvas.bind('<Shift-MouseWheel>',self.ShiftMouseWheel_Image)     
        #Read Ctrl + mouse wheel in the image canvas for Zooming
        self.frame.canvas.bind('<Control-MouseWheel>',self.CtrlMouseWheel_Image)     
        #Read Alt + mouse wheel in the image canvas for ROI Rotating
        self.frame.canvas.bind('<Alt-MouseWheel>',self.AltMouseWheel_Image)     
        #Read Alt + mouse wheel in the image canvas for ROI Rotating
        self.frame.canvas.bind('<Control-Alt-MouseWheel>',self.CtrlAltMouseWheel_Image)     
        #Scroll the image by mouse wheel
        self.frame.canvas.bind('<MouseWheel>',self.MouseWheel_Img)       
        
        #Add Rotation button        
        btn = tk.Button(self.frame2,text='Rotate',command=self.btn_ROTATE)
        btn.place(x=BtnX+BtnGapX*4+20,y=BtnY, width=BtnW, height=BtnH)            
        #Display an entry widget for Rotation
        text = tk.Text(self.frame2,relief=tk.FLAT,bg='SystemButtonFace')
        text.insert(tk.INSERT,"deg")
        text.place(x=BtnX+BtnGapX*5+55,y=BtnY+10, width=30, height=20)
        self.entRot = tk.Entry(self.frame2)
        self.entRot.place(x=BtnX+BtnGapX*5+15,y=BtnY+10, width=40, height=20)               
        self.entRot.insert(0, "0.0")
        #Update the rotation factor by mouse wheel
        self.MouseWheelCnt_Rot = float(self.entRot.get())  #Mouse Wheel Count
        self.entRot.bind('<MouseWheel>',self.MouseWheel_entRot)

        #Add Zoom button        
        btn = tk.Button(self.frame2,text='ZOOM',command=self.btn_ZOOM)
        btn.place(x=BtnX+BtnGapX*4+20,y=BtnY+BtnGapY, width=BtnW, height=BtnH)            
        #Display an entry widget for Zoom
        text = tk.Text(self.frame2,relief=tk.FLAT,bg='SystemButtonFace')
        text.insert(tk.INSERT,"x")
        text.place(x=BtnX+BtnGapX*5+15,y=BtnY+BtnGapY+10, width=15, height=20)
        self.entZm = tk.Entry(self.frame2)
        self.entZm.place(x=BtnX+BtnGapX*5+30,y=BtnY+BtnGapY+10, width=40, height=20)               
        self.entZm.insert(0, "1.0")
        #Update the zoom factor by mouse wheel
        self.MouseWheelCnt_Zm = float(self.entZm.get())  #Mouse Wheel Count
        self.entZm.bind('<MouseWheel>',self.MouseWheel_entZm)
        
        #Add Threshold Reset button        
        btn = tk.Button(self.frame2,text='Reset',command=self.btn_ResetThreshold)
        btn.place(x=BtnX+BtnGapX*4+30,y=BtnY+BtnGapY*2, width=BtnW-10, height=BtnH-10)            
        ##Add Stat button       
        #Add ResetROI checkbutton
        self.ResetROI_var = tk.IntVar()
        ckbtn = tk.Checkbutton(self.frame2,text='ROI',command=self.btn_ResetROI, bd=4,variable=self.ResetROI_var,onvalue=1,offvalue=0)
        ckbtn.place(x=BtnX+BtnGapX*5+10,y=BtnY+BtnGapY*2+5, width=BtnW-10, height=BtnH/2)         
        
        #Add Stretch checkbutton
        self.Stretch_var = tk.IntVar()
        ckbtn = tk.Checkbutton(self.frame2,text='Stretch:',command=self.btn_Stretch, bd=4,variable=self.Stretch_var,onvalue=1,offvalue=0)
        ckbtn.place(x=BtnX,y=BtnY+BtnGapY+5*BtnH+5)
        #Add Stretch radiobuttons for min/max selection
        Ranges = [('Percentile',0),('Min',1)] # (Range,iter)
        self.StretchMin_var = tk.IntVar() # variable of two buttons
        self.StretchMin_var.set(None)
        self.RdBtn = []        
        for Range,i in Ranges:  
            self.RdBtn.append(tk.Radiobutton(self.frame2,text=Range,variable=self.StretchMin_var,value=i,command=self.btn_StretchMin))
            self.RdBtn[i].place(x=BtnX+64+i*195,y=BtnY+BtnGapY+5*BtnH+5, height=BtnH) #, width=BtnW, height=BtnH)        
        
        #Add Heatmap Stretch checkbutton
        self.HeatmapStretch_var = tk.IntVar()
        ckbtn = tk.Checkbutton(self.frame2,text='Heatmap stretch',command=self.btn_HeatmapStretch, bd=4,variable=self.HeatmapStretch_var,onvalue=1,offvalue=0)
        ckbtn.place(x=BtnX,y=BtnY+BtnGapY+6*BtnH)
        self.HeatmapStretch_var.set(1)
        #btn = tk.Button(self.frame2,text='Stat',command=self.btn_STAT)
        #btn.place(x=self.frame_w+10,y=BtnY+self.scr_h-500, width=BtnW-8, height=BtnH-10)            
        #Add Hist button        
        btn = tk.Button(self.frame2,text='Hist',command=self.btn_HIST)
        btn.place(x=self.frame_w+10,y=BtnY+self.scr_h-395, width=BtnW-8, height=BtnH-10)            
        
        #Display an entry widget for Stretch
        text = tk.Text(self.frame2,relief=tk.FLAT,bg='SystemButtonFace')
        self.entStretchL = tk.Entry(self.frame2)
        self.entStretchL.place(x=BtnX+BtnGapX*2+20,y=BtnY+BtnGapY+5*BtnH+10, width=35) 
        text.insert(tk.INSERT,"-")
        text.place(x=BtnX+BtnGapX*3-5,y=BtnY+BtnGapY+5*BtnH+10, width=10, height=20)
        self.entStretchU = tk.Entry(self.frame2)
        self.entStretchU.place(x=BtnX+BtnGapX*3+7,y=BtnY+BtnGapY+5*BtnH+10, width=35) 
        text = tk.Text(self.frame2,relief=tk.FLAT,bg='SystemButtonFace')
        text.insert(tk.INSERT,"% or")
        text.place(x=BtnX+BtnGapX*4-20,y=BtnY+BtnGapY+5*BtnH+10, width=40, height=20)
        #Display an entry widget for Stretch min/max
        text = tk.Text(self.frame2,relief=tk.FLAT,bg='SystemButtonFace')
        self.entStretchMin = tk.Entry(self.frame2)
        self.entStretchMin.place(x=BtnX+BtnGapX*4+65,y=BtnY+BtnGapY+5*BtnH+10, width=35) 
        text.insert(tk.INSERT,"-")
        text.place(x=BtnX+BtnGapX*5+40,y=BtnY+BtnGapY+5*BtnH+10, width=10, height=20)
        self.entStretchMax = tk.Entry(self.frame2)
        self.entStretchMax.place(x=BtnX+BtnGapX*5+52,y=BtnY+BtnGapY+5*BtnH+10, width=35) 
        #Display an entry widget for HeatmapStretch
        text = tk.Text(self.frame2,relief=tk.FLAT,bg='SystemButtonFace')
        self.entHeatmapStretchL = tk.Entry(self.frame2)
        self.entHeatmapStretchL.place(x=BtnX+BtnGapX*2+20,y=BtnY+BtnGapY+6*BtnH+5, width=35) 
        text.insert(tk.INSERT,"-")
        text.place(x=BtnX+BtnGapX*3-5,y=BtnY+BtnGapY+6*BtnH+5, width=10, height=20)
        self.entHeatmapStretchU = tk.Entry(self.frame2)
        self.entHeatmapStretchU.place(x=BtnX+BtnGapX*3+7,y=BtnY+BtnGapY+6*BtnH+5, width=35) 
        text = tk.Text(self.frame2,relief=tk.FLAT,bg='SystemButtonFace')
        text.insert(tk.INSERT,"%")
        text.place(x=BtnX+BtnGapX*4-20,y=BtnY+BtnGapY+6*BtnH+5, width=10, height=20)        
        self.entHeatmapStretchL.insert(0, 1)
        self.entHeatmapStretchU.insert(0, 99)
                
        # add a ComboBox for Img View
        self.sel_Images = {'RGB':0,'HSV':1,'LAB':2,'NDVI':3,'Heatmap':4}
        self.combxViewImg = ttk.Combobox(self.frame2, state="readonly", value=list(self.sel_Images.keys()), width=4)
        self.combxViewImg.place(x=BtnX,y=BtnY+BtnGapY+0*BtnH, width=BtnW+25, height=BtnH)
        self.combxViewImg.current(0)    # {'RGB':0,'HSV':1,'LAB':2,'NDVI':3,'Heatmap':4}
        self.combxViewImg.bind("<<ComboboxSelected>>", self.Img_Select)
        
        # add a ComboBox for Heatmap View
        self.sel_Cmap = {'RdYlGn':0,'hsv':1,'brg':2,'gist_rainbow':3,'nipy_spectral':4}
        self.combxViewCmap = ttk.Combobox(self.frame2, state="readonly", value=list(self.sel_Cmap.keys()), width=4)
        self.combxViewCmap.place(x=BtnX+BtnGapX*4,y=BtnY+BtnGapY+6*BtnH+5, width=BtnW+25, height=20)
        self.combxViewCmap.current(0)    # {RdYlGn':0,'hsv':1,'brg':2,'gist_rainbow':3,'nipy_spectral':4}
        self.combxViewCmap.bind("<<ComboboxSelected>>", self.Cmap_Select)

        # add options for Batch process        
        text = tk.Text(self.frame2,relief=tk.FLAT,bg='SystemButtonFace')
        text.insert(tk.INSERT,"Batch:")
        text.place(x=BtnX,y=BtnY+BtnGapY+7*BtnH, width=100, height=20)
        #Add a checkbutton for Calibration
        self.BatchCal_var = tk.IntVar()
        ckbtn = tk.Checkbutton(self.frame2,text='Calibration',command=self.btn_BatchCal, bd=4,variable=self.BatchCal_var,onvalue=1,offvalue=0)
        ckbtn.place(x=BtnX+BtnGapX-10,y=BtnY+BtnGapY+7*BtnH-5)        
        # add a ComboBox for Batch\Masking source        
        text = tk.Text(self.frame2,relief=tk.FLAT,bg='SystemButtonFace')
        text.insert(tk.INSERT,", Mask")
        text.place(x=BtnX+BtnGapX*2+15,y=BtnY+BtnGapY+7*BtnH, width=100, height=20)
        self.sel_Mask = {'RGB':0,'HSV':1,'LAB':2,'Sat':3,'NDVI':4,'RGN':5}
        self.combxViewMask = ttk.Combobox(self.frame2, state="readonly", value=list(self.sel_Mask.keys()), width=4)
        self.combxViewMask.place(x=BtnX+BtnGapX*3+10,y=BtnY+BtnGapY+7*BtnH, width=BtnW, height=20)
        self.combxViewMask.current(0)    # {'RGB':0,'HSV':1,'Sat':2,'NDVI':3,'RGN':4}
        self.combxViewMask.bind("<<ComboboxSelected>>", self.Mask_Select)
        # add a ComboBox for Batch\Filtering source
        text = tk.Text(self.frame2,relief=tk.FLAT,bg='SystemButtonFace')
        text.insert(tk.INSERT,", Filter")
        text.place(x=BtnX+BtnGapX*4,y=BtnY+BtnGapY+7*BtnH, width=100, height=20)
        self.sel_Filter = {'None':0,'Opening':1,'Closing':2}
        self.combxViewFilter = ttk.Combobox(self.frame2, state="readonly", value=list(self.sel_Filter.keys()), width=4)
        self.combxViewFilter.place(x=BtnX+BtnGapX*5+10,y=BtnY+BtnGapY+7*BtnH, width=BtnW+10, height=20)
        self.combxViewFilter.current(0)    # {'Default':0,'Otsu':1,'Mean':2}
        self.combxViewFilter.bind("<<ComboboxSelected>>", self.Filter_Select)        
        
        #Add a checkbutton for Batch clip to ROI
        self.BatchClip2ROI_var = tk.IntVar()
        ckbtn = tk.Checkbutton(self.frame2,text='Clip to ROI,',command=self.btn_BatchClip2ROI, bd=4,variable=self.BatchClip2ROI_var,onvalue=1,offvalue=0)
        ckbtn.place(x=BtnX+5,y=BtnY+BtnGapY+8*BtnH-10)        
        self.BatchClip2ROI_var.set(0)
        #Add a checkbutton for Batch save: Heatmap image
        self.BatchSaveImgHeatmap_var = tk.IntVar()
        ckbtn = tk.Checkbutton(self.frame2,text='Heatmap,',command=self.btn_BatchSaveImgHeatmap, bd=4,variable=self.BatchSaveImgHeatmap_var,onvalue=1,offvalue=0)
        ckbtn.place(x=BtnX+BtnGapX+30,y=BtnY+BtnGapY+8*BtnH-10)
        self.BatchSaveImgHeatmap_var.set(0)
        #Add a checkbutton for Batch save: HeatmapSeged image
        self.BatchSaveImgHeatmapSeged_var = tk.IntVar()
        ckbtn = tk.Checkbutton(self.frame2,text='HeatmapSeged,',command=self.btn_BatchSaveImgHeatmapSeged, bd=4,variable=self.BatchSaveImgHeatmapSeged_var,onvalue=1,offvalue=0)
        ckbtn.place(x=BtnX+BtnGapX*2+46,y=BtnY+BtnGapY+8*BtnH-10)        
        self.BatchSaveImgHeatmapSeged_var.set(0)
        #Add a checkbutton for Batch save: NDVI image
        self.BatchSaveImgNDVI_var = tk.IntVar()
        ckbtn = tk.Checkbutton(self.frame2,text='NDVI,',command=self.btn_BatchSaveImgNDVI, bd=4,variable=self.BatchSaveImgNDVI_var,onvalue=1,offvalue=0)
        ckbtn.place(x=BtnX+BtnGapX*4+34,y=BtnY+BtnGapY+8*BtnH-10)        
        #Add a checkbutton for Batch save: nmask image
        self.BatchSaveImgMask_var = tk.IntVar()
        ckbtn = tk.Checkbutton(self.frame2,text='Mask',command=self.btn_BatchSaveImgMask, bd=4,variable=self.BatchSaveImgMask_var,onvalue=1,offvalue=0)
        ckbtn.place(x=BtnX+BtnGapX*5+28,y=BtnY+BtnGapY+8*BtnH-10)        
                
        #Display radiobuttons & sliders
        Channels = [('R','Red',0),('G','Green',1),('B','Blue',2),('N','Gray',3)] # (Channel,cmap,iter)
        self.Band_var = tk.IntVar() # variable of R,G,B,N buttons
        self.Band_var.set(None)
        self.RdBtn = []
        for Channel,cmap,i in Channels:  
            self.RdBtn.append(tk.Radiobutton(self.frame2,text=Channel,variable=self.Band_var,value=i,command=self.btn_Band,bg=cmap))
            self.RdBtn[i].place(x=BtnX,y=BtnY+BtnGapY+(i+1)*BtnH+5, width=BtnW, height=BtnH)
        
            # Display R,G,B,N sliders
            self.sldL.append(tk.Scale(self.frame2,from_=0,to=255, sliderlength=20, width=9,font=('arial',8),fg=cmap,orient=tk.HORIZONTAL))
            self.sldL[i].place(x=BtnX+BtnGapX,y=BtnY+2*BtnGapY-10+i*BtnH)        
            self.sldU.append(tk.Scale(self.frame2,from_=0,to=255, sliderlength=20, width=9,font=('arial',8),fg=cmap,orient=tk.HORIZONTAL))
            self.sldU[i].set(255)
            self.sldU[i].place(x=BtnX+BtnGapX+100,y=BtnY+2*BtnGapY-10+i*BtnH)        
            #self.sldU[i].config(font=('courier',5))
            #Read the slider position thru mouse motion and click
            self.sldL[i].bind('<B1-Motion>',self.MouseMotion_SliderL)
            self.sldU[i].bind('<B1-Motion>',self.MouseMotion_SliderU)            
            self.sldL[i].bind('<ButtonRelease-1>',self.MouseMotion_SliderL)
            self.sldU[i].bind('<ButtonRelease-1>',self.MouseMotion_SliderU)
            #self.sldL[i].bind('<Button-1>',self.MouseMotion_SliderL)  #skip to avoid twice run when clicked and released
            #self.sldU[i].bind('<Button-1>',self.MouseMotion_SliderU)            
        
    #show histrogrm 
    def init_histogram(self):
        from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
        fig = plt.Figure(figsize=(3.62,4))    #3.62 to match cursor move from 0-255 to the graph w offset=58
        gs = fig.add_gridspec(10,1)
        self.hist = fig.add_subplot(gs[1:,0])  # taking space 9/10 height            
        self.hist.yaxis.set_visible(False)  #remove y-axis values        
        self.canvas = FigureCanvasTkAgg(fig, master=self.frame2)
        #self.canvas.get_tk_widget().pack(padx=5,pady=5,side=tk.LEFT,fill=tk.Y)                
        self.canvas.get_tk_widget().place(x=0, y=self.scr_h-475)
        self.hist.plot([0,255],[0, 0], linestyle='')   #dummy to stretch x-axis to 0-255 in the initial view
        
        #Read the mouse position in the histogram
        self.canvas.get_tk_widget().bind('<Motion>',self.MouseMotion_Histo) 
        #Read the mouse click on thresholding lines in the histogram
        self.canvas.get_tk_widget().bind('<B1-Motion>',self.MouseMotion_Histo_Line) 

    def mainLoop(self): # This function starts an endless running thread through the gui         
        self.root.mainloop()

    def client_exit(self):
        exit()  

def main():        
    # init gui
    gui = Window()  # class instantiation             
    # execute gui
    gui.mainLoop()  #complete the image display
    
if __name__ == '__main__':
    main()